/*
    Copyright (C) 2001  Tsinghuaeolus

    Authors : ChenJiang, YaoJinyi, CaiYunpeng, Lishi

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

	If you make any changes or have any comments we would appreciate a 
	message to yjy01@mails.tsinghua.edu.cn.
*/

#include "Interception.h"
#include "strategy.h"
#include "global.h"

/***********************   Interception   *****************************************/
Interception::Interception(){
}

void Interception::SetInterceptionInfo(){
	int i, j, num_my, num_their;

	num_my = motion.Num_MyVisiblePlayers();
	num_their = motion.Num_TheirVisiblePlayers();

	for(i = 0; i < num_my; i ++){
		MyPlayers_Interception[i] = motion.MyVisiblePlayers[i];
		getinterceptioninfo(MyPlayer(MyPlayers_Interception[i]).pos,
			ball.pos, MyPlayer(MyPlayers_Interception[i]).global_vel, ball.global_vel, 
			MyPlayer(MyPlayers_Interception[i]).bodyfacing, MyPlayer(MyPlayers_Interception[i]).IT_inf);
	}

	for(i =0; i < num_their; i ++){
		TheirPlayers_Interception[i] = motion.TheirVisiblePlayers[i];
		getinterceptioninfo(TheirPlayer(TheirPlayers_Interception[i]).pos,
		ball.pos, TheirPlayer(TheirPlayers_Interception[i]).global_vel, ball.global_vel, 
		TheirPlayer(TheirPlayers_Interception[i]).bodyfacing, TheirPlayer(TheirPlayers_Interception[i]).IT_inf);
	}

	getinterceptioninfo(Self.pos, ball.pos, Self.global_vel, ball.global_vel, Self.bodyfacing, Self.IT_inf, false);
	
/* sort players by their ability to intercept the ball */
	for(i = 0; i < num_my - 1; i++)
		for ( j = i + 1; j < num_my; j ++){
			if (MyPlayer(MyPlayers_Interception[j]).IT_inf.IT_cycles < MyPlayer(MyPlayers_Interception[i]).IT_inf.IT_cycles){
				Swap(MyPlayers_Interception[j], MyPlayers_Interception[i]);
			}
		}
	
	for(i = 0; i < num_their - 1; i++)
		for ( j = i + 1; j < num_their; j ++){
			if (TheirPlayer(TheirPlayers_Interception[j]).IT_inf.IT_cycles < TheirPlayer(TheirPlayers_Interception[i]).IT_inf.IT_cycles){
				Swap(TheirPlayers_Interception[j], TheirPlayers_Interception[i]);
			}
		}

	situation.most_promising_controller = -1;

	if (num_my > 0){
		my_min_cycles = MyPlayer(MyPlayers_Interception[0]).IT_inf.IT_cycles;
	}
	else{
		my_min_cycles = InfCycles;
	}

	if(num_their > 0){
		their_min_cycles = TheirPlayer(TheirPlayers_Interception[0]).IT_inf.IT_cycles;
	}
	else{
		their_min_cycles = InfCycles;
	}
	
	situation.min_IT_cycles = InfCycles;

	if (their_min_cycles < situation.min_IT_cycles){
		if(TheirPlayers_Interception[0] <= SP_team_size){
			situation.most_promising_controller = SP_team_size + TheirPlayers_Interception[0];
		}
		else{
			situation.most_promising_controller = TheirPlayers_Interception[0];
		}

		situation.next_interception_point = TheirPlayer(TheirPlayers_Interception[0]).IT_inf.IT_point;
		situation.min_IT_cycles = TheirPlayer(TheirPlayers_Interception[0]).IT_inf.IT_cycles;
	}

	if (my_min_cycles < situation.min_IT_cycles){
		situation.most_promising_controller = MyPlayers_Interception[0];
		situation.next_interception_point = MyPlayer(MyPlayers_Interception[0]).IT_inf.IT_point;
		situation.min_IT_cycles = MyPlayer(MyPlayers_Interception[0]).IT_inf.IT_cycles;
	}

	if (Self.IT_inf.IT_cycles < situation.min_IT_cycles){
		situation.most_promising_controller = MyNumber;
		situation.next_interception_point = Self.IT_inf.IT_point;
		situation.min_IT_cycles = Self.IT_inf.IT_cycles;
	}

	situation.Team_Control = calculatecontrol(Min(Self.IT_inf.IT_cycles, my_min_cycles), their_min_cycles);
}

/*	get interception information*/

void Interception::getinterceptioninfo(const Player& player, InterceptionInfo& inf, bool testonly){
	getinterceptioninfo(player.pos, ball.pos, player.global_vel, ball.global_vel, player.bodyfacing, inf, true);
}

void Interception::getinterceptioninfo(Vector startpoint, Vector ballpos, Vector selfvel, Vector ballvel, float selffacing, InterceptionInfo& inf, bool testonly){
	getinterceptioninfo_new(startpoint, ballpos, selfvel, ballvel, selffacing, inf, testonly);
	return;
}

IT_test_res Interception::interception_test(Ray& ballcourse, Ray& playercourse, float ballspeed, float selffacing, float& distball, InterceptionInfo & inf){
	Vector intersection_pt;
	float distplayer, min_cycles, stretch_angle, dist, conf, speed;

	ballcourse.intersection(playercourse, intersection_pt);
	distball = ballcourse.DistanceFromOrigin(intersection_pt);
	distplayer = playercourse.DistanceFromOrigin(intersection_pt);
	if (!ballcourse.InRightDir(intersection_pt) || !playercourse.InRightDir(intersection_pt)){
		return IT_test_beyond;
	}
	
	stretch_angle = (float)fabs(NormalizeAngle(ballcourse.Dir() - playercourse.Dir()));

	min_cycles = (float)Max(int(distplayer / SP_player_speed_max), 0);
	conf = Exp(SP_ball_decay, min_cycles);
	distball -= ballspeed * (1 - conf) / (1 - SP_ball_decay);
	distplayer -= SP_player_speed_max * min_cycles;
	if (distball < 0){
	//go beyond
		dist = (float)sqrt(Sqr(distball) + Sqr(distplayer) + 2 * fabs(distball) * distplayer * Cos(stretch_angle));
		if (dist > CP_reliable_kickarea)
			return IT_test_beyond;
	}
	else{
	//see if it's too far away to wait for
		speed = ballspeed * conf;
		if (distball > (1 + SP_ball_decay) * speed + CP_reliable_kickarea)
			return IT_test_not_desired;
		dist = (float)sqrt(Sqr(distball) + Sqr(distplayer) - 2 * fabs(distball) * distplayer * Cos(stretch_angle));
		if (dist > CP_reliable_kickarea){
			if (distball - speed < -CP_reliable_kickarea){
				//go beyond this cycle
				return IT_test_beyond;
			}
		}
	}
	inf.IT_angle = selffacing;
	inf.IT_cycles = min_cycles;
	inf.IT_point = intersection_pt;
	inf.IT_is_closeball = false;
	inf.IT_rapidness = SP_player_speed_max;
	inf.IT_valid = true;
	return IT_test_ok;
}

inline void Interception::InfCycleProcess(float& cycles){
	if (fabs(cycles) > InfCycles || cycles < 0)
		cycles = InfCycles;
}

void Interception::getinterceptioninfo_new(Vector startpoint, Vector ballpos, Vector selfvel, Vector ballvel, float selffacing, InterceptionInfo& inf, bool testonly){
	IT_test_res it_test_res_noturn; 
	Ray ballcourse, playercourse;
	Vector gap;
	float distball_noturn, ballspeed, angle, ballangle;
	InterceptionInfo inf_noturn;

	if (!testonly){
	//test closeball interception
		if (getcloseballinterceptioninfo(startpoint, ballpos, selfvel, ballvel, selffacing, inf, testonly))
			return;
	}
	inf.IT_cycles = InfCycles;
	inf.IT_valid = false;
	inf.IT_rapidness = Self.max_speed;
	inf.IT_is_closeball = false;

	ballcourse.SetValues(ballpos, ballvel);
	playercourse.SetValues(startpoint, selffacing);
	ballspeed = ballvel.mod();
	ballangle = ballvel.Angle();
	gap = ballpos - startpoint;
	angle = gap.Angle();
	
	if (ballspeed < 0.15f){
		inf.IT_angle = angle;
		inf.IT_point = ballpos;
		inf.IT_cycles = float(int(gap.mod() / inf.IT_rapidness));
		InfCycleProcess(inf.IT_cycles);
		inf.IT_valid = true;
		return;
	}

	if (ballcourse.InRightDir(startpoint)){
		inf.IT_angle = angle;
		inf.IT_cycles = 0;
		inf.IT_point = ballpos;
		int cycles = 1, counter = 0; // initialize to be > 0
		float dist = gap.mod(), conf;
		while(dist > CP_reliable_kickarea && counter < 20){
			cycles = Max(int(dist / (ballvel.mod() + Self.max_speed)), 1);
			inf.IT_cycles += cycles;
			conf = Exp(ball.speed_decay, float(cycles));
			dist -= Self.max_speed * cycles + ballvel.mod() * (1 - conf) / ( 1- ball.speed_decay);
			inf.IT_point += ballvel * (1 - conf) / ( 1- ball.speed_decay);
			ballvel *= conf;
			counter ++;
		}
		InfCycleProcess(inf.IT_cycles);
		inf.IT_valid = true;
		return;
	}
	else if (ballcourse.InOppositeDir(startpoint)){
		inf.IT_angle = angle;
		inf.IT_cycles = 0;
		inf.IT_point = ballpos;
		int cycles, counter = 0;
		float dist = gap.mod(), conf;
		while(dist > CP_reliable_kickarea && counter < 20){
			cycles = Max(int(dist / Self.max_speed), 1);
			inf.IT_cycles += cycles;
			conf = Exp(ball.speed_decay, float(cycles));
			dist -= SP_player_speed_max * cycles - ballvel.mod() * (1 - conf) / ( 1- ball.speed_decay);
			inf.IT_point += ballvel * (1 - conf) / ( 1- ball.speed_decay);
			ballvel *= conf;
			counter ++;
		}
		InfCycleProcess(inf.IT_cycles);
		inf.IT_valid = true;
		return;
	}

	if (!testonly){
		if ((it_test_res_noturn = interception_test(ballcourse, playercourse, ballspeed, selffacing, distball_noturn, inf_noturn)) == IT_test_ok){
			DoLog(LOG_IT, "test ok");
			inf = inf_noturn;
			return;
		}
	}
	
	//should turn to intercept	
	ballpos += ballvel;
	ballvel *= SP_ball_decay;
	ballspeed *= SP_ball_decay;
	startpoint -= ballpos;
	startpoint = startpoint.Rotate(- ballangle);
	Vector inf_sup;
	float t;
	
	if (calc_interception_solution(startpoint, ballspeed, inf_sup)){
		t = inf_sup.y;
	}
	else{
		t = InfCycles;
		if (inf_sup.x > inf_sup.y){
			DoLog(LOG_IT, "Bug why inf > sup (%.2f %.2f)", inf_sup.x, inf_sup.y);
		}
		if (inf_sup.y == InfCycles){
			DoLog(LOG_IT, "see if really cannot get the ball(%.2f %.2f)", inf_sup.x, inf_sup.y);
		}
	}
	if (t == InfCycles){
		inf.IT_cycles = InfCycles;
		inf.IT_valid = true;
	}
	else{
		if (t < 0){
			DoLog(LOG_IT, "bug why t < 0");
		}
		inf.IT_angle = NormalizeAngle((Vector(t, 0) - startpoint).Angle() + ballangle);
		if (!testonly)
			DoLog(LOG_IT, "ballangle %.2f inf angle %.2f", ballangle, inf.IT_angle);
		ballcourse.SetValues(ballpos, ballangle);
		ballcourse.Normalize();
		inf.IT_point = ballcourse.GetPoint(t);
		inf.IT_cycles = (float)sqrt(Sqr(startpoint.x - t) + Sqr(startpoint.y)) / SP_player_speed_max;
		inf.IT_valid = true;
	}
}

bool Interception::getcloseballinterceptioninfo(Vector selfpos, Vector ballpos, Vector selfvel, Vector ballvel, float selffacing, InterceptionInfo &inf, bool testonly){
//more sophisticated design can be added here, compare onlydashintercept with turn intercept

	if (!ball.pos_valid()) return false;
	Command command;
	float cycles = (float)OnlyDashIntercept(5, selfpos, ballpos, selfvel, ballvel, selffacing, command, testonly);
	if (cycles != -1){
	//Only dashs will do
	//patches
		if (cycles == 2 && fabs(last_cycle_test_res.y) > 0.75){
			InterceptionInfo t_inf;
			int turn_cycles = TurnDashIntercept(3, selfpos, ballpos, selfvel, ballvel, selffacing, t_inf);
			if (turn_cycles > 0 && turn_cycles <= 3 && fabs(last_cycle_test_res.y) < 0.8){
				inf = t_inf;
				DoLog(LOG_IT, "patches applyed---turn dash");
				return true;
			}
		}

		DoLog(LOG_IT, "(close_ball) Only dash %.2f", cycles);
		inf.IT_is_closeball = true;
		inf.IT_dashpow = command.GetPower();
		inf.IT_angle = 360.0f; // disable turn
		inf.IT_valid = true;
		inf.IT_cycles = cycles;
		inf.IT_point = ballpos +ballvel * (1 - Exp(SP_ball_decay, cycles)) / (1 - SP_ball_decay);
		return true;
	}
//dash backwards
	cycles = (float)OnlyDashIntercept(3, selfpos, ballpos, selfvel, ballvel, selffacing + 180, command, testonly);
	if ( cycles != -1){
	//Only dashs will do
		DoLog(LOG_IT, "(close_ball) backward Only dash %.2f", cycles);
		inf.IT_is_closeball = true;
		inf.IT_dashpow = -command.GetPower();
		inf.IT_angle = 360.0f; // disable turn
		inf.IT_valid = true;
		inf.IT_cycles = cycles;
		inf.IT_point = ballpos +ballvel * (1 - Exp(SP_ball_decay, cycles)) / (1 - SP_ball_decay);
		return true;
	}
//if turn angle changes little, please
	cycles = (float)TurnDashIntercept(5, selfpos, ballpos, selfvel, ballvel, selffacing, inf);
	if (cycles != -1){
		return true;
	}
	
	DoLog(LOG_IT, "No CloseBall");
	return false;
}

bool Interception::OneCycleIntercept(Vector selfpos, Vector ballpos, Vector selfvel, Vector ballvel, float selfbodyfacing, Command& command, bool testonly){
	float dashpow = 2 * SP_max_power, min_dist, maxdasheffect, kickable_area;

	Vector ballrelpos = ballpos + ballvel- selfpos - selfvel;
	
	ballrelpos = ballrelpos.Rotate( - selfbodyfacing);
	
	if (testonly)
		kickable_area = CP_reliable_kickarea;
	else
		kickable_area = SP_kickable_area;

	if (fabs(ballrelpos.y) > kickable_area) return false;
	maxdasheffect = action.Max_DashEffect();
	if (ballrelpos.mod2() > Sqr(kickable_area + maxdasheffect)) return false;

	DoLog(LOG_IT, "(LastCycle) (%.2f %.2f) testonly %d", ballrelpos.x, ballrelpos.y, testonly);
	last_cycle_test_res = ballrelpos;
	
	if (maxdasheffect > fabs(ballrelpos.x)){
		if (testonly)
			return true;
		
		if (fabs(ballrelpos.y) >= CP_desired_kickarea){
			dashpow = float(fabs(ballrelpos.x) / action.Dashrate());
			DoLog(LOG_IT, "1");
		}
		else{
		//There are two solutions, choose the one that is more robust
			float solution[2], dist[2];
			int final_solution = 0;
			solution[0] = float(fabs(ballrelpos.x) - sqrt(Sqr(CP_desired_kickarea) - Sqr(ballrelpos.y)));
			solution[1] = 2 * (float)fabs(ballrelpos.x) - solution[0];
			pose_limitation(solution[0], -maxdasheffect, maxdasheffect);
			solution[1] = Min(solution[1], maxdasheffect);
			dist[0] = float(sqrt(Sqr(solution[0] - fabs(ballrelpos.x)) + Sqr(ballrelpos.y)));
			dist[1] = float(sqrt(Sqr(solution[1] - fabs(ballrelpos.x)) + Sqr(ballrelpos.y)));
			if (dist[0] < CP_collision_threshold || dist[1] < CP_collision_threshold){
				//if one solution leads to a collision, choose the bigger one
				if (dist[0] > dist[1])
					final_solution = 0;
				else
					final_solution = 1;
			}
			else{
			//try to intercept with front side
				ballvel = ballvel.Rotate(-selfbodyfacing);
				Vector next_ball_pos = ballvel + ballrelpos;
				if (Vector(solution[0], 0).dist2(next_ball_pos) > Vector(solution[1], 0).dist2(next_ball_pos))
					final_solution = 1;
				else
					final_solution =0;
			}
			
			dashpow = solution[final_solution] / action.Dashrate();
			DoLog(LOG_IT, "2");
		}
	}
	else{
		min_dist = (float)sqrt(Sqr(fabs(ballrelpos.x) - maxdasheffect) + Sqr(ballrelpos.y));
		if (min_dist < kickable_area){
			if (testonly)
				return true;
			if (min_dist >= CP_desired_kickarea){
				dashpow = SP_max_power;
				DoLog(LOG_IT, "3");
			}
			else{
				dashpow = Max(float(fabs(ballrelpos.x) - sqrt(Sqr(CP_desired_kickarea) - Sqr(ballrelpos.y))) / action.Dashrate(), -SP_max_power);
				DoLog(LOG_IT, "4");
			}
		}
	}

	if (testonly){
		DoLog(LOG_IT, "maxdasheffect %.2f min_dist %.2f", maxdasheffect, min_dist);
	}

	if (dashpow == 2 * SP_max_power) return false;
	if (ballrelpos.x < 0) dashpow *= -1.0f;
	command.dash(dashpow);
	return true;
}

int Interception::Close_to_ball_line_intercept(Vector selfpos, Vector ballpos, Vector selfvel, Vector ballvel, 
											   float selfbodyfacing, float dist_player, float dist_ball, Command& command, bool testonly){
//Try just 3 cycles, last cycle is a adjusting cycle
	DoLog(LOG_IT, "bdfy %.2f ballrelpos (%.2f %.2f) selfvel (%.2f %.2f) ballvel (%.2f %.2f) ", selfbodyfacing, (ballpos - selfpos).x, (ballpos - selfpos).y,
		selfvel.x, selfvel.y, ballvel.x, ballvel.y);
	DoLog(LOG_IT, "distplayer %.2f  dist_ball %.2f testonly %d", dist_player, dist_ball, testonly);
	float dashpow;
	for(int i = 1; i <= 3; i ++){
		if (OneCycleIntercept(selfpos, ballpos, selfvel, ballvel, selfbodyfacing, command, testonly || i != 1))
			break;
		//do not move beyond the ball line
		
		dashpow = Min(SP_max_power, (dist_player - selfvel.mod()) / action.Dashrate());
		
		simulate_one_dash_step(selfpos, ballpos, selfvel, ballvel, selfbodyfacing, dashpow);
		dist_player -= selfvel.mod()/SP_player_decay ;
		dist_ball -= ballvel.mod() / SP_ball_decay;
	}

	if (i == 4) // I am not able to intercept it in three cycles
		return -1;

	if (testonly)
		return i;

	if (i != 1){
	//more delicate control can be added here
		if (i == 2){
		// to avoid collision
			dashpow += TwoCycleAvoidEnemy_Adjust(selfpos, ballpos, selfvel, ballvel, selfbodyfacing);
		}
		DoLog(LOG_IT, "Close to ball line dash pow %.2f %d %.2f", dashpow, i, dist_player);
		command.dash(dashpow);
	}
	return i;
}

int Interception::OnlyDashIntercept(int maxcycles, Vector selfpos, Vector ballpos, Vector selfvel, Vector ballvel, float selfbodyfacing, Command& command, bool testonly){
//Ok, see if dashes will bring me close to the ball
	if (maxcycles <= 0)
		return -1;

	if (OneCycleIntercept(selfpos, ballpos, selfvel, ballvel, selfbodyfacing, command, testonly))
		return 1;

	Vector intersection_pt;
	Ray ball_course(ballpos, ballvel), player_course(selfpos, selfbodyfacing);
	float dist_ball, dist_player;

	//selffacing is parallel to ball course, unable to get intersection point
	if (ball_course.InRightDir(selfbodyfacing) || ball_course.InOppositeDir(selfbodyfacing)){
		if (ballvel.mod2() > 0.3f){
			if ((ball_course.InRightDir(selfbodyfacing) && ball_course.InRightDir(selfpos))
				|| (ball_course.InOppositeDir(selfbodyfacing) && ball_course.InOppositeDir(selfpos))){
				return -1;
			}
		}
		for(int i = 2; i <= maxcycles; i ++){
			simulate_one_dash_step(selfpos, ballpos, selfvel, ballvel, selfbodyfacing);
			if (OneCycleIntercept(selfpos, ballpos, selfvel, ballvel, selfbodyfacing, command, testonly || i != 1))
				break;				
		}
		if (i <= maxcycles){
			if (!testonly){
				if (i == 2){
					command.dash(SP_max_power + TwoCycleAvoidEnemy_Adjust(selfpos, ballpos, selfvel, ballvel, selfbodyfacing));
				}
				else
					command.dash(SP_max_power);
			}
			return i;
		}
		return -1;
	}
	
	ball_course.intersection(player_course, intersection_pt);
	dist_ball = ballpos.dist(intersection_pt);
	dist_player = selfpos.dist(intersection_pt);
	
	/*DoLog(LOG_IT, "bdfy %.2f ballrelpos (%.2f %.2f) selfvel (%.2f %.2f) ballvel (%.2f %.2f) ", selfbodyfacing, (ballpos - selfpos).x, (ballpos - selfpos).y,
		selfvel.x, selfvel.y, ballvel.x, ballvel.y);
	DoLog(LOG_IT, "distplayer %.2f  dist_ball %.2f", dist_player, dist_ball);*/
// intercept with more than maxcycles is beyond the reach of this function
	if (dist_player > maxcycles * Self.max_speed + SP_kickable_area)
		return -1;
	if (dist_ball  > ballvel.mod() * (1 - Exp(ball.speed_decay, (float)maxcycles)) / (1 - ball.speed_decay) + SP_kickable_area)
		return -1;
	if (!player_course.InRightDir(intersection_pt) || !ball_course.InRightDir(intersection_pt)) // no hope
		return -1;
//simulate running num_cycles cycles
	int num_cycles = int((dist_player - SP_kickable_area) / Self.max_speed);
	for(int i =0; i < num_cycles; i++){
		simulate_one_dash_step(selfpos, ballpos, selfvel, ballvel, selfbodyfacing);
		dist_ball -= ballvel.mod() / SP_ball_decay;
		dist_player -= selfvel.mod() / SP_player_decay;
		if (dist_ball < -SP_kickable_area) // the ball is passing by
			return -1;
	}
	
	num_cycles = Close_to_ball_line_intercept(selfpos, ballpos, selfvel, ballvel, selfbodyfacing, dist_player, dist_ball, command, testonly || i != 0);
	if (num_cycles == -1) return -1;
	if (i > 0)
		motion.run(Self.max_speed, command);

	return i + num_cycles;
}

int Interception::TurnDashIntercept(int maxcycles, Vector selfpos, Vector ballpos, Vector selfvel, Vector ballvel, float selfbodyfacing, InterceptionInfo & inf, bool test){
	Vector predicted_selfpos, predicted_ballpos, predicted_selfvel, predicted_ballvel;
	int cycles;
	Command cmd;

	simulate_one_dash_step(selfpos, ballpos, selfvel, ballvel, 0, 0);
 
	predicted_selfpos = selfpos; predicted_ballpos = ballpos;
	predicted_selfvel = selfvel; predicted_ballvel = ballvel;

	float turnangle = 0, lastturnangle;
	for(int i = 1; i < maxcycles; i ++){
		lastturnangle = turnangle;
		turnangle = (predicted_ballpos + predicted_ballvel - predicted_selfpos - predicted_selfvel).Angle();
		if (i != 1 && fabs(NormalizeAngle(lastturnangle - turnangle)) < 8.0f)
			break;
		//DoLog(LOG_IT, "turn ang %.2f", turnangle);
		cycles = OnlyDashIntercept(maxcycles, selfpos, ballpos, selfvel, ballvel, turnangle, cmd, true);
		//DoLog(LOG_IT, "cycles %.2f", cycles);
		if (cycles != -1 && cycles <= i){
			DoLog(LOG_IT, "(close_ball) turn dash %d", cycles +1);
			inf.IT_is_closeball = true;
			inf.IT_angle = turnangle;
			inf.IT_cycles = float(cycles + 1);
			inf.IT_point = ballpos + ballvel * (1 - Exp(SP_ball_decay, (float)cycles)) / (1 - SP_ball_decay);
			inf.IT_valid = true;
			return cycles;
		}
		simulate_one_dash_step(predicted_selfpos, predicted_ballpos, predicted_selfvel, predicted_ballvel, 0, 0);
	}
	return -1;
}

float Interception::TwoCycleAvoidEnemy_Adjust(Vector selfpos, Vector ballpos, Vector selfvel, Vector ballvel, float selfbodyfacing){
	Vector ballrelpos = ballpos + ballvel - selfpos - selfvel;
	float t, t_dashpow = 0.0f;
	if (ballrelpos.mod2() < Sqr(CP_collision_threshold)){
		ballrelpos = ballrelpos.Rotate(- selfbodyfacing);
		t = ballrelpos.x - (float)sqrt(Sqr(CP_desired_kickarea) - Sqr(ballrelpos.y));
		t_dashpow = t / ( action.Dashrate() * (1 + SP_player_decay));
		DoLog(LOG_IT, " two cycle adjust %.2f pow %.2f", t, t_dashpow);
	}
	return t_dashpow;
}

void Interception::simulate_one_dash_step(Vector& selfpos, Vector& ballpos, Vector& selfvel, Vector& ballvel, float bodyfacing, float dashpow){
	selfvel += Polar2Vector(dashpow, bodyfacing) * action.Dashrate();
	if (selfvel.mod2() > Sqr(SP_player_speed_max)){
		selfvel = selfvel / selfvel.mod() * SP_player_speed_max;
	}
	selfpos += selfvel;
	ballpos += ballvel;
	
	selfvel *= SP_player_decay;
	ballvel *= SP_ball_decay;
}

IT_mode Interception::intercept(InterceptionInfo& inf){
	Command command;
	IT_mode it_mode = intercept(inf, command);
	mediator.enroll(command, Action_interception, PriorityA);
	return it_mode;
}

IT_mode Interception::intercept(InterceptionInfo& inf, Command& command){
	//intercept action
	command.Reset();
	if (inf.IT_is_closeball){
		if (fabs(inf.IT_angle) > 180.0f){
			command.dash(inf.IT_dashpow);
		}
		else{
			motion.turn_to(inf.IT_angle, command);
		}
		DoLog(LOG_IT, "closeballinterception");
		return IT_closeball;		
	}
	else{
		if (inf.IT_valid && fabs(NormalizeAngle(inf.IT_angle - Self.bodyfacing)) > 15 - 8*exp(-ball.distance * 0.15f)){
			motion.turn_to(inf.IT_angle, command);
			return IT_turn;
		}
		else{
			motion.run(inf.IT_rapidness, command);
			return IT_dash;
		}
	}
}

bool Interception::SmartInterception(){
	//first decide whether the player should intercept the ball
	//if he is in interception, decide whether he should cancel it
	Command command;

	//the sight information is sufficient to make interception decision
	if (ball.pos_conf < 0.75 || ball.speed_conf < 0.75) return false;
	
	//ball is in someone's control, don't be hook up
	if (!situation.BallFree) return false;

	DoLog(LOG_IT, "most promising IT player%d, cycle%.2f at (%.2f %.2f)",situation.most_promising_controller,
		GetPlayer(situation.most_promising_controller).IT_inf.IT_cycles, GetPlayer(situation.most_promising_controller).pos.x, GetPlayer(situation.most_promising_controller).pos.y);
	DoLog(LOG_IT, "My Cycles %.2f", Self.IT_inf.IT_cycles);
	
	// can not get the ball
	if (Self.IT_inf.IT_cycles == InfCycles ) return false;

	//if the ball is definitely out of pitch before i can get there, let it go
	if(!fieldinfo.WithInExpandedField(Self.IT_inf.IT_point)){
		DoLog(LOG_IT, "Out of pitch(%.2f,%.2f)",Self.IT_inf.IT_point.x,  Self.IT_inf.IT_point.y);
		return false;
	}

	float control;
//see if I can get ball from opponent
	int num = motion.Num_TheirVisiblePlayers();
	for(int i = 0; i < num; i ++){
		control = calculatecontrol(Self.IT_inf, TheirPlayer_CtrlOfBall(i).IT_inf);
		if(control >= 0.45f){
		//there is great chance
			break;
		}

		if(Self.aggressiveness > 0.5f){
			if (control >= 0.3f)
				break;
		}
		if (WouldIntercept(situation.TheirSide, TheirPlayers_Interception[i])){
				// there is no chance for me to get the ball from opponent
			DoLog(LOG_IT, "There is no chance to get ball from %d whose cycles is %.2f", TheirPlayers_Interception[i], TheirPlayer_CtrlOfBall(i).IT_inf.IT_cycles);
			return false;
		}
	}

	num = motion.Num_MyVisiblePlayers();

//check if there is need for me to intercept that ball
	for(i = 0; i < num; i ++){

		if (motion.MyPlayer_CtrlOfBall(i).Is_goalie 
			|| MyNumber == MyPlayers_Interception[i]) continue;

		control = calculatecontrol(Self.IT_inf, MyPlayer_CtrlOfBall(i).IT_inf, true);

		if (control > 0.55f){
		//the player with the best chance of getting the ball should go
			DoLog(LOG_IT, "I have better chance than %d", MyPlayers_Interception[i]);
			break;
		}
		else if (WouldIntercept(situation.MySide, MyPlayers_Interception[i])){
			if (control < 0.45f){
				DoLog(LOG_IT, "%d has a better chance whose cycle is %f", MyPlayers_Interception[i], MyPlayer_CtrlOfBall(i).IT_inf.IT_cycles);
				return false;
			}
			else if (!Is_IT_Priored(MyNumber, MyPlayers_Interception[i])){
			//is it dangerous
				DoLog(LOG_IT, "%d is equal in chance but it is priored whose cycle is %.2f", MyPlayers_Interception[i], MyPlayer_CtrlOfBall(i).IT_inf.IT_cycles);
				return false;
			}
			else{
				DoLog(LOG_IT, "I am priored than %d", MyPlayers_Interception[i]);
				break;
			}
		}
	}
	

	/* Interception Action */
	intercept(Self.IT_inf, command);
	
	if (Self.IT_inf.IT_cycles == 1.0f && Self.IT_inf.IT_is_closeball){
		mediator.IsLastCycleIntercept = true;
	}
	else{
		mediator.IsLastCycleIntercept = false;
	}
	mediator.enroll(command, Action_interception, 2 * PriorityA);
	return true;
}

bool Interception::Is_IT_Priored(UNum player1, UNum player2){
	if (!MyPlayer(player1).IsRoleKnown()) return false;
	if (!MyPlayer(player2).IsRoleKnown()) return true;
	Vector point1, point2, ballpos;
	point1 = fm.GetFormationPoint(MyPlayer(player1));
	point2 = fm.GetFormationPoint(MyPlayer(player2));

	ballpos = situation.NextControlBallPos();
	return ballpos.dist2(point1) < ballpos.dist2(point2);
}


bool Interception::WouldIntercept(char side, UNum No){
	if (GetPlayer(side, No).Is_goalie){
		return fieldinfo.their_penaltyarea.IsWithin(GetPlayer(side, No).IT_inf.IT_point);
	}
	return true;
}

float Interception::calculatecontrol(InterceptionInfo& myinf, InterceptionInfo& theirinf, bool is_teammate){
	if (myinf.IT_is_closeball && !is_teammate){
		return calculatecontrol(myinf.IT_cycles, (float)int(theirinf.IT_cycles) + 2.0f);
	}
	return calculatecontrol(myinf.IT_cycles, theirinf.IT_cycles);
}

float Interception::calculatecontrol(float my_cycles, float their_cycles){
	if(my_cycles == InfCycles){
		if (their_cycles == InfCycles)
			return 0.5f;
		else
			return 0.0f;
	}
	else if (their_cycles == InfCycles)
		return 1.0f;
	double input[2];
	double control;
	float cycles_gap = (float)fabs(my_cycles - their_cycles), min_cycles = Min(my_cycles, their_cycles);
	input[0] = fabs(my_cycles - their_cycles) /10;
	input[1] = fabs(Min(my_cycles, their_cycles)) / 10;
	ControlBallNet.SimulateNet(&input[0], &control);
	control = Max(control, 0.5);
	
	if (my_cycles <= their_cycles)
		return  (float)control;
	else
		return 1 - (float)control;
}


Player& Interception::MyPlayer_CtrlOfBall(int idx){
	return MyPlayer(MyPlayers_Interception[idx]);
}

Player& Interception::TheirPlayer_CtrlOfBall(int idx){
	return TheirPlayer(TheirPlayers_Interception[idx]);
}
