/*
    Copyright (C) 2001  Tsinghuaeolus

    Authors : ChenJiang, YaoJinyi, CaiYunpeng, Lishi

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

	If you make any changes or have any comments we would appreciate a 
	message to yjy01@mails.tsinghua.edu.cn.
*/

#include "stdafx.h"

#include "Objects.h"
#include "global.h"

/*******************************   Object   **************************/
Object::Object(){
	pos_conf = 0;
	pos_time = -1;
	coordinate_time = -1;
	global_angle = 0;
}


void Object::set_polar(AngleDeg ang,Time time){
	//do not exchange the order of the next two statements
	seenposinfo.Setdata(Vector(seenposinfo.Data(seentime).x, ang), time);
	seentime = time;
	seen = true;
}

void Object::set_polar(AngleDeg ang,float d,Time time){
	seenposinfo.Setdata(Vector(d, ang), time);
	seentime = time;
	sdtime = time;
	seen = true;
}

void Object::set_pos(float X,float Y,bool rotate){
	if (!rotate)
		pos = Vector(X,Y);
	else
		pos = Vector(-X,-Y);
}

void Object::set_global_angle(AngleDeg ang,bool rotate){
	if(!rotate)
		global_angle = ang;
	else
		global_angle = NormalizeAngle(180 + ang);
}

void Object::set_conf_param(float max,float decay,float min){
	max_conf = max;
	conf_decay = decay;
	min_valid_conf = min;
}

void Object::set_objtype(ObjType objtype){
	this->objtype = objtype;
}

bool Object::pos_valid(){
	return bool(pos_conf >= min_valid_conf);
}
/***************************   Mobile Object       *********************/
MobileObject::MobileObject(){
	Object::Object();
	speed_conf = 0;
	speed_conf = 0;
	global_vel = 0;
	update_seen_time = -1;
	seentime = -1;
	sdtime = -1;
}

void MobileObject::set_chinfo(float distChng,float dirChng,Time time){
	seenchinfo.Setdata(Vector(distChng, dirChng), time);
	seenchtime = time;
	seenmoving = true;
}

Vector MobileObject::Pos(Time time){
	return gpos.IsDataKnown(time) ? gpos.Data(time) : pos;
}

Vector MobileObject::Vel(Time time){
	return gvel.IsDataKnown(time) ? gvel.Data(time) : global_vel;
}

float MobileObject::ActiveRadius(Time time){
	return rsradius.ChkData(time);
}

void MobileObject::update_end(){
	if(sensory.NewSight && !seen && inview(sensory.LastSightTime) && distance < SP_team_far_length){
		DoLog(LOG_UPDATE, "forget obj (%.2f %.2f)", pos.x, pos.y);
		forget();
	}

	pos = Pos(situation.CurrentTime);
	global_vel = Vel(situation.CurrentTime);
	rel_pos = pos - Self.pos;
	rel_pos_2_bodyfacing = rel_pos.Rotate(- Self.bodyfacing);
	rel_pos_2_headfacing = rel_pos.Rotate(- Self.headfacing);
	global_angle = rel_pos.Angle();
	distance = rel_pos.mod();
	rel_angle_2_bodyfacing = NormalizeAngle(global_angle - Self.bodyfacing);
	rel_angle_2_headfacing = NormalizeAngle(global_angle - Self.headfacing);

	speed = global_vel.mod();
	vel_angle = global_vel.Angle();
	rel_vel = global_vel.Rotate(- Self.bodyfacing);
	seen = false;
	seenmoving = false;
	heardupdate = false;
}

void MobileObject::update(Time time){
	newpos = false;
	newvel = false;

	update_seen(time);
	update_heard(time);
	
	estimate(time);
	pos = Pos(sensory.LastSightTime);
	rel_pos = pos - Self.Pos(sensory.LastSightTime); 
	distance = rel_pos.mod();
}

void MobileObject::update_seen(Time time){
	if (!sensory.NewSight) return;
	/*       Position         */
	if(seen){
		Time last_update_seen_time = Max(update_seen_time, time - 2);
		Vector pos, rel_vel, rpos;
		for(int t = last_update_seen_time + 1; t <= time; t ++){
			if (seenposinfo.IsDataKnown(t)){
				rpos = Polar2Vector(seen_distance(t), seen_rel_angle(t)); 
				pos = rpos.Rotate(Self.Headfacing(t)) + Self.Pos(t);
				estimate(t -1);
				gpos.Setdata(pos, t);
				pos_time = t;
				coordinate_time = t;
				pos_conf = max_conf;
				newpos = true;
				original_postime = t;
				if (objtype == OBJ_Ball)
					DoLog(LOG_UPDATE,"See Ball(%.2f %.2f) at %d", Pos(t).x, Pos(t).y, t);
				update_seen_time = t;
				
				/*****************/
				rspos.Setdata(pos, t);
				rsradius.Setdata(0,t);
				rspos_time = t;
				rs_valid = true;
				forgot = false;
				/*       Velocity         */
				if (seenchinfo.IsDataKnown(t)){
					rpos.Normalize();
					rel_vel.x = seen_distCh(t) * rpos.x - (seen_dirCh(t) * PI / 180.0f * seen_distance(t) * rpos.y);
					rel_vel.y = seen_distCh(t) * rpos.y + (seen_dirCh(t) * PI / 180.0f * seen_distance(t) * rpos.x);

					gvel.Setdata(Self.Vel(t) + rel_vel.Rotate(Self.Headfacing(t)), t);
					vel_time = t;
					speed_conf = max_conf;
					if (objtype == OBJ_Ball)
						DoLog(LOG_UPDATE, "See Ball Moving with (%.2f %.2f) at %d", Vel(t).x, Vel(t).y, t);
					
					newvel = true;
					original_veltime = t;
				}
			}
		}
	}
}

void MobileObject::estimate(Time time){
	Time t;
	if(vel_time < pos_time - 5){//for unknown player, maybe no old vel;
		vel_time = pos_time;
		gvel.Setdata(0, vel_time);
		speed_conf = 0;
	}
	for(t = vel_time; t < pos_time; t ++){
		estimate_vel(t + 1);
	}
	if(sensory.LastSightTime > coordinate_time && coordinate_time > 0 && gpos.IsDataKnown(coordinate_time)){
		float posconfbk = pos_conf;
		for(t = coordinate_time; t < pos_time; t++){			
			estimate_pos(t + 1);
		}
		pos_conf = posconfbk;
	}
	for(t = pos_time; t < time; t ++){
		estimate_vel(t + 1);
		estimate_pos(t + 1);
	}
}

void MobileObject::estimate_vel(Time time){
	Vector evel = gvel.ChkData(time -1);

	if (objtype == OBJ_Ball)
		evel += action.KickEffect(time -1);
	if (evel.mod() > max_speed){
		evel = evel / evel.mod() * max_speed;
	}
	evel *= speed_decay;
	gvel.Setdata(evel, time);
	speed_conf *= conf_decay;
	vel_time = time;
	if (objtype == OBJ_Ball){
		DoLog(LOG_UPDATE, "Estimate Ball Moving with (%.2f %.2f) at %d", Vel(time).x, Vel(time).y, time);
	}
}

void MobileObject::estimate_pos(Time time){
	gpos.Setdata(Pos(time -1) + gvel.ChkData(time) / speed_decay + Self.pos_gap.ChkData(time) , time);
	pos_conf *= conf_decay;
	pos_time = time;
	if(Self.pos_gap.IsDataKnown(time)){
		coordinate_time = time;
		if (objtype == OBJ_Ball){
			DoLog(LOG_UPDATE,"coordinate time %d", time);
		}
	}
	if (objtype == OBJ_Ball){
		DoLog(LOG_UPDATE, "Estimate Ball (%.2f %.2f) at %d", Pos(time).x, Pos(time).y, time);
	}
}

bool MobileObject::Heardupdate(){
	return heard;
}

void MobileObject::update_heard(Time time){
	if (Heardupdate()){
		gpos.Setdata(heardposinfo.data, heardposinfo.time);
		pos_time = heardposinfo.time;
		coordinate_time = pos_time;
		pos_conf = Exp(conf_decay, float(pos_time - heard_org_postime)) * max_conf;
		original_postime = heard_org_postime;
		
		newpos = true;
		heardupdate = true;
		
		rspos_time = heardposinfo.time;
		rspos.Setdata(heardposinfo.data, rspos_time);
		rsradius.Setdata((rspos_time - heard_org_postime) * max_speed, rspos_time);
		rs_valid = true;
		forgot = false;
		
		if(objtype == OBJ_Ball){
			DoLog(LOG_HEAR, "ball pos is heard at %d (%.2f %.2f)", original_postime, heardposinfo.data.x, heardposinfo.data.y);
			//take care, here we set player's iscontrolball
			for(int No = 1; No <= SP_team_size; No++){
				MyPlayer(No).Set_Iscontrolball_Fromheard(heardposinfo.time);
				TheirPlayer(No).Set_Iscontrolball_Fromheard(heardposinfo.time);
				unknowntheirplayerlist.Set_IsControlball_FromHeard(heardposinfo.time);
				unknownmyplayerlist.Set_IsControlball_FromHeard(heardposinfo.time);
				teamlessplayerlist.Set_IsControlball_FromHeard(heardposinfo.time);
			}
		}
		
		if (heard_org_veltime > original_veltime){
			gvel.Setdata(heardvelinfo.data, heardvelinfo.time);
			vel_time = heardvelinfo.time;
			speed_conf = Exp(conf_decay, float(vel_time - heard_org_veltime)) *  max_conf;
			original_veltime = heard_org_veltime;
			
			newvel = true;
			if(objtype == OBJ_Ball){
				DoLog(LOG_HEAR, "ball vel is heard at %d (%.2f %.2f)", original_veltime, heardvelinfo.data.x, heardvelinfo.data.y);
			}
		}
	}
	heard = false;
}

bool MobileObject::inview(Time time){
	if (!pos_valid()) return false;

	if (pos_time < time){
		estimate(time);
	}
	Vector tmp_pos = (Pos(time) - Self.Pos(time)).Rotate(-Self.Headfacing(time));
	if(tmp_pos.mod() < SP_feel_distance * 0.5f) return true; 
	//DoLog(7,"The Object is assumed to be at rel(%.2f %.2f)", tmp_pos.x, tmp_pos.y);
	float slide_dist = 1.0f / Sin(sensory.MyViewAngle(Self.viewwidth.Data(time)));
	tmp_pos -= Vector(slide_dist, 0.0f);
	return bool(fabs(tmp_pos.Angle()) <= sensory.MyViewAngle(Self.viewwidth.Data(time)));
}

void MobileObject::forget(){
	forgot = true;
	//pos_conf *= CP_min_valid_conf;
	//speed_conf = 0.0f;
}

void MobileObject::set_speed_param(float max,float decay){
	max_speed = max;
	speed_decay = decay;
}

bool MobileObject::speed_valid(){
	return bool(speed_conf > min_valid_conf);
}

Vector MobileObject::PredictPos(int cycles){
	Vector predictpos = pos;
	Vector vel = global_vel;
	for(int i = 0; i < cycles;i++){
		predictpos += vel;
		vel *= speed_decay;
	}
	return predictpos;
}

void MobileObject::PredictPos(Vector p[], int cycles){
	if(cycles <= 0) return;
	Vector vel = global_vel;
	p[0] = pos + vel;
	for(int i = 1; i < cycles;i++){
		vel *= speed_decay;
		p[i] = p[i-1] + vel;
	}
}

float MobileObject::GetOffsensitivity(){
	return GetOffsensitivity(situation.CurrentTime);
}

float MobileObject::GetOffsensitivity(Time t){
	if(!offsensitivity.IsDataKnown(t)){
		offsensitivity.Setdata(fieldinfo.GetOffensiveSensitivity(Pos(t)), t);
	}
	return offsensitivity.Data(t);
}

float MobileObject::GetDefsensitivity(){
	return GetDefsensitivity(situation.CurrentTime);
}

float MobileObject::GetDefsensitivity(Time t){
	if(!defsensitivity.IsDataKnown(t)){
		defsensitivity.Setdata(fieldinfo.GetDefensiveSensitivity(Pos(t)), t);
	}
	return defsensitivity.Data(t);
}

bool MobileObject::IsLastSeen(){
	return bool(Update_seen_time() == Self.Update_seen_time()) ;
}
/**************************   Player    ******************************/
Player::Player(){
	MobileObject::MobileObject();
	InsideNO = -1;
	headfacing = 0;
	bodyfacing = 0;
	IT_inf.IT_cycles = InfCycles;
	Is_goalie = false;
	No = -1;
	side = Side_Unknown;
	Iscontrolball = false;
	original_postime = 0;
	original_veltime = 0;
}

void Player::Reset(Time time){
	vel_time = time - 1;
	gvel.Setdata(0, time -1);
	speed_conf = 0;
	update_seen_time = Max(time - 1, 0);

	pos_time = time -1;
	gpos.Setdata(0, pos_time);
	
	coordinate_time = -1;
}

bool Player::IsSamePlayer(Player& newp){
	float dist;
	int timegap;
	if(gpos.IsDataKnown(original_postime)){ 
		dist = Pos(original_postime).dist(newp.Pos(newp.original_postime));
		timegap = abs(newp.original_postime - original_postime);
	}else{
		dist = pos.dist(newp.pos);
		timegap = 8;
	}
	if(dist > CP_too_different_distance) return false;
	if(dist < ((newp.distance + distance) * SP_dist_qstep + timegap * SP_player_speed_max + 1.0f))
		return true;
	else
		return false;
}

void Player::estimaters(Time time){
	if(!rs_valid) return;

	while(rspos_time < time){
		rspos.Setdata(rspos.Data(rspos_time), rspos_time+1);
		rsradius.Setdata(Min(rsradius.Data(rspos_time) + max_speed, CP_latent_max_area), rspos_time+1);
		rspos_time ++;
		//DoLog(LOG_VDEC, "estimate %d (%.1f, %.1f) r%.1f at %d", InsideNO,rspos.Data(rspos_time).x, rspos.Data(rspos_time).y, rsradius.Data(rspos_time), rspos_time);
	}
}

void Player::update(Time time){
	if(InsideNO == MyNumber) return;

	MobileObject::update(time);
	//check if it control the ball
	if (ball.IsSeen()){
	//see ball
		Iscontrolball = seen && (Pos(ball.seen_time()).dist2(ball.Pos(ball.seen_time())) < Sqr(SP_kickable_area));
	}
}

void Player::update_end(){
	if(sensory.NewSight){// && IsUNumKnown()){//don't update unknown player
		//update active area
		if(seen){			
			DoLog(LOG_VISUAL, "%d is seen", InsideNO);
		}else if(rs_valid){
			estimaters(sensory.LastSightTime);
			//split circle(rspos, rsradius) with my view angle(headfcing, viewangle)
			Vector relpos = rspos.Data(sensory.LastSightTime) - Self.Pos(sensory.LastSightTime);
			if(relpos.mod() > SP_team_far_length) return;
			
			relpos = relpos.Rotate(-Self.Headfacing(sensory.LastSightTime));
			float gapangle = relpos.Angle();
			float rayangle = gapangle > 0 ? sensory.MyViewAngle() : -sensory.MyViewAngle();
			Ray viewray(Vector(0,0), rayangle);
			Vector pos1,pos2; 
			if(fabs(gapangle) < sensory.MyViewAngle()+1.0f){
				DoLog(LOG_VISUAL,"Player%d (%.1f, %.1f) radius %.1f is out of range ", InsideNO, rspos.Data(sensory.LastSightTime).x, rspos.Data(sensory.LastSightTime).y, rsradius.Data(sensory.LastSightTime));
			}
			
			if(!viewray.CircleIntersect(rsradius.Data(sensory.LastSightTime), relpos,pos1, pos2)){//no split
				if(fabs(gapangle) < sensory.MyViewAngle()+1.0f){
					rs_valid = false;
					DoLog(LOG_VISUAL,"Player%d rs fail agl%.0f, dist%.1f conf %.1f", InsideNO, gapangle, relpos.mod(), pos_conf);
				}
			}else{
			}
		}
	}
	estimaters(situation.CurrentTime);
	MobileObject::update_end();

	rel_pos_2_ball = pos - ball.pos;
	balldist = rel_pos_2_ball.mod();
	ballangle = (ball.pos - pos).Angle();
	if(rs_valid){
		DoLog(LOG_UPDATE, "estimate %d (%.1f, %.1f) r%.1f", // sensitivity(%.2f, %.2f)",
			InsideNO, rspos.Data(rspos_time).x, rspos.Data(rspos_time).y, rsradius.Data(rspos_time));//, GetOffsensitivity(), GetDefsensitivity());
		//DoLog("rel_pos_2_ball %d (%.2f %.2f) r %.1f", InsideNO, rel_pos_2_ball.mod(), rel_pos_2_ball.Angle(), rsradius.Data(rspos_time));
	}
	//if(Is_goalie) pos_conf = CP_min_valid_conf;
}

bool Player::Heardupdate(){
	if(!heard || seen) return false;

	if (heard_org_postime <= original_postime) return false;
	if(pos_valid() && rs_valid && rspos.IsDataKnown(situation.CurrentTime -1)){
		if(rspos.Data(situation.CurrentTime-1).dist(heardposinfo.data) > (ActiveRadius(situation.CurrentTime -1) + SP_player_speed_max) + SP_dist_qstep * distance * 1.5f){
			DoLog(LOG_HEAR, "Wronginfo %d(%.2f, %.2f) at %d to (%.2f %.2f) at %d, dist%.2f, radius %.2f", InsideNO, heardposinfo.data.x, heardposinfo.data.y,heard_org_postime,
			rspos.Data(situation.CurrentTime-1).x, rspos.Data(situation.CurrentTime-1).y, original_postime,rspos.Data(situation.CurrentTime-1).dist(heardposinfo.data), (ActiveRadius(situation.CurrentTime -1) + SP_player_speed_max) * 1.2f);
			return false;
		}
	}
	return true;
}

AngleDeg Player::Bodyfacing(Time time){
	return bodyfc.IsDataKnown(time) ? bodyfc.Data(time) : bodyfacing;
}

AngleDeg Player::Headfacing(Time time){
	return headfc.IsDataKnown(time) ? headfc.Data(time) : headfacing;
}

AngleDeg Player::Headangle(Time time){
	return headang.IsDataKnown(time) ? headang.Data(time) : head_angle;
}

bool Player::IsBodyfacingKnown(Time time){
	return bodyfc.IsDataKnown(time);
}

bool Player::IsHeadfacingKnown(Time time){
	return headfc.IsDataKnown(time);
}

bool Player::IsHeadAngleKnown(Time time){
	return headang.IsDataKnown(time);
}

void Player::set_fcinfo(float facedir, float neckdir, Time time){
	fcinfo = TimedData<Vector>(Vector(facedir, neckdir), time);
}

void Player::set_inside_no(UNum No){
	this->InsideNO = No;
}

void Player::set_side(bool Is_myside){
	if (Is_myside){
		side = Side_My;
	}
	else{
		side = Side_Opp;
	}
}

void Player::set_side(PlayerSide side){
	this->side = side;
}

void Player::Set_Iscontrolball_Fromheard(Time time){
	if (heardposinfo.time == time && heard_iscontrolball){
		Iscontrolball = true;
	}
	else{
		Iscontrolball = false;
	}
}

bool Player::IsStrained(){
	return bool(stamina < 0.5f * SP_stamina_max);
}

bool Player::IsExtremStrained(){
	return bool(stamina < SP_stamina_max * SP_recover_dec_thr);
}

/**************************   Self   *********************************/
MySelf::MySelf(){
}

void MySelf::update(Time time){
/*********	time is assumed to be CurrentTime here	********/
	if (sensory.SensedInfKnown(time)){
		update_from_senseinfo(sensory.GetSenseBodyInfo(time));
	}else{
		estimate_headangle(time);
		estimate_stamina(time);
		estimate_effort(time);
	}

	update_seen_facing(time);
	estimate_facing(time);

	update_vel(time);// vel should be updated before pos
	update_seen_pos(time);
	estimate_pos(time);

	pos = Pos(time);
	global_vel = Vel(time);
	speed = global_vel.mod();
	vel_angle = global_vel.Angle();
	bodyfacing = Bodyfacing(time);
	headfacing = Headfacing(time);
	head_angle = Headangle(time);
	stamina = Stamina(time);
	effort = Effort(time);
	pos_time = time;
	vel_time = time;
}

void MySelf::update_from_senseinfo(SenseBodyInfo & sb){
	viewwidth.Setdata(sb.view_width, sb.time);
	viewqual.Setdata(sb.view_qual, sb.time);
	stmn.Setdata(sb.stamina, sb.time);
	spd.Setdata(sb.speed < max_speed * speed_decay ? sb.speed : max_speed * speed_decay, sb.time);
	spdang.Setdata(sb.rel_spd_angle, sb.time);
	headang.Setdata(sb.head_angle, sb.time);
	efft.Setdata(sb.effort, sb.time);
}

void MySelf::estimate_headangle(Time time){
	if (action.Valid(CMD_turn_neck)){
		headang.Setdata(Headangle(situation.LastCycle()) + action.TurnNeckEffect(situation.LastCycle()), time);
	}
}

void MySelf::estimate_stamina(Time time){
	stmn.Setdata(Stamina(situation.LastCycle()) - action.UsedPower(situation.LastCycle()), time);
}

void MySelf::estimate_effort(Time time){
	efft.Setdata(Effort(situation.LastCycle()), time);
}

void MySelf::update_seen_facing(Time time){
	if (!sensory.NewSight) return;

/******	update facings from seen	************/
	int last_update_seen_time = Max(update_seen_time, time -2);
	for(int t = last_update_seen_time + 1; t <= sensory.LastSightTime; t ++){
		if (sensory.num_seenlines.ChkData(t) > 0){
			int line = sensory.SeenLines.Data(0, t);
			float rel_angle_2_line = Lines[line].seen_rel_angle(t);
			if (rel_angle_2_line < 0){
				rel_angle_2_line += 180.0f;
			}

			headfc.Setdata(NormalizeAngle(Lines[line].global_angle - rel_angle_2_line), t);
			bodyfc.Setdata(NormalizeAngle(Headfacing(t) - Headangle(t)), t);
		}
		else if (sensory.num_seenmarkers.ChkData(t) >= 2 ){
			//find two markers with known distance
			MarkerType marker1, marker2;
			marker1 = sensory.SeenMarkers.Data(0, t);
			marker2 = sensory.SeenMarkers.Data(1, t);

			Markers[marker1].rel_pos = Polar2Vector(Markers[marker1].seen_distance(t), Markers[marker1].seen_rel_angle(t));
			Markers[marker2].rel_pos = Polar2Vector(Markers[marker2].seen_distance(t), Markers[marker2].seen_rel_angle(t));
			Vector gap1 = Markers[marker1].rel_pos - Markers[marker2].rel_pos;
			Vector gap2 = Markers[marker1].pos - Markers[marker2].pos;
			headfc.Setdata(NormalizeAngle(gap2.Angle() - gap1.Angle()),	t);
			bodyfc.Setdata(NormalizeAngle(Headfacing(t) - Headangle(t)), t);
		}
	}
}

void MySelf::estimate_facing(Time time){
	if(!IsBodyfacingKnown(time)){
		if (!IsBodyfacingKnown(situation.LastCycle())){
			/********	must miss some cycles	************/
			bodyfc.Setdata(NormalizeAngle(Bodyfacing(pos_time) + action.SummarizeTurneffect(pos_time, situation.LastCycle()))
				, situation.LastCycle());
			if(IsHeadAngleKnown(situation.LastCycle())){
				headang.Setdata(NormalizeAngle(Headangle(pos_time) + action.SummarizeTurnneckeffect(pos_time, situation.LastCycle()))
					, situation.LastCycle());
			}
			headfc.Setdata(NormalizeAngle(Bodyfacing(situation.LastCycle()) + Headangle(situation.LastCycle())), 
					situation.LastCycle());
		}

		bodyfc.Setdata(NormalizeAngle(Bodyfacing(situation.LastCycle()) + action.TurnEffect(situation.LastCycle())), time);
		headfc.Setdata(NormalizeAngle(Bodyfacing(time) + Headangle(time)), time); 
	}
}

void MySelf::update_vel(Time time){
	Vector evel = action.SummarizeDasheffectVel(Vel(vel_time), vel_time, time, speed_decay, max_speed);

	if (sensory.SensedInfKnown(time)){
		gvel.Setdata(Polar2Vector(Speed(time), SpeedAngle(time) + Headfacing(time)), time);

		float Velgap =(float)fabs(gvel.Data(time).mod() - evel.mod());

		Vector svgap = gvel.Data(time) - evel;

		if(Velgap >CP_selfvel_mingap ){
			DoLog(LOG_COLLIDE,"self coll Velgap%.2f,Vector Gap(%.2f,%.2f)",Velgap,svgap.x,svgap.y);
			selfcoll.Setdata(true,time);
		}else
			selfcoll.Setdata(false,time);
	}else{
		/*estimate vel*/
		gvel.Setdata(evel, time);
	}
	vel_time = time;
}

void MySelf::update_seen_pos(Time time){
/********	Update Pos from seen	**************/
	if(!sensory.NewSight) return;

	int last_update_seen_time = Max(update_seen_time, time -2);
	MarkerType closestmarker;
	for(int t = last_update_seen_time + 1; t <= sensory.LastSightTime; t ++){
		if (sensory.num_seenmarkers.ChkData(t) > 0){
			/***********	calculate pos gap	***************/
 			closestmarker = sensory.GetClosestMarker(t);
			if (closestmarker != No_Marker){
				Markers[closestmarker].rel_pos = Polar2Vector(Markers[closestmarker].seen_distance(t),
					Headfacing(t) + Markers[closestmarker].seen_rel_angle(t));
				gpos.Setdata(Markers[closestmarker].pos - Markers[closestmarker].rel_pos, t);
				DoLog(LOG_UPDATE, "Player Pos Updated (%.2f %.2f) at %d", Pos(t).x, Pos(t).y, t);

				if(pos_time < t -1){
					estimate_pos(t-1);
				}
				Vector posbk = Pos(t-1) + Vel(t) / speed_decay;
				pos_gap.Setdata(Pos(t) - posbk, t);
				if(pos_gap.Data(t).mod() > 4.0f){
					DoLog("I'm moved");
					pos_gap.Setdata(0, t);
				} 
				DoLog(LOG_UPDATE, "PrevCoord (%.2f %.2f)", posbk.x, posbk.y);				
				pos_time = t;
				update_seen_time = t;
			}
		}
	}
}


void MySelf::estimate_pos(Time time){
	if (pos_time < time){
		gpos.Setdata(Pos(pos_time) + SummarizeVelPos(pos_time, time), time);
		pos_time = time;
	}
}

Vector MySelf::SummarizeVelPos(Time prev, Time last){
	Vector moved_pos = 0;
	for(Time time = prev; time < last; time ++){
		moved_pos += Vel(time + 1) / speed_decay;
	}
	return moved_pos;
}

Vector MySelf::SummarizePosGap(Time prev, Time last){
	Vector moved_pos = 0;
	for(Time time = prev + 1; time <= last; time ++){
		moved_pos += pos_gap.ChkData(time);
	}
	return moved_pos;
}

float MySelf::Speed(Time time){
	return spd.IsDataKnown(time) ? spd.Data(time) : speed;
}

AngleDeg MySelf::SpeedAngle(Time time){
	return spdang.IsDataKnown(time) ? spdang.Data(time) : NormalizeAngle(vel_angle - headfacing);
}

float MySelf::Stamina(Time time){
	return stmn.IsDataKnown(time) ? stmn.Data(time) : stamina;
}

float MySelf::Effort(Time time){
	return efft.IsDataKnown(time) ? efft.Data(time) : effort;
}

bool MySelf::SelfColl(Time prevtime,Time lastime ){
	for(Time t=prevtime; t<=lastime;t++){
		if(selfcoll.ChkData(t)) return true;
	}
	return false;
}
bool MySelf::IsSpeedKnown(Time time){
	return spd.IsDataKnown(time);
}
bool MySelf::IsSpeedAngleKnown(Time time){
	return spdang.IsDataKnown(time);
}

bool MySelf::IsStaminaKnown(Time time){
	return stmn.IsDataKnown(time);
}

bool MySelf::IsEffortKnown(Time time){
	return efft.IsDataKnown(time);
}

Vector MySelf::PredictPos(float dash_pow, int cycles){
	Vector predictpos = pos;
	Vector vel = global_vel;
	for(int i = 0; i < cycles; i ++){
		vel += Polar2Vector(dash_pow, Self.bodyfacing) * action.Dashrate();
		if (vel.mod() > SP_player_speed_max){
			vel = vel / vel.mod() * SP_player_speed_max;
		}
		predictpos += vel;
		vel *= speed_decay;
	}
	return predictpos;
}

Vector MySelf::PredictPos(int cycles){
	return MobileObject::PredictPos(cycles);
}

Vector MySelf::PredictPosWithTurnDash(AngleDeg angle, float dash_pow){
	Vector predictpos = Self.PredictPos(2);
	predictpos += Polar2Vector(dash_pow * action.Dashrate(), angle);
	return predictpos;
}

/*********************     Ball     **************************/
Ball::Ball(){
	posgap_time = -1;
	velgap_time = -1;
}

void Ball::update_seengap(Time t){
	if(!seenposinfo.IsDataKnown(t)) return;

	Vector t_posgap;
	float headfacingap = 1.5f;
	float dist = seen_distance(t);
	float rposagl = NormalizeAngle(Self.Headfacing(t) +  seen_rel_angle(t));
	float fcosagl = (float)fabs(Cos(rposagl));
	float fsinagl = (float)fabs(Sin(rposagl));
	//ԾǶ
	float dis_gap = 0.05f + 0.513f * dist * SP_dist_qstep/0.9f;
	float agl_gap = Deg2Rad(0.5f + headfacingap);
	//ͶӰȫϵ
	t_posgap.x = dis_gap * fcosagl + (dist + dis_gap) * agl_gap * fsinagl;
	t_posgap.y = dis_gap * fsinagl + (dist + dis_gap) * agl_gap * fcosagl;
	posgap.Setdata(t_posgap,t);
	
	Vector rpos = Pos(t) - Self.Pos(t);
	DoLog(LOG_UPDATE,"ball rpos(%.4f,%.4f) gap(%.4f %.4f) hafcing %.1f at %d", rpos.x,rpos.y, t_posgap.x, t_posgap.y, Self.Headfacing(t),t);
	posgap_time = t;
	if(seenchinfo.IsDataKnown(t)){
		Vector t_velgap;
		float SP_disch_qstep = 0.02f;
		float SP_dirch_qstep = 0.1f;//SP_dir_qstep;
	
		//ٶ
		t_velgap.x = dis_gap / dist * (float)fabs(seen_distCh(t)) + (dist + dis_gap) * SP_disch_qstep * 0.5f;
		t_velgap.y  = dis_gap * Deg2Rad(seen_dirCh(t)) + (dist + dis_gap) * Deg2Rad(SP_dirch_qstep * 0.5f);
		
		t_velgap = t_velgap.Rotate(rposagl);
		t_velgap.x = (float)fabs(t_velgap.x);
		t_velgap.y = (float)fabs(t_velgap.y);
		//תrposagl
		float tmp = t_velgap.y;
		t_velgap.y += t_velgap.x * agl_gap;
		t_velgap.x += tmp * agl_gap;
		
		//Selfٶ
		Vector svelgap;
		float rvelagl = NormalizeAngle(Self.Headfacing(t) + Self.SpeedAngle(t));
		fcosagl = (float)fabs(Cos(rvelagl));
		fsinagl = (float)fabs(Sin(rvelagl));
		svelgap.x = (Self.Speed(t) + 0.005f) * agl_gap * fsinagl;
		svelgap.y = (Self.Speed(t) + 0.005f) * agl_gap * fcosagl;
		t_velgap += svelgap;
		velgap.Setdata(t_velgap,t);
		velgap_time = t;

		DoLog(LOG_UPDATE,"ball vel(%.4f,%.4f) gap(%.4f %.4f) selfvel gap(%.2f,%.2f)",Vel(t).x,Vel(t).y,t_velgap.x,t_velgap.y,svelgap.x,svelgap.y);
	}
}

void Ball::estimategap(Time t){
	Time i;
	Vector gap;
	for(i = velgap_time; i < t; i++){
		gap = (velgap.ChkData(i) + Vector(1,1) * SP_ball_rand * (velgap.ChkData(i).mod() + gvel.ChkData(i+1).mod()/speed_decay)) * speed_decay;
		gap.x = gap.x > max_speed ? max_speed : gap.x;
		gap.y = gap.y > max_speed ? max_speed : gap.y;
		velgap.Setdata(gap, i+1);
		DoLog(LOG_UPDATE,"Estimate velgap (%.3f,%.3f) at %d", velgap.ChkData(i+1).x, velgap.ChkData(i+1).y, i+1);
	}
	velgap_time = t;
	DoLog(LOG_UPDATE, "Estimate posgap (%.3f,%.3f) at %d(%.2f %.2f %.2f)",posgap.ChkData(posgap_time).x, posgap.ChkData(posgap_time).y, posgap_time, speed_decay, SP_player_rand, SP_player_decay);
	for(i = posgap_time; i < t; i++){
		posgap.Setdata(posgap.ChkData(i) + velgap.ChkData(i+1)/speed_decay + Vector(1,1) * SP_player_rand * Self.Vel(i+1).mod() / SP_player_decay, i+1);
		DoLog(LOG_UPDATE, "Estimate posgap (%.3f,%.3f) selfvel (%.2f,%.2f) at %d",posgap.ChkData(i+1).x, posgap.ChkData(i+1).y, Self.Vel(i+1).x, Self.Vel(i+1).y, i+1);
	}
	posgap_time = t;
}

void Ball::update_seen(Time time){
/* What's special about ball's updation?
	Collisions, and a more accurate velocity is needed*/
	if(!seen) return;
	Time last_update_time = Max(update_seen_time, time -2);	
	Vector predictpos, predictvel;
	Vector allowposgap, allowvelgap;
	Vector predictpos_gap, predictvel_gap;
	
	for(int t = last_update_time + 1; t <= time; t++){
		if(!seenposinfo.IsDataKnown(t)) continue;	
		estimate(t);
		predictpos = Pos(t) - Self.Pos(t);
		DoLog(LOG_UPDATE, "estimate relpos %.2f, %.2f", predictpos.x, predictpos.y);
		predictvel = Vel(t);
		allowposgap = posgap.ChkData(t);
		allowvelgap = velgap.ChkData(t);
		Time last_seen_time = original_postime;
		Time last_seench_time = original_veltime;
		MobileObject::update_seen(t);
		update_seengap(t);
		
		if(seen_distance(t) > 30.0f) continue;
		
		if (seen_distance(t) > (SP_kickable_area+0.04f)){
			action.ResetKickEffect(t); //then the kick command sent at seen_time must be a fault
		}
		int postime_gap = t - last_seen_time;
		if(postime_gap <= 0){
			DoLog(LOG_UPDATE,"why postime_gap = 0? %d", t);
			continue;
		}
		
		allowposgap += posgap.ChkData(t);
		predictpos_gap = Pos(t) - Self.Pos(t) - predictpos;// - Self.SummarizePosGap(last_seen_time, t);
		DoLog(LOG_UPDATE,"T gap%d, V gap%d, Predictpos Gap(%.2f,%.2f)", postime_gap, t - last_seench_time, predictpos_gap.x, predictpos_gap.y);
		
		bool selfcoll = Self.SelfColl(last_seen_time+1, t);
		float erroridx;
		bool PredictError = IsposError(allowposgap, predictpos_gap, erroridx);
		
		if(seenchinfo.IsDataKnown(t)){
			allowvelgap += velgap.ChkData(t);
			predictvel_gap = Vel(t) - predictvel;
			DoLog(LOG_UPDATE,"predictvel Gap(%.2f,%.2f)",predictvel_gap.x,predictvel_gap.y);
			float verroridx;
			bool PrevelError = IsposError(allowvelgap, predictvel_gap, verroridx);
			if(PredictError || PrevelError){
				erroridx += verroridx;
				DoLog(LOG_UPDATE, "Error Index %.3f",erroridx);
			}
			error_idx.Setdata(erroridx, t);
			continue;
		}
		if(predictpos.mod() < SP_ball_size + SP_player_size){
			DoLog(LOG_UPDATE, "predict coll, unhandled yet");
		}
		error_idx.Setdata(erroridx, t);
		float sqrconf = max_conf * Sqr(conf_decay);
		float CP_collision_maxdis = 0.75f;
		
		if(!PredictError ){
			if(postime_gap ==1 && !OtherKick(last_seen_time,t) && velgap.IsDataKnown(t) && (posgap.Data(t).mod() + posgap.Data(t-1).mod()) > velgap.Data(t).mod()){
				DoLog(LOG_UPDATE, "estimate maybe more accurate, reset velconf");
				speed_conf = max_conf;
				//wait for estimate
			}else if(selfcoll && postime_gap <= 2){
				DoLog(LOG_UPDATE, "self coll, pos still correct fortunately");
				//wait for estimate
			}else if(!gpos.IsDataKnown(last_seen_time)){
				DoLog(LOG_UPDATE, "last pos is lost, have to estimate");
				//wait for estimate
			}else if(postime_gap<4 && speed_valid()){
				ApplyPosbased_vel(last_seen_time,t,max_conf);
			}else if(postime_gap<6){
				ApplyPosbased_vel(last_seen_time,t,sqrconf);
			}else
				ApplyPosbased_vel(last_seen_time,t,sqrconf * sqrconf);
			
			if(!selfcoll && postime_gap == 1 && seen_distance(t) < 2.0f){
				original_veltime = t - 1;
				DoLog(LOG_UPDATE,"Reliable predict vel at %d", t);
			}
			continue;
		}else{
			DoLog(LOG_UPDATE,"why ball is here, Error Index %.3f at %d", erroridx, t);
		}

		//unmoved posgap limit;
		float dist_moved;
		if(gpos.IsDataKnown(last_seen_time)){
			dist_moved = (gpos.Data(last_seen_time) + Self.SummarizePosGap(last_seen_time, t)).dist(gpos.Data(t)) - allowposgap.mod();
		}else{
			dist_moved = predictpos.dist(gpos.Data(t)) - Vel(t).mod() * postime_gap  - allowposgap.mod();
		}
		if(dist_moved > max_speed * postime_gap){
			DoLog(LOG_UPDATE,"ball is moved");
			gvel.Setdata(0, t);
			vel_time = t;
			if(postime_gap ==1){
				speed_conf = max_conf;
			}else if(postime_gap == 2){
				speed_conf = sqrconf;
			}
		}else if(postime_gap ==1){
			if(OtherKick(last_seen_time)){
				ApplyPosbased_vel(last_seen_time,t,sqrconf);
				DoLog(LOG_UPDATE,"Predict other kick at %.0f",(float)last_seen_time);
			}else if(selfcoll){
				//wait for estimate
				DoLog(LOG_UPDATE,"Error because Self coll");
			}else if(last_seench_time < last_seen_time-1 ){
				ApplyPosbased_vel(last_seen_time, t, sqrconf );
				DoLog(LOG_UPDATE,"Error maybe too old vel");
			}else if(seen_distance(t) < CP_collision_maxdis){
				Collisionvel(t,sqrconf);
				DoLog(LOG_UPDATE,"It must be coll");
			}else{
				ApplyPosbased_vel(last_seen_time, t, sqrconf);
				DoLog(LOG_UPDATE,"Unknown reason");
			}
		}else if(postime_gap == 2){
			if(OtherKick(last_seen_time)){
				ApplyPosbased_vel(last_seen_time, t, sqrconf * conf_decay);
				DoLog(LOG_UPDATE,"Predict other kick2 at%d", last_seen_time);
			}else if(OtherKick(last_seen_time + 1)){
				ApplyPosbased_vel(last_seen_time + 1, t, sqrconf * conf_decay);
				DoLog(LOG_UPDATE,"Predict other kick3 at%d", last_seen_time+1);
			}else if(selfcoll){
				//wait for estimate
				DoLog(LOG_UPDATE,"Error because Self coll");
			}else if(last_seench_time<last_seen_time-1){
				ApplyPosbased_vel(last_seen_time, t, sqrconf * conf_decay );
				DoLog(LOG_UPDATE,"Error maybe too old vel");
			}else{
				if(seen_distance(t) > SP_feel_distance){
					DoLog(LOG_UPDATE,"too far ,see wrong");
				}else{
					float dis1 = (Pos(t) - Self.Pos(t)).mod() ;
					float dis2 = (Pos(t-1) - Self.Pos(t-1)).mod();
					DoLog(LOG_UPDATE,"t%d dis%.2f, t%d dis%.2f", t-1, dis2, t,dis1);
					if(dis1 < dis2 && dis1 < CP_collision_maxdis){
						Collisionvel(t, sqrconf * conf_decay);
						DoLog(LOG_UPDATE,"find coll at t%d", t);
					}else if(dis2 < dis1 && dis2 < CP_collision_maxdis){
						Collisionvel(t - 1, sqrconf * conf_decay);
						DoLog(LOG_UPDATE,"find coll at t%d", t-1);
					}else{
						if(action.KickEffect(last_seen_time).mod() > 0 && ball.Pos(last_seen_time).dist2(Self.Pos(last_seen_time)) > 1.0f){
							DoLog(LOG_UPDATE,"maybe kick%d invalid", last_seen_time+1);
							action.ResetKickEffect(last_seen_time+1);
						}else{
							DoLog(LOG_UPDATE,"maybe unfound kick %d", t);
						}

						ApplyPosbased_vel(last_seen_time, t, sqrconf * conf_decay );
						//invalid kick not be clear out. or other kick haven't been found
					}
				}
			}
		}else if(!gpos.IsDataKnown(last_seen_time)){
			//wait for estimate
		}else if(postime_gap<6){
			ApplyPosbased_vel(last_seen_time, t, sqrconf * conf_decay);
		}else{
			ApplyPosbased_vel(last_seen_time, t, CP_min_valid_conf * conf_decay);
		}
	}
}

void Ball::ApplyPosbased_vel(Time prevtime,Time latestime,float conf){
	bool valid;
	Vector evel_posbased;
	evel_posbased = posbased_vel_estimate(prevtime, latestime, valid);
	if(valid){
		gvel.Setdata(evel_posbased, latestime);
		vel_time = latestime;
		speed_conf = conf;
		DoLog(LOG_UPDATE,"t%d, Posbased vel applied", latestime);
	}else{
		DoLog(LOG_UPDATE,"Why posbased_vel Invalid (%.2f %.2f)", evel_posbased.x, evel_posbased.y);
	}
	velgap.Setdata(posgap.ChkData(prevtime) + posgap.ChkData(latestime), latestime);
	velgap_time = latestime;
	DoLog(LOG_UPDATE, "reset velgap (%.2f, %.2f)",velgap.Data(latestime).x, velgap.Data(latestime).y);
}

bool Ball::Heardupdate(){
	if(!heard) return false;	

	if (pos_conf > 0.8f || distance < 2.0f){
		//it is critical for me, if the confidence of ball is greater than 0.8f and it ball is within 2 meter of my reach
		if (heardvelinfo.time != heardposinfo.time)	return false;

		if (ball.global_vel.dist2(heardvelinfo.data) < Sqr(0.6f)) return false;
	}

	if(ball.heardpass.time == heardposinfo.time	&& heard_org_postime >= original_postime)
		return true;

	if(heard_org_postime <= original_postime)
		return false;
		
	if(IsLastSeen() && !OtherKick(update_seen_time)) return false;

	return true;
}

void Ball::update_end(){	
	MobileObject::update_end(); 
	if(distance > CP_kickable_area || !pos_valid() || !Rs_valid()){
		kick_able = false;
	}else{
		kick_able = (Update_seen_time() >= Self.Update_seen_time() || 
			ball.Pos(Self.Update_seen_time()).dist(Self.Pos(Self.Update_seen_time())) + ball.posgap.ChkData(Self.Update_seen_time()).mod() > SP_feel_distance);
	}
	if(distance < CP_kickable_area && !kick_able){
		if(!pos_valid()) DoLog(LOG_UPDATE, "kick disable for pos invalid");
		else if(!Rs_valid()) DoLog(LOG_UPDATE, "kick disable for Rs invalid");
		else{
			DoLog(LOG_UPDATE, "kick disable for ball( %d) is unseen at %d", Update_seen_time(), Self.Update_seen_time());
		}
	}
	theirgoalangle = (fieldinfo.theirgoal - pos).Angle();
	mygoalangle = (fieldinfo.mygoal - pos).Angle();
}
 
void Ball::Collisionvel(Time time,float conf){
	Vector vel = action.SummarizeKickeffectVel(Vel(vel_time),vel_time,time,speed_decay,max_speed);
	gvel.Setdata(vel * -0.1f, time);
	vel_time = time;
	speed_conf = conf;
	DoLog(LOG_UPDATE,"coll handle t%.0f,v(%.2f,%.2f)",(float)time,Vel(time).x,Vel(time).y);
}

bool Ball::OtherKick(Time time){
	if(IscrposKnown(time)){
		if(closestrpos(time).mod() < SP_kickable_area+0.05f )
			return true;
	}
	return false;
}

 bool Ball::OtherKick(Time prevtime,Time lastime){
	for(Time t=prevtime;t<=lastime;t++){
		if(OtherKick(t)) return true;
	}
	return false;
}

bool Ball::IsposError(Vector allowposgap, Vector predictgap,float& gapmod){
	if(fabs(predictgap.x) > allowposgap.x || fabs(predictgap.y) > allowposgap.y){
		gapmod = fabs(predictgap.x) > allowposgap.x ? Sqr((float)fabs(predictgap.x)-allowposgap.x) : 0;
		gapmod += fabs(predictgap.y) > allowposgap.y ? Sqr((float)fabs(predictgap.y)-allowposgap.y) : 0;
		gapmod = (float)sqrt(gapmod / allowposgap.mod());
		return true;
	}else{
		gapmod = 0;
		return false;
	}
}

Vector Ball::posbased_vel_estimate(Time prev_time, Time last_time, bool& valid){
	/* estimate by postion difference */
	valid = false;
	Vector gap = Pos(last_time) - Pos(prev_time);
	Vector evel;
	gap -= Self.SummarizePosGap(prev_time, last_time);

	DoLog(LOG_UPDATE,"prev pos (%.2f %.2f) last pos (%.2f %.2f) gap (%.2f %.2f)", Pos(prev_time).x, Pos(prev_time).y,
		Pos(last_time).x, Pos(last_time).y, gap.x, gap.y);
	//if (gap.mod() < (float)fabs(last_time - prev_time) * max_speed){
	Vector kickeffectpos = action.SummarizeKickeffectPos(prev_time, last_time, speed_decay);
	evel = (gap - kickeffectpos) * (1 - speed_decay) / (1 - Exp(speed_decay, (float)(last_time - prev_time)));
	//evel is now that at time prev_time
	if(evel.mod() > max_speed){
		evel = evel / evel.mod() * max_speed;
	}
	evel = action.SummarizeKickeffectVel(evel, prev_time, last_time, speed_decay, max_speed);
	valid = true;
	return evel;
}

void Ball::estimate(Time time){
	MobileObject::estimate(time);
	estimategap(time);
}

void Ball::estimate_pos(Time time){
	MobileObject::estimate_pos(time);
}

void Ball::CollisionHandle(Time time){
	DoLog(LOG_UPDATE,"Coolisionhandle");
	estimate_pos(time);
	if (speed_valid()){
		Vector rpos = Pos(time) - Self.Pos(time);
		float r = SP_ball_size + SP_player_size;
		float d = rpos.mod();
		if (d>=r) return; 
		AngleDeg th = float(fabs(NormalizeAngle(rpos.Angle() - Self.Bodyfacing(time -1) - 180.0f)));
		float l1 = d * Cos(th);
		float h = d * Sin(th);
		float cosp = h / r;
		float sinp = (float)sqrt(1.0 - Sqr(cosp));
		float l2 = r * sinp;
		gpos.Data(time) += Vel(time) *float( speed_decay * (-(l1+l2) / Max(Vel(time).mod() * speed_decay, 0.000001f)));
		gvel.Data(time) *= -0.1f;
		vel_time = time;
		speed_conf = min_valid_conf / Sqr(conf_decay);
	}
}

bool Ball::catchable(){
	if(Self.Is_goalie)
		return bool(pos_conf > 0.8f && distance < SP_catch_area_l && fieldinfo.my_penaltyarea.IsWithin(Self.pos) && action.cancatch(situation.CurrentTime));
	else
		return bool(pos_conf > 0.8f && distance < SP_catch_area_l && fieldinfo.my_penaltyarea.IsWithin(pos) && action.cancatch(situation.CurrentTime));
}

bool Ball::shootable(){
	return bool(fieldinfo.shootablearea.IsWithin(pos) && kickable());
}

bool Ball::kickable(const Vector& selfpos){
	return bool(pos_valid() && ball.pos.dist2(selfpos) < Sqr(SP_kickable_area));
}

bool Ball::catchable(const Vector& selfpos){
	return bool(pos_valid() && pos.dist2(selfpos) < Sqr(SP_catch_area_l) && fieldinfo.my_penaltyarea.IsWithin(selfpos) && action.cancatch(situation.CurrentTime));
}

bool Ball::shootable(const Vector& ballpos){
	return fieldinfo.shootablearea.IsWithin(ballpos);
}

void Ball::Setball_CPlayer(UNum idx,Vector pos,Time time){
	if(idx > 0 && idx <= 2*SP_team_size)
		ClosestPlayer.Setdata(idx, time);
	else
		ClosestPlayer.Setdata(idx, time-1);
	ClosestRPos.Setdata(pos-Pos(time),time);
}

bool Ball::IsOtherKick(){
	if(IsOtherKick(situation.CurrentTime) || IsOtherKick(situation.CurrentTime - 1)){
		return true;
	}else
		return false;
}

bool Ball::IsOtherKick(Time t){
	if(error_idx.IsDataKnown(t) && error_idx.Data(t) > 0.01f )
		return true;
	else
		return false;
}
/*	UnknownPlayer	*/
UnknownPlayerNode::UnknownPlayerNode(){
	next = NULL;
}

/*	UnknownPlayerList	*/
UnknownPlayerList::UnknownPlayerList(){
	Head.next = NULL;
	Tail = &Head;
	numplayers = 0;
}

UnknownPlayerList::~UnknownPlayerList(){
	UnknownPlayerNode* tmp;
	while(Head.next != NULL){
		tmp = Head.next->next;
		delete Head.next;
		Head.next = tmp;
	}
}

void UnknownPlayerList::EnList(Player& player){
	if(sensory.TwiceSight){//check if repeat player
		for(int i=oldnumplayers; i<numplayers; i++){
			if(player.seen_time() == List[i]->seen_time()) break;
			if(List[i]->IsSamePlayer(player)){
				DoLog(LOG_UNKPLAYER,"find two same unknown opp");
				return;
			}
		}
	}
	if(numplayers >= SP_team_size){
		DoLog(LOG_UNKPLAYER,"num_unk_players reach limts");
		return;
	}

	Tail->next = new UnknownPlayerNode;
	Tail->next->unknownplayer = player;
	if(!IsUnknownside()){
		Tail->next->unknownplayer.set_side(IsMyside());
	}
	Tail = Tail->next;
	Tail->next = NULL;
	ArraySet(&(Tail->unknownplayer));
}

bool UnknownPlayerList::Clearable(Player& player){
	if(situation.ClockStopped) return true;
	if(player.pos_conf * Exp(CP_conf_decay, (float)(situation.CurrentTime - player.pos_time)) < 0.8f) return true;

	if(!sensory.NewSight) return false;
	if (player.inview(sensory.LastSightTime) || (sensory.TwiceSight && player.inview(sensory.PrevSightTime))){
		if(IsUnknownside()) return true;
		if(player.distance < SP_team_far_length) return true;
	}
	int i;
	float max_deviation = CP_different_distance;
	if(IsMyside() || IsUnknownside()){
		for(i=0; i<SP_team_size; i++){
			if(!MyTeam[i].IsSeen()) continue;
			if(player.pos.disaeolus(MyTeam[i].pos) < max_deviation) return true;
		}
		for(i=0; i<num_unknown_myplayers; i++){
			if (!UnknownMyPlayers[i].IsDistanceSeen()) continue;
			if(player.pos.disaeolus(UnknownMyPlayers[i].pos) < max_deviation) return true;
		}
	}
	if(IsOppside() || IsUnknownside()){
		for(i=0; i<SP_team_size; i++){
			if(!TheirTeam[i].IsSeen()) continue;
			if(player.pos.disaeolus(TheirTeam[i].pos) < max_deviation) return true;
		}
		for(i=0; i<num_unknown_theirplayers; i++){
			if (!UnknownTheirPlayers[i].IsDistanceSeen()) continue;
			if(player.pos.disaeolus(UnknownTheirPlayers[i].pos) < max_deviation) return true;
		}
	}
	if(IsUnknownside()){
		for(i=0; i<num_teamlessplayers; i++){
			if (!TeamlessPlayers[i].IsDistanceSeen()) continue;
			if(player.pos.disaeolus(TeamlessPlayers[i].pos) < max_deviation) return true;
		}
	}
	return false;
}

void UnknownPlayerList::ArraySet(Player* player){
	List[numplayers ++] = player;
}

void UnknownPlayerList::update_end(){
	for(int i=0; i< numplayers; i++){
		List[i]->update_end(); 
	}
} 

void UnknownPlayerList::update(Time time){
	UnknownPlayerNode* p = &Head, *tmp;
	numplayers = 0;	
	while(p->next != NULL){		
		if (Clearable(p->next->unknownplayer)){
			tmp = p->next->next;
			delete p->next;
			p->next = tmp;
		}else{
			p = p->next;
			p->unknownplayer.update(time);
			ArraySet(&(p->unknownplayer));
		}
	}
	Tail = p;
	oldnumplayers = numplayers;
}

void UnknownPlayerList::LogList(char* msg){
	UnknownPlayerNode* p = &Head;	
	int logcount=0;
	if(msg != NULL)
		logcount += sprintf(logbuf,msg);
	else{
		if(IsMyside()){
			logcount += sprintf(logbuf,"Teammate %d", numplayers);
		}else if(IsOppside()){
			logcount += sprintf(logbuf,"Opp %d", numplayers);
		}else{
			logcount += sprintf(logbuf,"Unknown %d", numplayers);
		}
	}
	while(p->next != NULL){
		p = p->next;
		logcount += sprintf(&logbuf[logcount]," pos(%.2f,%.2f)",p->unknownplayer.pos.x, p->unknownplayer.pos.y);
	}
	DoLog(LOG_UNKPLAYER,logbuf);
}

int UnknownPlayerList::NumUnknownPlayers(){
	return numplayers;
}

void UnknownPlayerList::Set_IsControlball_FromHeard(Time time){
	UnknownPlayerNode* p = &Head;
	while(p->next != NULL){
		p->next->unknownplayer.Set_Iscontrolball_Fromheard(time);
		p = p->next;
	}
}