/*
    Copyright (C) 2001  Tsinghuaeolus

    Authors : ChenJiang, YaoJinyi, CaiYunpeng, Lishi

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

	If you make any changes or have any comments we would appreciate a 
	message to yjy01@mails.tsinghua.edu.cn.
*/

#include "stdafx.h"

#include "global.h"
#include "strategy.h"


/*********************  Situation  *************************/
/************** Keep record of situation  *******************************/
Situation::Situation():MyScore(0),TheirScore(0),CurrentTime(-1),playmode(PM_No_Mode),
           LastInterruptTime(-1),InterruptsThisCycle(0),ClockStopped(false),ModeChangedTime(-1),
		   Sense_Interval(0){
	mode_changed = false;
	num_hearmodes = 0;
}

Time Situation::LastCycle(){
	return (!ClockStopped || situation.ClockStoppedTime == 0) ?
		situation.CurrentTime - 1 : situation.CurrentTime;
}

void Situation::HearMode(PlayMode playmode){
	if(num_hearmodes >= CP_max_hearmodes){
		num_hearmodes = 0;
	}
	heardmode[num_hearmodes ++] = playmode;
	mode_changed = true;
}

void Situation::UpdateMode(){
	if(mode_changed){
		DoLog(LOG_UPDATE, "mode changed");
		for(int i = 0; i < num_hearmodes; i++){
			SetPlayMode(heardmode[i]);
		}
		num_hearmodes = 0;
		mode_changed = false;
	}
}

void Situation::SetPlayMode(PlayMode playmode){
	this->playmode = playmode;
	OnChangeMode(playmode);
}

void Situation::ForgetAll(){
	for(int i=1; i<= SP_team_size; i++){
		MyPlayer(i).Setvalid(false);
		MyPlayer(i).pos_conf = 0;
		TheirPlayer(i).Setvalid(false);
		TheirPlayer(i).pos_conf = 0;
	}
}

void Situation::OnChangeMode(PlayMode playmode){
	ModeChangedTime = CurrentTime;
	if( ClockStopped){
		ClockStopped = false;
	}

	if(playmode != PM_Play_On)
		ForgetAll();

	switch (playmode){
	case PM_Before_Kick_Off :
	case PM_My_Offside_Kick :
		IsGoalieKick = false;
		PlaceKickState=SPST_Initialize;
		DoLog(LOG_SETPLAY,"initialized");
		ClockStopped = true;
		ClockStoppedTime = 0;
		break;	
	case PM_Play_On:
		break;
	case PM_My_Goalie_Free_Kick:
		IsGoalieKick = true;
		fm.goalieplacekick.movecount=SP_goalie_max_moves; 
		break;
	case PM_My_Goal_Kick :
		IsGoalieKick = true;
	case PM_My_Kick_Off :
	case PM_My_Corner_Kick :
	case PM_My_Kick_In :
	case PM_My_Free_Kick :
		DoLog(LOG_SETPLAY,"Mode Changed");
		PlaceKickState=SPST_Initialize;
		break;	

	case PM_Their_Offside_Kick :
		ClockStopped = true;
		ClockStoppedTime = 0;
		break;
	case PM_Drop_Ball :
		DoLog(LOG_SETPLAY,"Drop Ball");
		PlaceKickState=SPST_AfterExec;
		ball.forget();
		situation.IsGoalieKick = false;
		break;
	default:
		break;
	}
}

void Situation::LogGameState(){
	switch (PlaceKickState){
	case SPST_Initialize:
		DoLog(18,"Initialization");
		break;
	case SPST_Prego:
		DoLog(18,"PreExecute Going");
		break;
	case SPST_ExecWait:
		DoLog(18,"Waiting For Execute");
		break;
	case SPST_Executing:
		DoLog(18,"Executing");
		break;
	case SPST_AfterExec:
		DoLog(18,"Play Mode");
		break;
	default:
		DoLog(18,"UnExpected");
		break;
	}
	if (IsGoalieKick)
		DoLog(18,"Is Goalie Kick");
}

GameState Situation::JudgeGameState(){
	float CP_playon_ball_speed_min = 0.3f;
	if (situation.IsPlaceKick()&& (playmode==PM_Play_On||playmode==PM_My_Goal_Kick&&PlaceKickState > SPST_Prego) 
		  &&(situation.BallFree||ball.pos_valid() && motion.Num_TheirVisiblePlayers()>0&&motion.TheirPlayer_Close2Ball(0).pos.dist(ball.pos)<7))
	{
		DoLog(18,"Change To Playing");
		situation.ChangeSetPlay(SPST_AfterExec);
		situation.IsGoalieKick = false;
	}
	switch (playmode)
	{
		case PM_Before_Kick_Off :
			gamestate = GS_Before_Kick_Off;
			break;
		case PM_Their_Kick_Off :
		case PM_Their_Kick_In :
		case PM_Their_Corner_Kick :
		case PM_Their_Goal_Kick :
		case PM_Their_Free_Kick :
			DoLog(18,"Their kick");
			PlaceKickState = SPST_AfterExec;
			IsGoalieKick = false;
			gamestate = GS_Their_PlaceKick;
			break;
		case PM_My_Kick_Off :
		case PM_My_Kick_In :
		case PM_My_Corner_Kick :
		case PM_My_Goal_Kick :
		case PM_My_Free_Kick :
			if (IsPlaceKick())
				gamestate = GS_My_PlaceKick;
			else
				gamestate = GS_Playing;
			break;
		case PM_Play_On :
			if (!situation.IsGoalieKick) 
			{
				situation.ChangeSetPlay(SPST_AfterExec);
				gamestate = GS_Playing;
			}
			break;
		case PM_Half_Time :
		case PM_Time_Up :
		case PM_Extended_Time :
		case PM_My_Offside_Kick :
		case PM_Their_Offside_Kick :
		default :
			gamestate = GS_Other;	
			break;
	}
	situation.LogGameState();
	return gamestate;
}

void Situation::UpdateScore(char Side){
	if(Side == MySide)
		MyScore++;
	else
		TheirScore++;
}

void Situation::SetKickOffMode(KickOff_Mode kickoff_mode){
	this->kickoff_mode = kickoff_mode;
}

void Situation::UpdateTime(Time time){
	CurrentTime = time;
}

void Situation::JudgeSituation(){
	int num, i;
	motion.SetArounds();
//Judge about control of the ball
	num_my_ballcontrollers = 0;
	num_their_ballcontrollers = 0;
	if(ball.distance < 30.0f || ball.pos_conf > 0.9f){
		num = motion.Num_MyVisiblePlayers();
		for(i = 0; i < num; i ++){
			if (motion.MyPlayer_Close2Ball(i).balldist > 4.0f){
			//to far to be the controller of ball
				break;
			}
			if (motion.MyPlayer_Close2Ball(i).Iscontrolball){
				My_Ballcontrollers[num_my_ballcontrollers++] = motion.MyPlayers_Close2Ball[i];
			}
		}
		num = motion.Num_TheirVisiblePlayers();
		for(i =0; i < num; i ++){
			if (motion.TheirPlayer_Close2Ball(i).balldist > 4.0f){
			//to far to be the controller of ball
				break;
			}
			if (motion.TheirPlayer_Close2Ball(i).Iscontrolball){
				Their_Ballcontrollers[num_their_ballcontrollers++] = motion.TheirPlayers_Close2Ball[i];
			}
		}		
		DoLog(LOG_SEE, "judge 1");
	}

	num_ballcontrollers = num_my_ballcontrollers + num_their_ballcontrollers;

//for the sake of compatibility
	if (num_their_ballcontrollers > 0){
		ballcontroller = Their_Ballcontrollers[0];
		if (ballcontroller <= SP_team_size)
			ballcontroller += SP_team_size;
	}		
	else if (num_my_ballcontrollers > 0){
		ballcontroller = My_Ballcontrollers[0];
	}
	else if (ball.kickable()){
		ballcontroller = MyNumber;
	}
	else{
		ballcontroller = -1;
	}

	BallFree = bool(ballcontroller == -1);

//TeamControl
	conf_of_control = ball.pos_conf;
	if(!BallFree){
		if (num_their_ballcontrollers > 0){
			if (num_my_ballcontrollers > 0 || ball.kickable()){
				Team_Control = 0.5f;
			}
			else{
				Team_Control = 0.0f;
			}
		}
		else{
			Team_Control = 1.0f;
		}
	}
	else{		
		DoLog(LOG_SEE, "judge 2");
		motion.SetInterceptionInfo();
	}

	if (Self.attackness > 0.5f)
		IsAttack = bool(Team_Control >= 0.5f);
	else
		IsAttack = bool(Team_Control >= 0.6f);
	if(!BallFree){
		DoLog(LOG_SITUATION, "Ball is controlled by %d", ballcontroller);
	}
	else{
		DoLog(LOG_SITUATION, "most promising contoller is %d at (%.2f %.2f)", most_promising_controller, GetPlayer(situation.most_promising_controller).pos.x,
			GetPlayer(situation.most_promising_controller).pos.y);
	}
}

bool Situation::Ballwithincontrol(char side, UNum No){
	if (ball.pos_conf < 0.85 || GetPlayer(side, No).pos_conf < 0.8) return false;
	return bool(GetPlayer(side, No).balldist < CP_control_margin);
}

bool Situation::IsControllerOrFutureController(char side, UNum No){
	if (side == TheirSide && No > 0 && No <= SP_team_size){
		return IsControllerOrFutureController(No + SP_team_size);
	}
	return IsControllerOrFutureController(No);
}

bool Situation::IsControllerOrFutureController(UNum No){
	return (No == ballcontroller || No == most_promising_controller) && No > 0;
}

Vector Situation::NextControlBallPos(){
	return (!BallFree || ball.speed_conf < 0.8f || !IsValidNo(most_promising_controller)) ? ball.pos : next_interception_point;
}
/***************    Role   ******************/
Role::Role(float attackness, float leftness, float aggressiveness, float creative){
	SetRole(attackness, leftness, aggressiveness, creative);
}

void Role::SetRole(float attackness, float leftness, float aggressiveness, float creative){
	this->attackness = attackness;
	this->leftness = leftness;
	this->aggressiveness = aggressiveness;
	this->creativeness = creativeness;
}

bool Role::RoleMatched(Role& role){
	return bool(attackness == role.attackness && leftness == role.leftness);
}


/*****************  Set Play   ****************************/
void Set_Play::AddRole(float attackness, float leftness, const Vector& pos){
	roles[total_roles].attackness = attackness;
	roles[total_roles].leftness = leftness;
	roles[total_roles].strategical_pos = pos;
	total_roles ++;
}

void Set_Play::AddRole(UNum No, const Vector& pos){
	roles[total_roles].No = No;
	roles[total_roles].strategical_pos = pos;
	total_roles ++;
}

bool Set_Play::IsRole(UNum No){
	for(int i = 0; i< total_roles; i++){
		if(No == roles[i].No) return true;
	}
	return false;
}

bool Set_Play::IsExecutor(UNum No){
	return bool(No == Executor);
}

Vector Set_Play::GetSetPlayPos(UNum No){
	for(int i = 0; i< total_roles; i++){
		if (No == roles[i].No){
			return roles[i].strategical_pos;
		}
	}
	return Vector(-1, -1);
}

Vector Set_Play::GetMySetPlayPos(){ 
	return GetSetPlayPos(CP_Number);
}

bool Set_Play::RecentSeen(){
	return bool(situation.SightTime>situation.CurrentTime-2); 
	
}

/********************  Before Kick Off  **********************/
void BeforeKickOff::set_roles(){
	total_roles=0;
	Counter =0;
	AddRole(1, Vector(-47.0f,0.0f));
	AddRole(3, Vector(-25.0f,-12.0f));
	AddRole(5, Vector(-25.0f,12.0f));
	AddRole(2, Vector(-20.0f,20.0f));
    AddRole(4, Vector(-20.0f,-20.0f));
	AddRole(7, Vector( -5.0f,8.0f));
	AddRole(8, Vector( -5.0f,-8.0f));

	if (situation.kickoff_mode == KickOff_Mine)
	{
		AddRole(6, Vector(-15.0f,0.0f));
		AddRole(10, Vector(-0.5f, 0.0f));
		AddRole(9, Vector(-0.1f,18.0f));
		AddRole(11,Vector(-0.1f,-18.0f));
	}
	else{
		AddRole(6,Vector(-15.0f,0.0f));
		AddRole(10, Vector(-10.0f, 0.0f));
		AddRole(9, Vector(-0.5f,18.0f));
		AddRole(11,Vector(-0.5f,-18.0f));
	}
	situation.ChangeSetPlay(SPST_Prego);
}

void BeforeKickOff ::going(){
	if (situation.IsSetPlayState(SPST_Initialize)){
		set_roles();
	}
	Vector initialpos = GetMySetPlayPos();
	if (!motion.infield(initialpos))
		return;
	if (!motion.within_range(initialpos)){
		if (fill_pos_move_time >= 8){
			action.move(initialpos);
			fill_pos_move_time = 0;
		}
		fill_pos_move_time ++;
	}
}


/*****************   Formation   ***************************/
Formation::Formation(){
}


Vector Formation::GetOffensivePoint(float attackness, float leftness, Vector ballpos){
	input[0] = ballpos.x / SP_semi_pitch_length;
	input[1] = -fabs(ballpos.y) / SP_semi_pitch_width;
	input[2] = attackness;
	if (ballpos.y < 0)
		input[3] = leftness;
	else
		input[3] = 1 - leftness;
	FMOffenseNet.SimulateNet(&input[0], &output[0]);
	if (ballpos.y > 0)
		output[1] *=-1;
	return Vector(float(SP_semi_pitch_length * output[0]), float(SP_semi_pitch_width * output[1]));
}

Vector Formation::GetDefensivePoint(const Role & role, Vector ballpos){
	return GetDefensivePoint(role.attackness, role.leftness, ballpos);
}

Vector Formation::GetDefensivePoint(float attackness, float leftness, Vector ballpos){
	Vector fm_defense_point;
	
	pose_limitation(ballpos.x, - SP_semi_pitch_length, SP_semi_pitch_length);
	pose_limitation(ballpos.y, - SP_semi_pitch_width, SP_semi_pitch_width);
	input[0] = ballpos.x / SP_semi_pitch_length;
	input[1] = fabs(ballpos.y) / SP_semi_pitch_width;
	input[2] = attackness;
	if (ballpos.y < 0)
		input[3] = leftness;
	else
		input[3] = 1 - leftness;
	FMDefenseNet.SimulateNet(&input[0], &output[0]);
	if (ballpos.y > 0){
		fm_defense_point = Vector(SP_semi_pitch_length * float(output[0]), SP_semi_pitch_width * float(output[1]));
	}
	else{
		fm_defense_point = Vector(SP_semi_pitch_length * float(output[0]), -SP_semi_pitch_width * float(output[1]));
	}
	return fm_defense_point;
}


Vector Formation::GetFormationPoint(const Player& player){
	Vector ballpos = situation.NextControlBallPos();

	if (situation.IsAttack){
		return GetOffensivePoint(player.attackness, player.leftness, ballpos);
	}
	else{
		return GetDefensivePoint(player.attackness, player.leftness, ballpos);
	}
}

Vector Formation::GetFormationPoint(){
	return GetFormationPoint(Self);
}

float Formation::CalcRelativeDegree(){
	float distance = ball.pos.dist(GetFormationPoint());
	if (distance < 7.0f)
		return 1.0f;
	else if (distance < 20.0f)
		return 1.0f - 0.5f / 13 * distance;
	else
		return 0.5f - 0.5f / 10 *distance > 0 ? 0.5f - 0.5f / 10 *distance : 0;
}

/******************      Field  ***************************/
Field::Field(){
}

void Field::Initialize(){

	field = Rect(-SP_semi_pitch_length, SP_semi_pitch_length,
		-SP_semi_pitch_width,SP_semi_pitch_width);
	l_field = Rect(-SP_semi_pitch_length, SP_semi_pitch_length,
		0, -SP_semi_pitch_width);
	r_field = Rect(-SP_semi_pitch_length, SP_semi_pitch_length,
		0, SP_semi_pitch_width);

	backfield = Rect(-SP_semi_pitch_length, SP_backfieldline,
		-SP_semi_pitch_width,SP_semi_pitch_width);
	forwardfield = Rect(SP_midfieldline, SP_semi_pitch_length,
		-SP_semi_pitch_width, SP_semi_pitch_width);
	midfield = Rect(SP_backfieldline, SP_midfieldline,
		-SP_semi_pitch_width, SP_semi_pitch_width);

	l_backfield = Rect(-SP_semi_pitch_length, SP_backfieldline, 
		-SP_semi_pitch_width, SP_left_wing);
	m_backfield = Rect(-SP_semi_pitch_length, SP_backfieldline,
		SP_left_wing, SP_right_wing);
	r_backfield = Rect(-SP_semi_pitch_length, SP_backfieldline,
		SP_right_wing, SP_semi_pitch_width);

	l_midfield = Rect(SP_backfieldline, SP_midfieldline,
		-SP_semi_pitch_width, SP_left_wing);
	m_midfield = Rect(SP_backfieldline, SP_midfieldline,
		SP_left_wing,SP_right_wing);
	r_midfield = Rect(SP_backfieldline, SP_midfieldline,
		SP_right_wing, SP_semi_pitch_width);

	l_forwardfield = Rect(SP_midfieldline, SP_semi_pitch_length,
		-SP_semi_pitch_width, SP_left_wing);
	m_forwardfield = Rect(SP_midfieldline, SP_semi_pitch_length,
		SP_left_wing, SP_right_wing);
	r_forwardfield = Rect(SP_midfieldline, SP_semi_pitch_length,
		SP_right_wing, SP_semi_pitch_width);

	my_penaltyarea = Rect(-SP_semi_pitch_length, SP_penalty_area_length -SP_semi_pitch_length,
		SP_penalty_area_width /2, - SP_penalty_area_width /2);
	their_penaltyarea = Rect(SP_semi_pitch_length, - SP_penalty_area_length + SP_semi_pitch_length,
		SP_penalty_area_width /2, - SP_penalty_area_width /2);

	my_goaltenderbox = Rect(-SP_semi_pitch_length, -SP_semi_pitch_length + SP_goal_area_length,
		SP_goal_area_width / 2, - SP_goal_area_width / 2);

	their_goaltenderbox = Rect(SP_semi_pitch_length, SP_semi_pitch_length - SP_goal_area_length,
		SP_goal_area_width / 2, - SP_goal_area_width / 2);

	my_goaldefendbox = Rect(-SP_semi_pitch_length, -SP_semi_pitch_length + SP_goal_area_length - SP_catch_area_l /2, 
		SP_goal_area_width / 2, - SP_goal_area_width / 2); 

	shootablearea = Rect(SP_semi_pitch_length, SP_semi_pitch_length - CP_max_shoot_distance,
		-SP_semi_pitch_width, SP_semi_pitch_width);

	expanded_field = Rect(-SP_semi_pitch_length - CP_out_of_pitch_error, SP_semi_pitch_length + CP_out_of_pitch_error,
		-SP_semi_pitch_width - CP_out_of_pitch_error, SP_semi_pitch_width + CP_out_of_pitch_error);

	shrinked_field = Rect(-SP_semi_pitch_length + CP_out_of_pitch_error, SP_semi_pitch_length - CP_out_of_pitch_error,
		-SP_semi_pitch_width + CP_out_of_pitch_error, SP_semi_pitch_width - CP_out_of_pitch_error);

	mygoal = Vector(-SP_semi_pitch_length, 0);
	theirgoal = Vector(SP_semi_pitch_length, 0);

	l_mygoalpost = Vector(-SP_semi_pitch_length, -SP_goal_width / 2.0f);
	r_mygoalpost = Vector(-SP_semi_pitch_length, SP_goal_width / 2.0f);
	l_theirgoalpost = Vector(SP_semi_pitch_length, -SP_goal_width / 2.0f);
	r_theirgoalpost = Vector(SP_semi_pitch_length, SP_goal_width / 2.0f);

	SideLines[SL_Right].LineFromTwoPoints(Vector(SP_semi_pitch_length, 1), Vector(SP_semi_pitch_length, 0));
	SideLines[SL_Left].LineFromTwoPoints(Vector(-SP_semi_pitch_length, 1), Vector(-SP_semi_pitch_length, 0));
	SideLines[SL_Bottom].LineFromTwoPoints(Vector(1, SP_semi_pitch_width), Vector(0, SP_semi_pitch_width));
	SideLines[SL_Top].LineFromTwoPoints(Vector(1, -SP_semi_pitch_width), Vector(0, -SP_semi_pitch_width));

	theirgoalie = -1;
	if(Self.Is_goalie)
		mygoalie = MyNumber;
	else if(FP_goalie_number == MyNumber)
		mygoalie = -1;
	else
		mygoalie = FP_goalie_number;
}

bool Field::IsInvalidZone(Vector Pos){
	switch (situation.playmode){
	case PM_My_Kick_Off:
		return (Pos.x>0);
		break;
	case PM_Their_Kick_Off:
		return (Pos.x>0||ball.pos_valid()&&Pos.dist(ball.pos)<SP_free_kick_buffer);
		break;
	case PM_Their_Goal_Kick :
			return their_penaltyarea.IsWithin(Pos);
			break;
	case PM_Their_Kick_In :
	case PM_Their_Corner_Kick :
	case PM_Their_Free_Kick :
	case PM_Their_Offside_Kick :
		return (Pos.dist(ball.pos)<SP_free_kick_buffer);
		break;
	default :
		return false;
		break;
	}
}

float Field::GetOffsideLine(){
	if(offsideline.IsDataKnown(situation.CurrentTime))	
		return offsideline.Data(situation.CurrentTime);
	
	float offsidex = 0;
	UNum offsidep = -1;
	
	int oppgoalie = fieldinfo.IsValidtheirgoalie() ? fieldinfo.theirgoalie  :  1;
	for(int i = 1; i <= SP_team_size; i ++){
		if(i == oppgoalie) continue;
		if(TheirPlayer(i).Rs_valid() || TheirPlayer(i).pos_valid()){
			if(TheirPlayer(i).pos.x > offsidex){
				offsidex = TheirPlayer(i).pos.x;
				offsidep = i;
			}
		}
	}
	
	float ballposx = ball.pos.x;
	if(situation.BallFree && ball.pos_valid() && ball.Rs_valid())
		ballposx = Max(ballposx, situation.NextControlBallPos().x - 2.0f);
	
	offsideline.Setdata(Max(offsidex, ballposx), situation.CurrentTime);
	if(offsidep > 0) Behindestopp.Setdata(offsidep, situation.CurrentTime);
	
	return offsideline.Data(situation.CurrentTime);
}

bool Field::IsOffside(){
	return bool(Self.pos.x > (GetOffsideLine() - 1.0f));
}

bool Field::Maybeoffside(){
	if(IsOffside()) return true;

	if(!Susoffsideopp.IsDataKnown(situation.CurrentTime)){
		Susoffsideopp.Setdata(GuessOffside(), situation.CurrentTime);
	}

	return bool(Susoffsideopp.Data(situation.CurrentTime) > 0);
}

UNum Field::GuessOffside(){
	if(Self.pos.x <= 0 || Self.pos.x < ball.pos.x) return -1;

	UNum goalieno = fieldinfo.IsValidtheirgoalie() ? fieldinfo.theirgoalie : 1;
	goalieno += SP_team_size;
	UNum susopp = -1;
	for(int i = 0; i < motion.Num_TheirVisiblePlayers(); i++){
		if(motion.TheirPlayer_InFront(i).InsideNO == goalieno) continue;

		if(motion.TheirPlayer_InFront(i).pos.x < Self.pos.x - 10.0f) 
			return susopp;

		if(motion.TheirPlayer_InFront(i).pos.x - motion.TheirPlayer_InFront(i).ActiveRadius(situation.CurrentTime) > Self.pos.x) 
			return -1;		

		if(motion.TheirPlayer_InFront(i).pos.x + motion.TheirPlayer_InFront(i).ActiveRadius(situation.CurrentTime) <= Self.pos.x)
			continue;
		
		if(susopp < 0) susopp = motion.TheirPlayers_InFront[i];
	}
	return susopp;
}

float Field::GetOffensiveSensitivity(Vector pos){
	sensitivity_input[0] = pos.x / 52.5;
	sensitivity_input[1] = fabs(pos.y) / 35;
	OffSensitivityNet.SimulateNet(&sensitivity_input[0], &sensitivity_output);
	return (float)(0.5f + 0.5f * sensitivity_output * 1.111f);
}

float Field::GetDefensiveSensitivity(Vector pos){
	sensitivity_input[0] = pos.x / 52.5;
	sensitivity_input[1] = fabs(pos.y) / 34;
	DF_SensitivityNet.SimulateNet(&sensitivity_input[0], &sensitivity_output);
	return float((sensitivity_output - 0.1) * 1.25);
}

bool Field::WithInField(Vector pos){
	return field.IsWithin(pos);
}

bool Field::WithInExpandedField(Vector pos){
	return expanded_field.IsWithin(pos);
}

bool Field::WithInLeftField(Vector pos){
	return l_field.IsWithin(pos);
}

bool Field::WithInRightField(Vector pos){
	return r_field.IsWithin(pos);
}

bool Field::WithInShootableArea(Vector pos){
	return shootablearea.IsWithin(pos);
}

int Field::NumOpponentsInCone(AngleDeg direction, float distance, AngleDeg angle_range){
	int num = motion.Num_TheirVisiblePlayers();
	int count = 0;
	for(int i = 0; i < num; i ++){
		if (motion.TheirPlayer_Close2Me(i).distance > distance)
			break;
		else{
			if (fabs(NormalizeAngle(motion.TheirPlayer_Close2Me(i).global_angle - direction)) < angle_range / 2.0f){
				count ++;
			}
		}
	}
	return count;
}

int Field::NumOpponentsInCone(Vector pos, AngleDeg direction, float distance, AngleDeg angle_range){
	int num = motion.Num_TheirVisiblePlayers();
	int count = 0;
	Vector gap;
	for(int i = 0; i < num; i ++){
		gap = motion.TheirPlayer_Close2Me(i).pos - pos;
		if (gap.mod2() < Sqr(distance) && fabs(gap.Angle() - direction) <= angle_range /2)
				count ++;
	}
	return count;
}

int Field::NumTeammatesInCone(AngleDeg direction, float distance, AngleDeg angle_range){
	int num = motion.Num_MyVisiblePlayers();
	int count = 0;
	for(int i = 0; i < num; i ++){
		if (motion.MyPlayer_Close2Me(i).distance > distance)
			break;
		else{
			if (fabs(NormalizeAngle(motion.MyPlayer_Close2Me(i).global_angle - direction)) < angle_range / 2.0f){
				count ++;
			}
		}
	}
	return count;	
}


int Field::NumTeammatesInCone(Vector pos, AngleDeg direction, float distance, AngleDeg angle_range){
	int num = motion.Num_MyVisiblePlayers();
	int count = 0;
	Vector gap;
	for(int i = 0; i < num; i ++){
		gap = motion.MyPlayer_Close2Me(i).pos - pos;
		if (gap.mod2() < Sqr(distance) && fabs(NormalizeAngle(gap.Angle() - direction)) <= angle_range /2)
			count ++;
	}
	return count;
}

int Field::NumNoneGoalieTeammatesInCone(Vector end1, Vector end2, AngleDeg angle_range){
	Vector polar = end2 - end1;
	return NumNoneGoalieTeammatesInCone(end1, polar.Angle(), polar.mod(), angle_range);
}

int Field::NumNoneGoalieTeammatesInCone(Vector pos, AngleDeg direction, float distance, AngleDeg angle_range){
	int num = motion.Num_MyVisiblePlayers();
	int count = 0;
	Vector gap;
	for(int i = 0; i < num; i ++){
		if (motion.MyPlayers_Close2Me[i] == FP_goalie_number) continue;
		gap = motion.MyPlayer_Close2Me(i).pos - pos;
		if (gap.mod2() < Sqr(distance) && fabs(NormalizeAngle(gap.Angle() - direction)) <= angle_range /2)
			count ++;
	}
	return count;
}

int Field::NumOpponentsWithin(float distance){
	int num = motion.Num_TheirVisiblePlayers();
	int count = 0;
	for(int i = 0; i < num; i ++){
		if (motion.TheirPlayer_Close2Me(i).distance > distance)
			break;
		count ++;
	}
	return count;
}

int Field::NumTeammatesWithin(float distance){
	int num = motion.Num_MyVisiblePlayers();
	int count = 0;
	for(int i = 0; i < num; i ++){
		if (motion.MyPlayer_Close2Me(i).distance > distance)
			break;
		count ++;
	}
	return count;
}

int Field::NumOpponentsWithin(float Dist, AngleDeg angle, float ofDist, AngleDeg ofangle){
	int num = motion.Num_TheirVisiblePlayers();
	int count = 0;
	for(int i = 0; i < num; i ++){
		if (motion.TheirPlayer_Close2Me(i).distance > ofDist)
			break;
		if (motion.TheirPlayer_Close2Me(i).distance > Dist 
			&& Angle_between(motion.TheirPlayer_Close2Me(i).global_angle, angle, ofangle)){
			count ++;
		}
	}
	return count;
}

int Field::NumTeammatesWithin(float Dist, AngleDeg angle, float ofDist, AngleDeg ofangle){
	int num = motion.Num_MyVisiblePlayers();
	int count = 0;
	for(int i = 0; i < num; i ++){
		if (motion.MyPlayer_Close2Me(i).distance > ofDist)
			break;
		if (motion.MyPlayer_Close2Me(i).distance > Dist 
			&& Angle_between(motion.MyPlayer_Close2Me(i).global_angle, angle, ofangle)){
			count ++;
		}
	}
	return count;
}

int Field::NumOpponentsWithin(float distance, Vector pos){
	int num = motion.Num_TheirVisiblePlayers();
	int count = 0;
	for(int i = 0; i < num; i ++){
		if ((motion.TheirPlayer_Close2Me(i).pos - pos).mod2() < Sqr(distance))
			count ++;
	}
	return count;
}

int Field::NumTeammatesWithin(float distance, Vector pos){
	int num = motion.Num_MyVisiblePlayers();
	int count = 0;
	for(int i = 0; i < num; i ++){
		if ((motion.MyPlayer_Close2Me(i).pos - pos).mod2() < Sqr(distance))
			count ++;
	}
	return count;
}


bool Field::IsBlocked(UNum No){
	if (IsTheirPlayer(No)){
		return IsBlocked(situation.TheirSide, No - SP_team_size);
	}
	else if (IsMyPlayer(No)){
		return IsBlocked(situation.MySide, No);
	}
	return false;
}

bool Field::IsBlocked(char side, UNum No){ 
	if (situation.TheirSide == side)
	{
		Vector OPos = TheirPlayer(No).pos;
		AngleDeg l_goal_ang = (fieldinfo.l_mygoalpost - OPos).Angle();
		AngleDeg r_goal_ang = (fieldinfo.r_mygoalpost - OPos).Angle();
		AngleDeg m_ang = MidAngle2(l_goal_ang,r_goal_ang);
		return (fieldinfo.NumTeammatesInCone(OPos,m_ang,CP_block_dist,AngleDif(l_goal_ang,r_goal_ang)) > 0);
	}
	else{
		if (MyPlayer(No).pos_valid()){
			float angle = (fieldinfo.theirgoal - MyPlayer(No).pos).Angle();
			return bool(fieldinfo.NumNoneGoalieTeammatesInCone(MyPlayer(No).pos, angle, CP_block_dist, 120.0f) > 0);
		}
	}
	return false;
}

float Field::FitPosDegree(Vector CurrentPos, Vector TargetPos, float radius){
	if (radius<1.0f) radius=1.0f;
	return (TargetPos.dist(CurrentPos) < radius ? 1 : 
              1/(1+Sqr(TargetPos.dist(CurrentPos)/radius-1.0f) * 4.0f));	
}

 float Field::FitPosDegree(UNum No, Vector TargetPos, float radius){
	return FitPosDegree(MyPlayer(No).pos, TargetPos, radius);
}

float Field::CalcThreat(UNum No){
	//estimate threat of an opponent
	if (!IsTheirPlayer(No)) return 0.0f;
	return TheirPlayer(No).GetDefsensitivity();
}

/*** The3rdCurve ********/

The3rdCurve::The3rdCurve(float a, float b, float c, float d, float inf, float sup){
	SetABCD(a, b, c, d);
	SetInfSup(inf, sup);
}

void The3rdCurve::SetABCD(float a, float b, float c, float d){
	this->a = a;
	this->b = b;
	this->c = c;
	this->d = d;
}

void The3rdCurve::SetInfSup(float inf, float sup){
	this->inf = inf;
	this->sup = sup;
}

float The3rdCurve::GetOutput(float x, bool limited){
	float result = a *x * x *x + b * x * x + c * x + d;
	if (limited){
		pose_limitation(result, inf, sup);
	}
	return result;
}

bool The3rdCurve::Interpolation(float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3){
	if (x0 == x1 || x0 == x2 || x0 == x3 || x1 == x2 || x1 == x3 || x2 == x3) return false;
	float t0, t1, t2, t3;
	t0 = y0 / ((x0 - x1) * (x0 - x2) * (x0 - x3));
	t1 = y1 / ((x1 - x0) * (x1 - x2) * (x1 - x3));
	t2 = y2 / ((x2 - x0) * (x2 - x1) * (x2 - x3));
	t3 = y3 / ((x3 - x0) * (x3 - x1) * (x3 - x2));

	a = t0 + t1 + t2 + t3;
	b = - (t0 * (x1 + x2 + x3) + t1 * (x0 + x2 + x3) + t2 * (x0 + x1 + x3) + t3 * (x0 + x1 +x2));
	c = t0 * (x1 * x2 + x1 * x3 + x2 * x3) + t1 * (x0 * x2 + x0 * x3 + x2 * x3)
		+ t2 * (x1 * x0 + x3 * x1 + x3 * x0) + t3 * (x0 * x1 + x0 * x2 + x1 * x2);
	d = -(t0 * (x1 * x2 * x3) + t1 * (x0 * x2 * x3) + t2 * (x0 * x1 * x3) + t3 * (x0 * x1 * x2));
	return true;
}

/******************** Place Kick Arrangement ***********/
void PlaceKick::CheckHeadingAndStamina()
//check if the executer is going to the ball properly
{	UNum i;
	Vector meanvel,balldir;

	for (i=2;i<=SP_team_size;i++)
	{
	if (Trackmark[i].x>SP_pitch_length/2 || !MyPlayer(i).pos_valid()) 
	{
		HeadingToBall[i] = 1.0f - 2.0f * (situation.CurrentTime - situation.ModeChangedTime)/ SP_drop_ball_time;
	}
	else
	{
		meanvel=MyPlayer(i).pos-Trackmark[i];
		balldir=ball.pos-MyPlayer(i).pos;
		if (balldir.mod()<1.0f)
			HeadingToBall[i]=1.0f;
		else
		{
			balldir.Normalize();
			HeadingToBall[i]=
				((meanvel.x*balldir.x+meanvel.y*balldir.y)/CP_setplay_speed+1)/2;
		}
	}
	pose_limitation(HeadingToBall[i],0,1);
	if (MyPlayer(i).pos_valid())
				Trackmark[i]=MyPlayer(i).pos;
	else 
				Trackmark[i]=Vector(SP_pitch_length,SP_pitch_width);
	}

}

UNum PlaceKick::JudgeExecutor()
{	
	Vector meanvel,balldir;
	UNum i,select=MyNumber;
	float min_pos_dis,min_cur_dis,player_i_pos_dis,player_i_cur_dis;
	Vector ofpos;
	if (situation.IsGoalieKick) return FP_goalie_number;
	if (ball.kickable()) return MyNumber;
	if (IsMyPlayer(situation.ballcontroller)&&(fieldinfo.mygoalie != situation.ballcontroller))
		return situation.ballcontroller;
	if (Counter%6==0&&Counter > 0) CheckHeadingAndStamina();
	ofpos=fm.GetOffensivePoint(Self.attackness, Self.leftness, ball.pos);
	min_pos_dis=ofpos.dist(ball.pos);
	min_cur_dis=ball.distance;
	for (i=2;i<=SP_team_size;i++)
		if (i!=MyNumber)
	{
		if (HeadingToBall[i]<0.75) continue;
		//filter out players that do not intend to get ball								
		ofpos=fm.GetOffensivePoint(FP_attackness[Idx(i)],FP_leftness[Idx(i)],ball.pos);
		player_i_pos_dis=ofpos.dist(ball.pos);
		player_i_cur_dis=MyPlayer(i).pos.dist(ball.pos);
		if (player_i_pos_dis>1.2*min_pos_dis) continue;
		if (min_pos_dis>1.2*player_i_pos_dis)
		{
			select=i;
			min_pos_dis=player_i_pos_dis;
			min_cur_dis=player_i_cur_dis;
		}
		else
		{
			if (player_i_cur_dis>1.2*min_cur_dis) continue;
			if (min_cur_dis>1.2*player_i_cur_dis||i>select)
			{
				select=i;
				min_pos_dis=player_i_pos_dis;
				min_cur_dis=player_i_cur_dis;
			}
		}
	}
	return select;
}


bool PlaceKick::CanExecute(){
	float passpriority;
	int i,bestpass;
	if (!(situation.IsSetPlayState(SPST_Executing)||RecentSeen()))
	{
		DoLog(18,"Cannot Exec 1");
		return false;
	}
	//Wait for best chance to pass
	canpass=motion.SmartPass();
	float maxpriority=0.5f;
	DoLog(18,"num pass %.0f",(float)motion.num_passinfo);
	for(i = 0; i < motion.num_passinfo; i++){
		if (motion.passinfo[i].PassSpeed()<SP_kickable_margin*0.7f)
			continue;
		passpriority =motion.GetPassPriority(motion.passinfo[i]);
		if(passpriority>maxpriority){
			maxpriority=passpriority;
			bestpass=i;
		}
	}
	if( maxpriority >0.5f ){
		motion.logpass=motion.passinfo[bestpass];
		DoLog(18,"Can pass");
		canpass = true;
	}
	else
		canpass = false;
	if (canpass) return true;
	if (Counter>CP_setplay_wait_time||situation.IsSetPlayState(SPST_Executing)) return true; 
		else 
		{
			return false;
		}

}

bool PlaceKick::ExecutorGo(){
	if (ball.kickable())
		return false;
	Vector Execpos=ball.pos+Polar2Vector(CP_reliable_kickarea-SP_ball_size,GetFaceAngle());
	bool success = motion.smartgoto(Execpos,CP_setplay_speed);
	if (!success){
		float runspeed = ball.kickable() ? Execpos.dist(Self.pos)*(1 - SP_player_decay):Execpos.dist(Self.pos);

		if (Execpos.dist(Self.pos)>0.1f || !ball.kickable())
		{
			if (fabs(NormalizeAngle((Execpos-Self.pos).Angle()-Self.bodyfacing))>15)
				motion.turn_to(Execpos);
			else
				motion.run(runspeed);
			return true;
		}
		else
		{
			DoLog(18,"Already get ball");
			return false;
		}
	}
	else
		return true;
}

AngleDeg PlaceKick::GetFaceAngle(){
	Vector pos0(Self.pos.x,0),pos1,pos2;
	AngleDeg ang0 = 0,ang1,ang2;
	float w0,w1,w2;

	if (Self.pos.y>0)
	{
		pos1 = Vector(Self.pos.x,SP_pitch_width/4);
		ang1 = (fieldinfo.l_theirgoalpost-pos1).Angle();
		pos2 = Vector(Self.pos.x,SP_pitch_width/2);
		ang2 = -90;
	}
	else{
		pos1 = Vector(Self.pos.x,-SP_pitch_width/4);
		ang1 = (fieldinfo.r_theirgoalpost-pos1).Angle();
		pos2 = Vector(Self.pos.x,-SP_pitch_width/2);
		ang2 = 90;
	}
	w0 =1.0f/( (pos0 - Self.pos).mod2()+0.01f);
    w1 =1.0f/( (pos1 - Self.pos).mod2()+0.01f);
    w2 =1.0f/( (pos2 - Self.pos).mod2()+0.01f);
    return (w0*ang0+w1*ang1+w2*ang2)/(w0+w1+w2);
}

void PlaceKick::Execute(){
	Command command;
	//move to the ball
	if (ExecutorGo()) 
	{
		DoLog(18,"Going to ball");
		return;
	}
	
	if (CanExecute())
	{
		situation.ChangeSetPlay(SPST_Executing);
		if (canpass)
		{
			DoLog(18,"Pass");
			KT_Res passok = motion.pass(motion.logpass);
		}
		else
		{
			DoLog(18,"Dribble");
			motion.k_StepDribble(GetFaceAngle(), 90.0f, 1.0f); 
		}
	}
	else
	{
		DoLog(18,"Cannot execute");
		my_error("Cannot execute");
	}
}


void PlaceKick::going(){
	UNum i;
	if (!ball.pos_valid())
		return;
	else
	{		
		switch (situation.PlaceKickState) {
		case SPST_Initialize:
			Counter=0;
			for (i=1;i<=SP_team_size;i++)
				if (MyPlayer(i).pos_valid())
					Trackmark[i]=MyPlayer(i).pos;
				else 
					Trackmark[i]=Vector(SP_pitch_length,SP_pitch_width);
			for (i=1;i<=SP_team_size;i++)
				HeadingToBall[i]=1.0f;
			situation.ChangeSetPlay(SPST_Prego);
		case SPST_Prego:	
			
			Executor=JudgeExecutor();
			DoLog(18,"Executor %.0f",(float)Executor);
			motion.B_evaluate();	//initialization
			
			if (Executor==MyNumber)
			{
				if (!ExecutorGo())
				{
					situation.ChangeSetPlay(SPST_ExecWait);
					Counter = Max(2 * CP_setplay_wait_time - (SP_drop_ball_time + situation.ModeChangedTime - situation.CurrentTime),0);
				}
			}
			
			else
			{
				motion.Offense();
						
				if (IsMyPlayer(situation.ballcontroller)&&(situation.ballcontroller==MyNumber||
					MyPlayer(situation.ballcontroller).balldist<0.9*ball.distance)
					&&(situation.IsGoalieKick || fieldinfo.mygoalie != situation.ballcontroller))
				{
					situation.ChangeSetPlay(SPST_ExecWait);
					Counter = Max(2 * CP_setplay_wait_time - (SP_drop_ball_time + situation.ModeChangedTime - situation.CurrentTime),0);
				}
			}
			
			break;
		case SPST_ExecWait:
		case SPST_Executing:
			
			my_error("On Execute %d",Executor);
			DoLog(18,"On Execute %.0f",(float)Executor);
			if (Executor==MyNumber)
				Execute();			
			else
				motion.Offense();	
			break;
		case SPST_AfterExec:
			DoLog(18,"(SP)Should Not get here!!");
			my_error("Should not get here!!");
			break;
		default:
			break;
		}
		Counter++;
		DoLog(18,"(SP) Counter %.0f",(float)Counter);

	}
}

GoaliePlaceKick::GoaliePlaceKick(){
MovePoint[0]=Vector(-47,-15);
MovePoint[1]=Vector(-47,15);
MovePoint[2]=Vector(-40,-10);
MovePoint[3]=Vector(-40,10);
MovePoint[4]=Vector(-43.5,-15);
MovePoint[5]=Vector(-43.5,15);
movetime = -1;
}

float GoaliePlaceKick::CalcAngle(Vector ATpos,Vector DFpos){
	/*ATPOSĽǶԱŽ֮
		ΪΣնȵĹ
		1ԱԶƶ
		2ԱԶƶ
		3ԱԽԱ½ǶԽãֵԽС
	*/
	AngleDeg dfangle=float(fabs(Ray(DFpos,(fieldinfo.l_mygoalpost-DFpos).Angle()).StrechAngle(fieldinfo.r_mygoalpost)));
	if (ATpos.dist(DFpos)<9) 
		ATpos=DFpos+Polar2Vector(9,(ATpos-DFpos).Angle());
	Ray lcourse(ATpos,(fieldinfo.l_mygoalpost-ATpos).Angle());
	Ray rcourse(ATpos,(fieldinfo.r_mygoalpost-ATpos).Angle());
	AngleDeg atangle1=float(fabs(lcourse.StrechAngle(DFpos)));
	AngleDeg atangle2=float(fabs(rcourse.StrechAngle(DFpos)));
	if (atangle1+atangle2>dfangle) return 1;
	else return (atangle1+atangle2)/dfangle;
}

float GoaliePlaceKick::CalcPass(Vector pos){
	PassInfo currinfo;
	Ray ballcourse;
	float bestpr,currpr;
	float factpasspr,factintrdange,factreceivedange,factcontrol;
	float banang,bangoalcross,banstrive,banclose,bancriticarea;

	Vector dangerpos;
	PassInfo potninfo;
	bestpr=-1.0f;
	//Ray goalieray(fieldinfo.mygoal,(pos-fieldinfo.mygoal).Angle());
	DoLog(18,"Numpasses %.0f",(float)motion.num_passinfo);
	for (UNum i =0;i<motion.num_passinfo;i++)
	{
		currinfo=motion.passinfo[i];
		banang = LowPassFilter(85,95,(float)fabs(currinfo.PassAngle()));
		if (fabs((pos-fieldinfo.mygoal).Angle())< 10)
			bangoalcross = 1.0f;
		else
		bangoalcross = pos.y<0 ? LowPassFilter(-5,5,currinfo.PassAngle()):HighPassFilter(-5,5,currinfo.PassAngle());
		banclose = HighPassFilter(CP_closeinterception_max_cycles-1.0f,(float)CP_closeinterception_max_cycles,currinfo.MinReceiverCycles())
					*HighPassFilter(CP_goalie_kick_min_vel-0.1f,CP_goalie_kick_min_vel,currinfo.PassSpeed());
		bancriticarea = LowPassFilter(0.85f,0.9f,fieldinfo.GetDefensiveSensitivity(currinfo.ReceiveBallPoint()));
		//banstrive = HighPassFilter(currinfo.MinReceiverCycles(),currinfo.MinReceiverCycles()+1.0f,currinfo.MinOpponentCycles());
		banstrive = 1.0f;
		factcontrol=(float)(1-Sqr(currinfo.MinReceiverCycles()/currinfo.MinOpponentCycles()));
		ballcourse.SetValues(pos,currinfo.PassAngle());
		dangerpos=ballcourse.GetClosestPoint(currinfo.MostDangerOpponentPos());
										//;صΣ
		factpasspr=motion.GetPassPriority(currinfo);//
		if (factpasspr < 0) factpasspr = 0;
		if (ballcourse.OnRay(dangerpos) && currinfo.MostDangerOpponent()!=-1)
				factintrdange=2*CalcAngle(dangerpos,pos)*(float)sqrt(1-fieldinfo.GetDefensiveSensitivity(dangerpos));
			else 
				factintrdange=0;
		factreceivedange=2*CalcAngle(currinfo.ReceiveBallPoint(),pos)*fieldinfo.GetDefensiveSensitivity(currinfo.ReceiveBallPoint());
										//ڵΣ
		currpr=0.3f*factpasspr+0.2f*(1 - factintrdange)+0.3f*(1 - factreceivedange)+0.2f*factcontrol;
		
		currpr *= banang * bangoalcross * banstrive * banclose * bancriticarea;

		if (bestpr<currpr)
		{
			bestpr=currpr;
			motion.logpass=currinfo;
		}

	}
	return bestpr;

}

float GoaliePlaceKick::GetControl(int mynum,int oppnum)
{
	if (oppnum==0) return 1.0f;
	float factop,factmine;
	if (oppnum==1) factop=0.2f;
	else if (oppnum==2) factop=0.5f;
	else factop=0.8f;

	if (mynum==1) factmine=0.1f;
	else if (oppnum==2) factmine=0.3f;
	else factmine=0.4f;

	if (factmine>factop) return 1.0f;
	else
		return 1.0f+factmine-factop;

}

float GoaliePlaceKick::CalcPoint(Vector pos){
	if (fieldinfo.my_goaltenderbox.IsWithin(pos))
		return 0;
	AngleDeg direction=NormalizeAngle((pos-fieldinfo.mygoal).Angle());
	int N_N_Teammates=fieldinfo.NumTeammatesInCone(pos,direction,15,150);
	int N_M_Teammates=fieldinfo.NumTeammatesInCone(pos,direction,30,150)-N_N_Teammates;
	int N_F_Teammates=fieldinfo.NumTeammatesInCone(pos,direction,40,150)-N_M_Teammates-N_N_Teammates;
	int N_N_Opponents=fieldinfo.NumOpponentsInCone(pos,direction,15,150);
	int N_M_Opponents=fieldinfo.NumOpponentsInCone(pos,direction,30,150)-N_N_Opponents;
	int N_F_Opponents=fieldinfo.NumOpponentsInCone(pos,direction,40,150)-N_M_Opponents-N_N_Opponents;
	return GetControl(N_N_Teammates,N_N_Opponents)*0.5f+GetControl(N_M_Teammates,N_M_Opponents)*0.3f
		+GetControl(N_F_Teammates,N_F_Opponents)*0.2f;
}

AngleDeg GoaliePlaceKick::MyFaceAngle(Vector pos){
	return 0;
	AngleDeg res = NormalizeAngle((pos-fieldinfo.mygoal).Angle());
	if (res>45) res = 45;
	if (res<-45) res = -45;
	return res;
}

bool GoaliePlaceKick::ShouldMove(){
	UNum j,bestpnt=0;
	float calcpr,movepr,nativepr;
	if (situation.SightTime <= movetime)
		return false;
	if (situation.playmode!=PM_My_Free_Kick||movecount == 0)
		return false;
	else
	{
		if (passpriority>0.3f) return false;
		if (!RecentSeen())
		return false;

		nativepr = CalcPoint(Self.pos);
		DoLog(18,"(GSP)native (%f %f) calcvalue %f",Self.pos.x,Self.pos.y,nativepr);
		movepr=0.0f;
		for (j=0;j<CP_NumGoalieMovePoint;j++)
		{
			calcpr=CalcPoint(MovePoint[j]);
			DoLog(18,"(GSP)point (%f %f) calcvalue %f",MovePoint[j].x,MovePoint[j].y,calcpr);
			if (movepr<calcpr)
			{	bestpnt=j;
				movepr=calcpr;
			}
		}

		if ((movecount==SP_goalie_max_moves && movepr>1.1*nativepr || movepr>1.2*nativepr
			||deadtime>=CP_setplay_wait_time-3)&&MovePoint[bestpnt].dist(Self.pos)>4)
		{
			DoLog(18,"(GSP)Move %.0f to (%f %f) ",3.0f-movecount,MovePoint[bestpnt].x,MovePoint[bestpnt].y);
			action.move(MovePoint[bestpnt]);
			movetime = situation.CurrentTime;
			deadtime = Max(2 * CP_setplay_wait_time - (SP_drop_ball_time + situation.ModeChangedTime - situation.CurrentTime),0);
			movecount--;
			return true;
		}
		else
		{
			return false;
		}
	}
}

bool GoaliePlaceKick::CanKickOut(){
	if (motion.Num_MyVisiblePlayers() > 0 && motion.MyPlayer_Close2Me(0).distance < 6.0f)
	{
		DoLog(18,"Teammate too close ");
		return false;
	}
	if (!(RecentSeen()||situation.IsSetPlayState(SPST_Executing))||(situation.SightTime <= movetime))
		{
			DoLog(18,"No Start without sight %d",situation.SightTime);
			return false;
		}
	if ((action.cancatch(situation.CurrentTime) || situation.playmode == PM_My_Goal_Kick) && passpriority>0.2f)
		return true;
	else 
		return false;

}

bool GoaliePlaceKick::KickOut(){
	DoLog(18,"Pass to angle %.2f receiver %.0f speed %.2f",motion.logpass.PassAngle(),
		float(motion.logpass.Receiver()),motion.logpass.PassSpeed());
	KT_Res passok = motion.pass(motion.logpass);
	motion.GoalieJustKicked = true;
	situation.ChangeSetPlay(SPST_Executing);
	mediator.SetViewWidth(VW_Normal);
	return true;
}

 void GoaliePlaceKick::going(){
	if (!ball.pos_valid())
		return;
	UNum i;
	float mvspeed=0;
	if (situation.IsGoalieKick)
	{				
		if (situation.IsSetPlayState(SPST_Initialize))
			deadtime = 0;
		if (ball.kickable())
		{	
		    if (!situation.IsSetPlayState(SPST_Executing))
				situation.ChangeSetPlay(SPST_ExecWait);
						
			motion.B_evaluate();
			motion.SmartPass();
			passpriority=CalcPass(Self.pos);
			DoLog(18,"pass priority %f",passpriority);

			if (!ShouldMove()) 
			{			
				DoLog(18,"(GSP)No move required");
				if (fabs(NormalizeAngle(MyFaceAngle(Self.pos) - Self.bodyfacing)) > 15.0f  
					&& !situation.IsSetPlayState(SPST_Executing))  //additional of visual
					motion.turn_to(MyFaceAngle(Self.pos));
				else
				if(CanKickOut())
				{
					DoLog(18,"(GSP)Good Chance to start");
					KickOut();
				}
				else
				{  
					/*ûдʱı仯Աƶ
					յȴ򿪴ƻ*/
					for (i=0;i<motion.Num_MyVisiblePlayers();i++)
						mvspeed+=MyPlayer(motion.MyVisiblePlayers[i]).speed;
					mvspeed/=motion.Num_MyVisiblePlayers();
					if (mvspeed>0.1)  deadtime=Max(2 * CP_setplay_wait_time - (SP_drop_ball_time + situation.ModeChangedTime - situation.CurrentTime),0); 
					else deadtime++;
					DoLog(18,"Teammate Unresponsed %.0f",(float)deadtime);
					if (deadtime>CP_setplay_wait_time&&RecentSeen()
						//ball.speed_valid()&&ball.speed>0.1f||
					    ||situation.IsSetPlayState(SPST_Executing)) 
					{
						if (motion.Clear_Ball())
						{
							motion.GoalieJustKicked = true;
							situation.ChangeSetPlay(SPST_Executing);
							deadtime=Max(2 * CP_setplay_wait_time - (SP_drop_ball_time + situation.ModeChangedTime - situation.CurrentTime),0); 
							DoLog(18,"Clear Ball");
						}
						else
						{
							if (situation.IsSetPlayState(SPST_Executing))
								kick.stopball();
							DoLog(18,"Clear Unconfirmed");
						}
					}
					else
					{
						DoLog(18,"(GSP)Hehe,Wait a minute.");
					}
				}
			}
		}
		else
		{
			if (situation.IsSetPlayState(SPST_Executing))
				motion.SmartInterception();
			else
			{	
				situation.ChangeSetPlay(SPST_Prego);
				motion.smartgoto(ball.pos+Vector(0.7f,0),CP_setplay_speed);
				
			}
		}
	}
	else
	{
		situation.ChangeSetPlay(SPST_ExecWait);
		DoLog(18,"(GSP) none of my business");
		motion.GoalieRunto(motion.GetGuardPoint(),0,0.4f);
	}

}