/*
    Copyright (C) 2001  Tsinghuaeolus

    Authors : ChenJiang, YaoJinyi, CaiYunpeng, Lishi

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

	If you make any changes or have any comments we would appreciate a 
	message to yjy01@mails.tsinghua.edu.cn.
*/

#include "kick.h"
#include "global.h"

/**************	My OA	****************/
void MyOA::Initvpos(char* filename){
	int i,j,k,l;
	for(i=0; i<KICKDIS; i++){
		for(j=0; j<KICKAGL; j++){
			oapos[i][j].idx_dis = i;
			oapos[i][j].idx_agl = j;
			oapos[i][j].enabled = true;
			oapos[i][j].pos = Polar2Vector(SP_kickable_margin*(i+0.5f)/KICKDIS + SP_ball_size + SP_player_size,360.0f*j/KICKAGL);
		}
	}

	FILE* fp;
	if(filename!=NULL) fp = fopen(filename,"r");
	if(fp == NULL){
		for(i=0; i<KICKDIS; i++){
			for(j=0; j<KICKAGL; j++){
				for(k=0; k<KICKDIS; k++){
					for(l=0; l<KICKAGL; l++){
						oapos[i][j].nextp[k*KICKAGL+l] = &oapos[k][l];
					}
				}
			}
		}	
		return;
	}
	fscanf(fp,"%f",&Tvel.x);
	fscanf(fp,"%f",&Tvel.y);

	Vector vel;
	float maxvel = Sqr(SP_ball_speed_max);
	float minv = -1.0f;
	float tmp;
	for(i=0; i<KICKDIS; i++){
		for(j=0; j<KICKAGL; j++){
			for(k=0; k<KICKDIS; k++){
				for(l=0; l<KICKAGL; l++){
					oapos[i][j].nextp[k*KICKAGL+l] = &oapos[k][l];
					fscanf(fp,"%f",&tmp);
					oapos[i][j].qvalue[k*KICKAGL+l] = tmp;
				}
			}
			//
			for(k = 0; k<KICKDIS*KICKAGL-1; k++){
				int max = k;
				float maxv = oapos[i][j].qvalue[max];
				for(l = k+1; l<KICKDIS*KICKAGL; l++){
					if(oapos[i][j].qvalue[l] > maxv){
						max = l;
						maxv = oapos[i][j].qvalue[l];
					}
				}
				if(max != k){
					Swap(oapos[i][j].qvalue[k],oapos[i][j].qvalue[max]);
					Swap(oapos[i][j].nextp[k],oapos[i][j].nextp[max]);
				}
			}
			oapos[i][j].nextp[KICKDIS*KICKAGL] = NULL;
		}
	}

	if(fp!=NULL) fclose(fp);
} 

void MyOA::SetInitState(Vector pos, Vector vel, Vector pvel, float bodyfacing){
	s_pos = pos;
	s_vel = vel;
	p_vel[0] = pvel;
	for(int i=1; i<maxdepth; i++) p_vel[i] = p_vel[i-1] * SP_player_decay;
	bdfacing = bodyfacing;
	//float fvalue = motion.GetStateV(pos,vel);
	float hestimate = 0;
	int idx1= getdisidx(pos.mod());
	int idx2 = getaglidx(pos.Angle());

	kstate s(getkickpos(idx1,idx2));
	s.depth = 0;
	SetInitial(s,hestimate);
	DoLog(LOG_OAKICK, "setinitial");
}

void MyOA::Extend(OAnode<kstate>* pnode){
	if(pnode == NULL) return;
	if(pnode->data.depth >= maxdepth) return;

	kstate* pstate = &pnode->data;
	OApos* npos;	
	float hvalue,cost;
	Vector pos,vel,pmove;
	if(pnode->Getwaypre() == NULL){
		pos = s_pos;
		vel = s_vel;
	}else{
		pos = pnode->data.pos->pos;
		Vector lpos;
		if(pnode->Getwaypre()->Getwaypre() == NULL) lpos = s_pos;
		else lpos = pnode->Getwaypre()->data.pos->pos;
		vel = (pos + p_vel[pnode->data.depth-1] - lpos) * SP_ball_decay;
	}
	int counts = 0;
	while((npos = pstate->Subnode()) != NULL && (counts++)< 100){
		pstate->subnode ++;
		if(!npos->enabled) {
			DoLog(LOG_OAKICK,"skip (%.2f %.2f) %d", npos->pos.x, npos->pos.y, counts);
			continue;
		}
		if(!Canreachpos(pos, vel, bdfacing, npos->pos+p_vel[pnode->data.depth]))	continue;
		kstate nstate(npos);
		if(closelist.FindNode(nstate) != NULL) continue;
		//enqueue cstate's valid nextsubnode
		OAnode<kstate>* ptmp = openlist.FindNode(nstate);
		hvalue = pstate->pos->qvalue[pstate->subnode-1];
		cost = pnode->Getcost() + stepcost;
		if(ptmp == NULL){
			kstate s(npos);
			OAnode<kstate>* node = new OAnode<kstate>(nstate);
			node->Sethvalue(hvalue);
			node->Setcost(cost);
			node->Setwaypre(pnode);
			node->data.depth = pnode->data.depth + 1;
			openlist.enqueue(node);
			break;
		}else if(ptmp->Gethvalue() < hvalue){
			ptmp->Sethvalue(hvalue);
			OAnode<kstate>* ptmp2 = ptmp->Getwaypre(); 
			ptmp->Setwaypre(pnode);
			ptmp->data.depth = ptmp2->data.depth + 1;
			//extend another pstate's sub node
			pnode = ptmp2;
			pstate = &pnode->data ;
		}
	}
	DoLog(LOG_OAKICK,"extend counts %d",counts);
} 

void MyOA::Extendcurnode(){
	if(curnode == NULL) return;
	DoLog(LOG_OAKICK,"oa extend 1");
	Extend(curnode->Getwaypre());
	DoLog(LOG_OAKICK,"oa extend 2");
	Extend(curnode);
}

bool MyOA::IsTargetState(){
	Vector pos,vel;
	if(curnode->Getwaypre() == NULL){
		pos = s_pos;
		vel = s_vel;
	}else{
		Vector lpos;
		pos = curnode->data.pos->pos;
		if(curnode->Getwaypre()->Getwaypre() == NULL){
			lpos = s_pos;
		}else
			lpos = curnode->Getwaypre()->data.pos->pos;
		vel = (pos + p_vel[curnode->data.depth-1] - lpos) * SP_ball_decay;
	}
	return Canreachvel(pos,vel,bdfacing,dvel);
}

void MyOA::Enableall(){
	for(int i=0; i<KICKDIS; i++){
		for(int j=0; j<KICKAGL; j++){
			oapos[i][j].enabled = true;
		}
	}
} 

Vector MyOA::Getnextpos(){
	OAnode<kstate>* tmp1 = curnode;
	OAnode<kstate>* tmp2 = tmp1->Getwaypre(); 
	while(tmp2 != NULL){
		if(tmp2->Getwaypre() == NULL){
			return tmp1->data.pos->pos;
		}
		tmp1 = tmp2;
		tmp2 = tmp1->Getwaypre();
	}
	return Vector(0,0);
}

Vector MyOA::Getfirstkick(){
	OAnode<kstate>* tmp1 = curnode;
	OAnode<kstate>* tmp2 = tmp1->Getwaypre(); 
	while(tmp2 != NULL){
		if(tmp2->Getwaypre() == NULL){
			return tmp1->data.pos->pos + p_vel[0] - s_pos - s_vel;
		}
		tmp1 = tmp2;
		tmp2 = tmp1->Getwaypre();
	}
	return Vector(-s_vel.x, -s_vel.y);
}

bool MyOA::Canreachvel(Vector& bpos, Vector& bvel, float bdfacing, Vector& next_bvel, float* effort){
	float dist_ball = bpos.mod() - SP_player_size - SP_ball_size;
	if(dist_ball > SP_kickable_margin) return false;
	float dir_diff = (float)fabs(Deg2Rad(NormalizeAngle(bpos.Angle() - bdfacing)));
	float eff_power = SP_max_power * SP_kick_power_rate * 
		(1 - 0.25f*dir_diff/PI - 0.25f*dist_ball/SP_kickable_margin);//to be modified

	float gap = (next_bvel- bvel).mod() ;
	if(effort!=NULL) *effort = gap / eff_power ;
	if(gap > eff_power)	return false;
	else return true;
}

bool MyOA::Canreachpos(Vector& bpos, Vector& bvel, float bdfacing, Vector& next_bpos, float* effort){
	Vector tvel = next_bpos - bpos;
	if(tvel.mod2() > Sqr(SP_ball_speed_max)) return false;
	return Canreachvel(bpos,bvel,bdfacing,tvel,effort);
}

/***********	Kick	********************/
Kick::Kick(){
	speedinc[0] = 0.0f;
	maxinc_angle[0] = 0.0f;
	speedinc[1] = 0.0078f;
	maxinc_angle[1] = 5.7296f;
	speedinc[2] = 0.0316f;
	maxinc_angle[2] = 11.4592f;
	speedinc[3] = 0.0718f;
	maxinc_angle[3] = 17.7617f;
	speedinc[4] = 0.1293f;
	maxinc_angle[4] = 24.0642f;
	speedinc[5] = 0.2051f;
	maxinc_angle[5] = 30.9397f;
	speedinc[6] = 0.3000f;
	maxinc_angle[6] = 38.3882f;
	speedinc[7] = 0.4143f;
	maxinc_angle[7] = 46.4096f;
	speedinc[8] =  0.5475f;
	maxinc_angle[8] = 55.0039f;
	speedinc[9] = 0.6974f;
	maxinc_angle[9] = 64.7442f;
	speedinc[10] = 0.8601f;
	maxinc_angle[10] = 75.0575f;
	speedinc[11] = 1.0274f;
	maxinc_angle[11] = 83.0789f;
	speedinc[12] = 1.1455f;
	maxinc_angle[12] = 83.0789f;
	speedinc[13] = 1.1928f;
	maxinc_angle[13] = 83.0789f;
	speedinc[14] = 1.1677f;
	maxinc_angle[14] = 83.0789f;
	speedinc[15] = 1.0711f;
	maxinc_angle[15] = 83.0789f;
	speedinc[16] = 0.9060f;
	maxinc_angle[16] = 83.0789f;
	speedinc[17] = 0.6774f;
	maxinc_angle[17] = 83.0789f;
	speedinc[18] = 0.3951f;
	maxinc_angle[18] = 83.0789f;
}
/* Function smartkick  kick to desire angle with desire speed*/
float Kick::smartkick(float kickspeed, AngleDeg kickangle, Command& command, KK_Mode mode){
	command.Reset();
	if (!ball.kickable()) return 0.0f;

	Vector dvel;
	if (ball.speed_valid())
		dvel = Polar2Vector(kickspeed, kickangle) - ball.global_vel;
	else
		dvel = Polar2Vector(kickspeed, kickangle);

	float angle = NormalizeAngle(dvel.Angle() - Self.bodyfacing);
	float power = dvel.mod() / action.Kickrate();

	if (power > SP_max_power || power < SP_min_power){
		/* try to kick harder than allowed*/
		if (!ball.speed_valid() || ball.speed < FLOAT_EPS){
			pose_limitation(power, SP_min_power, SP_max_power);
		}
		else{
			/* try to kick to the right direction, the kickspeed may not be desire, but
			the angle is correct. And we do our best with the kickspeed.
			It is a ray intersection problem. The idea is from CMU.*/
			Vector solution1, solution2, gap;
			Ray vel(ball.pos, kickangle);
			int num_of_solutions = vel.CircleIntersect(SP_max_power * action.Kickrate(), 
				ball.PredictPos(1), solution1, solution2);
			if (num_of_solutions == 0){
				/* can not kick to the right trajectory. just try to stop it*/
				power = ball.speed / action.Kickrate();
				pose_limitation(power, SP_min_power, SP_max_power);
				angle = NormalizeAngle(ball.vel_angle - Self.bodyfacing + 180);
				command.kick(power, angle);
				my_error("No solution. So stop it!");
			}
			else{
				if (mode == KK_fast)
					gap = solution2 - ball.PredictPos(1);
				else
					gap = solution1 - ball.PredictPos(1);
				power = gap.mod() / action.Kickrate();
				angle = NormalizeAngle(gap.Angle() - Self.bodyfacing);
				pose_limitation(power, SP_min_power, SP_max_power);
			}
		}
	}

	Vector kickvel;
	if (ball.speed_valid()){
		kickvel = ball.global_vel + Polar2Vector(power * action.Kickrate(), angle + Self.bodyfacing);
	}
	else{
		kickvel = Polar2Vector(power * action.Kickrate(), angle + Self.bodyfacing);
	}
	command.kick(power, angle);
	Vector kickeffect = Polar2Vector(power, angle + Self.bodyfacing) * action.Kickrate();
	return kickvel.mod();
}

float Kick::smartkick(float speed, AngleDeg ang, KK_Mode mode){
	Command command;
	float kickspeed;
	kickspeed = smartkick(speed, ang, command, mode);
	mediator.enroll(command, Action_other, PriorityA);
	return kickspeed;
}

KT_Res Kick::turn_ball(TurnDir dir, AngleDeg turn_angle, float turnball_radius, KK_Mode kk_mode){
	Command command;
	KT_Res res = turn_ball(dir, turn_angle, command, turnball_radius, kk_mode);
	mediator.enroll(command, Action_turn_ball, PriorityA);
	return res;
}

KT_Res Kick::turn_ball(TurnDir dir, AngleDeg turn_angle, Command& command, float turnball_radius, KK_Mode kk_mode){
	if(!ball.kickable()){
		return KT_lostball;
	}
	else {
		float kickspeed, kickangle;
		if (ball.distance < CP_turnball_radius_threshold) {
			// the ball is too close
			kickspeed = CP_kickballaway_speed;
			kickangle = ball.global_angle;

			CorrectForMyVel(kickspeed, kickangle);
			smartkick(kickspeed, kickangle, command, kk_mode);
			return KT_kickaway;
		}
		else {
			Vector vel;
			vel = Polar2Vector((ball.distance + turnball_radius) /2, dir * turn_angle) - 
				 Polar2Vector(ball.distance , 0);
			kickspeed = vel.mod();
			kickangle = NormalizeAngle(ball.global_angle + vel.Angle());
	
			if (!CanKickOut(ball.pos - Self.pos, kickspeed, kickangle)){
				// the ball course is too close to the player
				kickangle = NormalizeAngle(180.0f + ball.global_angle - dir * (float)ASin(CP_turnball_radius_threshold/ ball.distance));
				kickspeed = (float)(sqrt(Sqr(ball.distance) - Sqr(CP_turnball_radius_threshold))
					+ sqrt(Sqr(turnball_radius) - Sqr(CP_turnball_radius_threshold)));
			}

			CorrectForMyVel(kickspeed, kickangle);

			float candospeed = smartkick(kickspeed, kickangle, command, kk_mode);
/*			if ((Polar2Vector(candospeed, kickangle) - Self.global_vel).mod2() > Sqr(SP_kickable_area) && kk_mode != KK_fast){
				motion.smartkick(0.0f, 0.0f, command);
				return KT_stopball;
			}
			else{*/
				return KT_success;
			//}
		}
	}
}

KT_Res Kick::turn_ball(AngleDeg turn_angle, float turnball_radius, KK_Mode kk_mode){
	Command command;
	KT_Res res = turn_ball(turn_angle, command, turnball_radius, kk_mode);
	mediator.enroll(command, Action_turn_ball, PriorityA);
	return res;
}

KT_Res Kick::turn_ball(AngleDeg turn_angle, Command& command, float turnball_radius, KK_Mode kk_mode){
	if (turn_angle < 0)
		return turn_ball(Turn_cw, turn_angle, command, turnball_radius);
	else
		return turn_ball(Turn_cw, turn_angle, command, turnball_radius);
}

void Kick::CorrectForMyVel(float& kickspeed, float& kickangle){
	Vector kickvel = Polar2Vector(kickspeed, kickangle) + Self.global_vel;
	kickspeed = kickvel.mod();
	kickangle = kickvel.Angle();
}

bool Kick::stopball(){
	Command command;
	bool success = stopball(command);
	mediator.enroll(command, Action_other, PriorityA);
	return success;
}

bool Kick::stopball(Command& command){
	if (!ball.kickable()) return false;
	Vector gap = ball.pos - Self.PredictPos(1);
	gap.Normalize();
	gap = gap * CP_desired_kickarea + Self.PredictPos(1) - ball.pos;
	smartkick(gap.mod(), gap.Angle(), command);
	if((command.GetPower() > 0.1f * SP_max_power) || ball.distance > 0.9f * CP_reliable_kickarea){

	}else
		command.stay(); 
	return true;
}

bool Kick::CanKickOut(Vector ballrelpos, float speed, float angle){
	ballrelpos += Polar2Vector(speed, angle);
	return bool(ballrelpos.mod2() >= Sqr(CP_turnball_radius_threshold));
}

KT_Res Kick::wisekick(float kickspeed, AngleDeg kickangle, KK_Mode kk_mode){
	Command command;
	KT_Res success = wisekick(kickspeed, kickangle, command, kk_mode);
	mediator.enroll(command, Action_dribble, PriorityA);
	return success;
}

KT_Res Kick::wisekick(float kickspeed, AngleDeg kickangle, Command& command, KK_Mode kk_mode){
	command.Reset();
	if (!ball.kickable()) return KT_lostball;
	float candospeed;
	if (kk_mode == KK_fast)
		candospeed = smartkick(SP_ball_speed_max, kickangle, command);
	else
		candospeed = smartkick(kickspeed, kickangle, command);
	if (candospeed > CP_kick_threshold_factor * kickspeed || candospeed >= 1.90f){
		if ((ball.pos + Polar2Vector(kickspeed, kickangle)).dist2(Self.PredictPos())
			< Sqr(SP_player_size + SP_ball_size)){
			//can not kick out, so turn the ball
			DoLog(11,"Adjust 30");
			if (NormalizeAngle(ball.global_angle - kickangle) < 0){
				turn_ball(30.0f, command);
				return KT_adjustball;
			}
			else{
				turn_ball(-30.0f, command);
				return KT_adjustball;
			}
		}
		else{
			// can kick out
			return KT_success;
		}
	}
	else{
		//see if an acceleration helps
		//first see the range of angle i can turn the ball to, it is a two circle intersection
		//problem
		Vector ballpos = ball.pos - Self.pos;
		Vector ballpred = ballpos + ball.global_vel;

		ballpos = ballpos.Rotate(-kickangle); // Rotate to relative coords
		ballpred = ballpred.Rotate(-kickangle); // Rotate to relative coords
		float ballangle = ballpos.Angle();
		float ball_on_left = bool(ballangle <= 0.0f) ? 1.0f : -1.0f;
		ballangle = (float)fabs(ballangle);
		float circle2_angle = ballpred.Angle();
		if (ball_on_left == 1.0f)
			circle2_angle = - circle2_angle;
		float dist_between_circles = ballpred.mod();
		
		float maxkickeffect = action.Max_KickEffect();

		float upangle = 0.0f, lowangle = 0.0f;

		if (dist_between_circles <= CP_turnball_radius + maxkickeffect
			&& dist_between_circles > CP_turnball_radius - maxkickeffect){
			lowangle = 0.0f;
			upangle = 360.0f;

			if (dist_between_circles > maxkickeffect - CP_turnball_radius){
				float tmpangle = ACos((Sqr(dist_between_circles) + Sqr(CP_turnball_radius) - Sqr(maxkickeffect))
					/ (2 * CP_turnball_radius * dist_between_circles));
				lowangle = NormalizeAngle(ballangle - circle2_angle - tmpangle, 0);
				upangle = NormalizeAngle(ballangle - circle2_angle + tmpangle, 0);
			}
		}
		float maxincangle = getmaxincangle(ballangle);

		if (!Angle_between(maxincangle, lowangle, upangle)){
			maxincangle = NormalizeAngle(getspeedinc(ballangle, upangle) > getspeedinc(ballangle, lowangle)
				? upangle : lowangle);
		}
		
		float predictkickeffect = (float) action.CalcKickRate(ball.global_angle + ball_on_left * maxincangle - Self.bodyfacing, CP_turnball_radius) * SP_max_power;


		float max_accelerated_speed = getspeedinc(ballangle, (float)fabs(maxincangle)) + predictkickeffect;
		if (max_accelerated_speed >= kickspeed || max_accelerated_speed > 1.95f){
				// a kick can get enough acceleration
			turn_ball(ball_on_left * maxincangle, command, CP_powerkick_radius, KK_fast);
			return KT_accelerate;
		}

		float speedincneeded = kickspeed - maxkickeffect;
		if(speedincneeded < 0.05f){
			turn_ball(0.0f, command, CP_powerkick_radius);
		}
		else if (speedincneeded <= 0.1f){
			if (ballangle < 40.0f)
				turn_ball(-ball_on_left * (40.0f - ballangle), command, CP_powerkick_radius);
			else
				turn_ball(0.0f, command);
		}
		else if (speedincneeded <= 0.3f){
			if (ballangle < 65.0f)
				turn_ball(-ball_on_left * (65.0f - ballangle), command, CP_powerkick_radius);
			else
				turn_ball(0.0f, command, CP_powerkick_radius);
		}
		else if (speedincneeded <= 0.6f){
			if (ballangle < 130.0f){
				if (ballangle < 90.0f)
					turn_ball(-ball_on_left * (90.0f - ballangle), command, CP_powerkick_radius);
				else
					turn_ball(0.0f, command, CP_powerkick_radius);
			}
			else{
				if (ballangle > 170.0f)
					turn_ball(-ball_on_left * (170.0f - ballangle), command, CP_powerkick_radius);
				else
					turn_ball(0.0f, command, CP_powerkick_radius);
			}
		}
		else{
			//if (ballangle < 130.0f){
				//if(ballangle < 135.0f)
					turn_ball(-ball_on_left * (135.0f - ballangle < 80 ? 135.0f - ballangle : 80.0f), command, CP_powerkick_radius);
				//else
				//	turn_ball(0.0f, command, CP_powerkick_radius);
			/*}
			else{
				if (ballangle > 150.0f)
					turn_ball(-ball_on_left * (150.0f - ballangle), command, CP_powerkick_radius);
				else
					turn_ball(0.0f, command, CP_powerkick_radius);
			}*/
		}
	}
	return KT_adjustball;
}


float Kick::getmaxincangle(AngleDeg startangle){
	startangle = (float)fabs(NormalizeAngle(startangle)) / 10;
	return (maxinc_angle[int(startangle)] * (1 + int(startangle) - startangle) +
		maxinc_angle[int(startangle) + 1] * (startangle - int(startangle)));
}

AngleDeg Kick::getspeedinc(AngleDeg startangle, AngleDeg turnangle){
	float d = 2 * CP_turnball_radius * Sin(turnangle / 2);
	float u = Sqr(CP_mnmx_kickeffect) - Sqr(d * Cos(startangle - turnangle / 2));
	if (u < 0) return 0.0f;
	return float(sqrt(u) + d * fabs(Sin(startangle - turnangle / 2))) - CP_mnmx_kickeffect;
}


void Kick::SetOainf(){
	oainf_time = situation.CurrentTime; 
	oainf_Miny = - SP_semi_pitch_width - Self.pos.y;
	oainf_Maxy = SP_semi_pitch_width - Self.pos.y;
	oainf_Maxx = SP_semi_pitch_length - Self.pos.x;
	oainf_Minx = - SP_semi_pitch_length - Self.pos.x;
	if(oainf_Miny > -SP_kickable_area || oainf_Maxy < SP_kickable_area
		|| oainf_Minx > -SP_kickable_area || oainf_Maxx < SP_kickable_area){
		line_limit = true;
		DoLog(LOG_OAKICK,"oa limited line(%.1f, %.1f, %.1f, %.1f)", oainf_Miny, oainf_Maxy, oainf_Minx, oainf_Maxx );
	}else
		line_limit = false;

	int num = Max(3, motion.Num_TheirVisiblePlayers());
	for(int i=0; i<num; i++){
		if(motion.TheirPlayer_Close2Me(i).distance > SP_feel_distance || !motion.TheirPlayer_Close2Me(i).pos_valid()) break;
		float maxradius = SP_kickable_area + SP_player_speed_max + 0.6f * (situation.CurrentTime - motion.TheirPlayer_Close2Me(i).original_postime);
		float dis = motion.TheirPlayer_Close2Me(i).distance - maxradius;
		if(dis > SP_kickable_area )	break;
		oainf_pos[i] = motion.TheirPlayer_Close2Me(i).rel_pos;
		if(dis > 0)
			oainf_radius[i] = maxradius;
		else
			oainf_radius[i] = Max(SP_kickable_area, motion.TheirPlayer_Close2Me(i).distance);
		DoLog(LOG_OAKICK,"oa limited pos(%.1f, %.1f), radius %.1f", oainf_pos[i].x, oainf_pos[i].y, oainf_radius[i]);
	}
	pos_limit= i;

	if(!line_limit && pos_limit ==0){
		DoLog(LOG_OAKICK,"kick no limited");
		if(!oainf_allenabled){
			Enableall();
			oainf_allenabled = true;
		}
		return;
	}
	oainf_allenabled = false;
}

void Kick::Oa_limitpos(float rel_angle){
	for(int j=0; j<KICKAGL; j++){
		for(int i=0; i<KICKDIS; i++){
			Vector pos = KickOA1_5.getkickpos(i,j)->pos.Rotate(rel_angle) + Self.global_vel;
			if(line_limit){
				if(pos.x > oainf_Maxx || pos.x < oainf_Minx || pos.y > oainf_Maxy || pos.y < oainf_Miny){
					Disablepos(i,j);
					//DoLog(LOG_OAKICK,"disable (%.2f %.2f)", pos.x,pos.y);
					continue;
				}
			}
			for(int k=0; k<pos_limit; k++){
				if(pos.dist(oainf_pos[k]) <= oainf_radius[k]){
					Disablepos(i,j);
					break;
					//DoLog(LOG_OAKICK,"disable (%.2f %.2f)", pos.x,pos.y);
				}
			}
			if(k == pos_limit) Enablepos(i,j);
		}
	}
}

void Kick::Enablepos(int disidx, int aglidx){
	KickOAlast.Enable(disidx,aglidx);
	KickOA2_5.Enable(disidx,aglidx);
	KickOA2_0.Enable(disidx,aglidx);
	KickOA1_5.Enable(disidx,aglidx);
}

void Kick::Disablepos(int disidx, int aglidx){
	KickOAlast.Disable(disidx,aglidx);
	KickOA2_5.Disable(disidx,aglidx);
	KickOA2_0.Disable(disidx,aglidx);
	KickOA1_5.Disable(disidx,aglidx);
}

void Kick::Enableall(){
	KickOAlast.Enableall();
	KickOA2_5.Enableall();
	KickOA2_0.Enableall();
	KickOA1_5.Enableall();
} 

KT_Res Kick::oakick(float kickspeed, AngleDeg kickangle, KK_Mode kk_mode){
	Command command;
	KT_Res success = oakick(kickspeed, kickangle, command, kk_mode);
	mediator.enroll(command, Action_dribble, PriorityA);
	return success;
}

KT_Res Kick::oakick(float kickspeed, AngleDeg kickangle, Command& command, KK_Mode kk_mode){	
	Vector bpos = ball.rel_pos.Rotate(-kickangle);
	Vector bvel = ball.global_vel.Rotate(-kickangle);
	Vector pvel = Self.global_vel.Rotate(-kickangle);
	
	float bdfacing = NormalizeAngle(Self.bodyfacing - kickangle);
	float difficult = kickspeed + (float) fabs(bdfacing) / 180.0f * 0.25f * 2.7f;
	MyOA* useoa;
	if(difficult > 2.5f)
		useoa = &KickOAlast;
	else if(kickspeed > 2.0f)
		useoa = &KickOA2_5;
	else if(kickspeed > 1.5f)
		useoa = &KickOA2_0;
	else
		useoa = &KickOA1_5;
	useoa->SetInitState(bpos, bvel, pvel, bdfacing);
	useoa->SetDesirevel(Polar2Vector(kickspeed,0));
	if(useoa->IsTargetState()){//only one kick
		if (kk_mode == KK_fast){
			smartkick(SP_ball_speed_max, kickangle, command, kk_mode);
		}
		else
			smartkick(kickspeed,kickangle,command,kk_mode);
		return KT_success;
	} 

	if(oainf_time != situation.CurrentTime) SetOainf(); 
	if(!oainf_allenabled) Oa_limitpos(kickangle);

	DoLog(LOG_OAKICK,"start oa find");
	bool find = useoa->Findoptimalway();
	if(!find){//impossible
		DoLog(LOG_OAKICK, "oa find fail");
		return wisekick(kickspeed,kickangle,command,kk_mode);
	}else{
		DoLog(LOG_OAKICK,"find success");
		Vector kick = useoa->Getfirstkick();
		float power = kick.mod() / action.Kickrate();
		float angle = NormalizeAngle(kick.Angle() + kickangle - Self.bodyfacing);
		pose_limitation(power, SP_min_power, SP_max_power);
		command.kick(power,angle);

		Vector kickeffect = ball.global_vel + Polar2Vector(power * action.Kickrate(), NormalizeAngle(angle+ Self.bodyfacing));
		Vector nextbpos = ball.rel_pos + kickeffect;
		Vector oabpos = useoa->Getnextpos();
		oabpos = oabpos.Rotate(kickangle);
		DoLog(LOG_OAKICK, "oa find after %d, oapos(%.2f, %.2f), kick to(%.2f %.2f)", useoa->Getsearchcounts(), oabpos.x, oabpos.y, nextbpos.x, nextbpos.y);
		if(useoa->Getneedkicks() == 1) return KT_accelerate;
		else return KT_adjustball;
	}
}