/*
    Copyright (C) 2001  Tsinghuaeolus

    Authors : ChenJiang, YaoJinyi, CaiYunpeng, Lishi

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

	If you make any changes or have any comments we would appreciate a 
	message to yjy01@mails.tsinghua.edu.cn.
*/

#include "pass.h"
#include "global.h"

/********************	Pass	*******************************/
Pass::Pass(){	
}

bool Pass::SmartPass(){
	
	ResetPassInfo();
	SetConsideredRoutes();
	SetConsideredPlayers();
	DoLog(LOG_PASS, "Considered players %d considered routes %d", num_players_considered,
		num_routes_considered);
	for(int i = 0; i < num_routes_considered; i ++){
		Explore(i);
	}
	DoLog(LOG_PASS, "numpassinfos %d", num_passinfo);
	return num_passinfo > 0;
	
}

KT_Res Pass::pass(PassInfo& passinf, Command& command){
	command.Reset();
	return kick.oakick(passinf.PassSpeed(), passinf.PassAngle(), command);
}

KT_Res Pass::pass(PassInfo& passinf){
	Command command;
	KT_Res res = pass(passinf, command);
	mediator.enroll(command, Action_pass, PriorityA);
	return res;
}

void Pass::AddPassInfo(UNum receiver, AngleDeg PS_angle, float PS_speed, bool passable){
	PassInfo inf;
	if (getpassinfo(receiver, PS_angle, PS_speed, inf)){
		if (get_pass_confidence(inf) > 0.5)
			AddPassInfo(inf, passable);
	}
}

void Pass::AddPassInfo(const PassInfo& passinf, bool passable){
	//if (passable){//passable means whether being confidential
	if(true){
		if(num_passinfo >= CP_max_passinfos) return;
		passinfo[num_passinfo++] = passinf;	
	}
	else{
		if(num_potential_passinfo >= CP_max_passinfos) return;
		potential_passinfo[num_potential_passinfo++] = passinf;
	}
}

void Pass::ResetPassInfo(){
	num_passinfo = 0;
	num_potential_passinfo = 0;
}

bool Pass::Is_Angle_Passable(AngleDeg angle){
	return sensory.DirConf(angle) > 0.77f && sensory.DirConf(angle + CP_passable_angle) > 0.77f
		&& sensory.DirConf(angle - CP_passable_angle) > 0.77f;
}

float Pass::PASSWeightByDistance(float distance){
	float weight = CP_passweight_conf_a * Sqr(distance) + CP_passweight_conf_b * distance + CP_passweight_conf_c;
	pose_limitation(weight, 0, 1);
	return weight;
}

float Pass::get_pass_confidence(PassInfo &inf){	
	return motion.calculatecontrol(inf.min_receiver_cycles, inf.min_opponent_cycles);
}


bool Pass::getpassinfo(UNum receiver, AngleDeg PS_angle, float PS_rapidness, PassInfo& passinfo){
	InterceptionInfo inf;
	float min_receiver_cycles = InfCycles, min_opponent_cycles = InfCycles, cycles;
	UNum most_danger_opponent = -1;
	Vector ballpos, ballvel, receive_ball_point;
	ballpos = ball.pos;
	ballvel = Polar2Vector(PS_rapidness, PS_angle);

	motion.getinterceptioninfo(MyPlayer(receiver).pos, ballpos, MyPlayer(receiver).global_vel, ballvel, MyPlayer(receiver).bodyfacing, inf);
	receive_ball_point = inf.IT_point;

	if (!fieldinfo.WithInField(receive_ball_point) || inf.IT_cycles >= InfCycles || inf.IT_cycles < 0)
		return false;

	min_receiver_cycles = inf.IT_cycles;

	int num = motion.Num_TheirVisiblePlayers();
	for (int i =0; i < num; i++){
		cycles = receive_ball_point.dist(TheirPlayer(motion.TheirVisiblePlayers[i]).pos) / 
			CP_assumed_interceptspeed;
		if (cycles < min_opponent_cycles){
			min_opponent_cycles = cycles;
			most_danger_opponent = motion.TheirVisiblePlayers[i];
		}
	}

	passinfo = PassInfo(receiver, PS_rapidness, PS_angle, min_receiver_cycles,
		min_opponent_cycles, most_danger_opponent, receive_ball_point);
	return true;
}


Line& Pass::Pline(int idx1, int idx2){
	if (!pline.IsDataKnown(idx1, idx2, situation.CurrentTime)){
		Line line;
		if (idx1 > idx2){
			Swap(idx1, idx2);
		}

		if (Players_considered[idx1] == MyNumber){
			line.LineFrompline(0, GetPlayer_considered(idx2).rel_pos_2_ball);
		}
		else{
			line.LineFrompline(GetPlayer_considered(idx1).rel_pos_2_ball, GetPlayer_considered(idx2).rel_pos_2_ball);
		}
		pline.SetData(line, idx1, idx2, situation.CurrentTime);
	}

	return pline.Data(idx1, idx2);
}

void Pass::SetConsideredPlayers(){
	num_players_considered = 0;
	Players_considered[num_players_considered ++] = MyNumber;

	int num = motion.Num_VisiblePlayers();
	
	for(int i = 0; i < num; i ++){
		if (num_players_considered >= CP_max_considerd_pass_players)
			break;
		Players_considered[num_players_considered ++] = motion.Players_Close2Ball[i];
	}
}

void Pass::SetConsideredRoutes(){
//set routes
	ResetRoutesBuffer();

	Ordered_Soft_Queue<float, SP_team_size> my_angles_queue;
	
	for(int No = 1; No <= SP_team_size; No ++){
		if (No != MyNumber && MyPlayer(No).pos_valid()){
			my_angles_queue.enqueue(MyPlayer(No).global_angle);
		}
	}

	int p = my_angles_queue.ActualHead();

	if (my_angles_queue.IsEmpty()) return;
	
	
	float ang, ang1, ang2, prev_ang = 360.0f;
	while(my_angles_queue.IsValid(p)){
		if (prev_ang == 360.0f){
			ang2 = my_angles_queue.GetData(my_angles_queue.CircularPrev(p));
		}
		else{
			ang2 = prev_ang;
		}

		ang1 = my_angles_queue.GetData(p);
		if (fabs(NormalizeAngle(prev_ang - ang1, 0.0f)) < 5.0){
			goto  GoNext;
		}

		prev_ang = ang1;
		
		if(!AddRoutes(ang1))
			break;

		ang = NormalizeAngle(ang2 - ang1, 0.0f);

		if (ang < 10.0f){
		//no route
		}
		else if (ang < 30.0f){
		//one route
			AddRoutes(NormalizeAngle(ang1 + ang /2.0f));
		}
		else if (ang < 60.0f){
			AddRoutes(NormalizeAngle(ang1 + ang / 3.0f));
			AddRoutes(NormalizeAngle(ang2 - ang / 3.0f));
		}
		else if (ang < 90.0f){
			AddRoutes(NormalizeAngle(ang1 + ang / 3.0f));
			AddRoutes(NormalizeAngle(ang1 + ang / 2.0f));
			AddRoutes(NormalizeAngle(ang2 - ang / 3.0f));
		}
		else if (ang < 120.0f){
			AddRoutes(NormalizeAngle(ang1 + ang / 6.0f));
			AddRoutes(NormalizeAngle(ang1 + ang / 3.0f));
			AddRoutes(NormalizeAngle(ang1 + ang / 2.0f));
			AddRoutes(NormalizeAngle(ang2 - ang / 3.0f));
			AddRoutes(NormalizeAngle(ang2 - ang / 6.0f));
		}
		else{
			AddRoutes(NormalizeAngle(ang1 + 15.0f));
			AddRoutes(NormalizeAngle(ang1 + 30.0f));
			AddRoutes(NormalizeAngle(ang1 + 60.0f));
			AddRoutes(NormalizeAngle(ang2 - 60.0f));
			AddRoutes(NormalizeAngle(ang2 - 30.0f));
			AddRoutes(NormalizeAngle(ang2 - 15.0f));
		}
GoNext:
		p = my_angles_queue.Next(p);
	}

	p = routes_queue.ActualHead();
	num_routes_considered = 0;
	while(routes_queue.IsValid(p)){
		if (num_routes_considered >= CP_max_considered_pass_routes)
			break;
		Routes_considered[num_routes_considered ++] = routes_queue.GetData(p);
		p = routes_queue.Next(p);
	}
}

bool Pass::AddRoutes(AngleDeg angle){
	if (routes_queue.NumDatas() >= CP_max_considered_pass_routes)
		return false;
	return routes_queue.enqueue(angle);
}

void Pass::ResetRoutesBuffer(){
	routes_queue.cleanup();
	num_routes_considered = 0;
}

Player& Pass::GetPlayer_considered(int idx){
#ifdef _DEBUG_MOD
	if (idx < 0 || idx >= num_players_considered){
		DoLog(LOG_PASS, "Bug here");
	}
#endif
	return GetPlayer(Players_considered[idx]);
}


int Pass::GetNearestExploredRoute(int routeidx){
	if (routeidx < 0 || routeidx >= num_routes_considered)
		return -1;
	int clockwise_idx, counter_idx;
	clockwise_idx = counter_idx = -1;
	for(int i = routeidx + 1; i < num_routes_considered; i ++){
		if (IsExplored(i)){
			clockwise_idx = i;
			break;
		}
	}

	for(i = routeidx -1; i >= 0; i --){
		if (IsExplored(i)){
			counter_idx = i;
			break;
		}
	}

	if (clockwise_idx == -1 && counter_idx == -1){
	//no route explored
		return -1;
	}

	if (clockwise_idx == -1){
		for(i = 0; i < counter_idx; i ++){
			if (IsExplored(i)){
				clockwise_idx = i;
				break;
			}
		}
		if (clockwise_idx == -1)
			return counter_idx;
	}

	if (counter_idx == -1){
		for(i = num_routes_considered; i > clockwise_idx; i --){
			if (IsExplored(i)){
				counter_idx = i;
				break;
			}
		}
		DoLog(LOG_PASS, "clockwise_idx %d counter %d", clockwise_idx, counter_idx);
		if (counter_idx == -1)
			return clockwise_idx;
	}

	return fabs(NormalizeAngle(Routes_considered[routeidx] - Routes_considered[clockwise_idx]))
			< fabs(NormalizeAngle(Routes_considered[routeidx] - Routes_considered[counter_idx])) ?
		clockwise_idx : counter_idx;
}

bool Pass::IsExplored(int idx){
	if (idx < 0 || idx >= num_routes_considered) return false;
	return num_controllers[idx].time == situation.CurrentTime;
}

bool Pass::Explore(int routeidx){
	if (routeidx < 0 || routeidx >= num_routes_considered) return false;

	Vector point;

	ctrl_list.cleanup();

	ballcourse = Ray(ball.pos, Routes_considered[routeidx]);
	//set pitch constraint
	if (fieldinfo.shrinked_field.RayIntersection(ballcourse, point)){
		ctrl_list.GetData(ctrl_list.Head()) = ballcourse.DistanceFromOrigin(point);
	}
	else{
	//max distances
		ctrl_list.GetData(ctrl_list.Head()) = SP_pitch_diameter;
	}

	ballcourse = Ray(0, Routes_considered[routeidx]);
	ballcourse.Normalize();

//first see if there is some information that can foster the search
	int nearest_explored_route = GetNearestExploredRoute(routeidx), i, begin;
//add controllers in the nearest route first
	if (nearest_explored_route != -1){
		begin = ctrl_list.Head();
		for(i = 0; i < num_controllers[nearest_explored_route].data; i ++){
			AddtoCtrlList(controllers_of_line[nearest_explored_route][i], begin);
		}
	}

	begin = ctrl_list.Head();
	for(i = 0; i < num_players_considered; i ++){
		AddtoCtrlList(i, begin);
	}

	num_controllers[routeidx].data = 0;
	num_controllers[routeidx].time = situation.CurrentTime;

	int q = ctrl_list.ActualHead();
	while(ctrl_list.IsValid(q)){
		if (num_controllers[routeidx].data > CP_max_considerd_pass_players)
			break;
		deliminator[routeidx][num_controllers[routeidx].data] = ctrl_list.GetData(q);
		controllers_of_line[routeidx][num_controllers[routeidx].data ++] = q;
		q = ctrl_list.Next(q);
	}

	GetGT_speedinfo(routeidx);
	return true;
}


bool Pass::GetGT_speedinfo(int routeidx){
	if (!IsExplored(routeidx))
		return false;

	Vector point;
	float min_gt_speed = 0.0f, peak_point;
	int peakpoint_considered_idx =  -1, i, j, num;
	num = num_controllers[routeidx].data;
	/*DoLog(LOG_PASS, "angle %.2f controllers %d", Routes_considered[routeidx], num);
	char string[1000];
	int len = 0;
	string[0] = 0;
	for(i = 0; i < num; i ++){
		len += sprintf(string + len, "%d %.2f ", Players_considered[controllers_of_line[routeidx][i]], 
			deliminator[routeidx][i]);
	}
	DoLog(LOG_PASS, string);*/

	for(i = 0; i < num; i++){
		if (GetPlayer_considered(controllers_of_line[routeidx][i]).IsMyside()){
			for(j = peakpoint_considered_idx + 1; j < i; j ++){
				if(i == 0){
					//impossible branch
					break;
				}
				if (Players_considered[controllers_of_line[routeidx][j]] == MyNumber)
					continue;
				point = GetPlayer_considered(controllers_of_line[routeidx][j]).rel_pos_2_ball.Rotate(- Routes_considered[routeidx]);
				peak_point = calc_peakpoint(point);
				if(peak_point < deliminator[routeidx][i - 1]){
					min_gt_speed = Max(calc_spottingspeed(point, peak_point), min_gt_speed);
				}
				if (min_gt_speed >= SP_ball_speed_max)
					break;
			}

			if (min_gt_speed >= SP_ball_speed_max)
					break;

			point = GetPlayer_considered(controllers_of_line[routeidx][i]).rel_pos_2_ball.Rotate(- Routes_considered[routeidx]);
			min_gt_speed = Max(calc_spottingspeed(point, deliminator[routeidx][i - 1]),
				min_gt_speed);

			if (min_gt_speed >= SP_ball_speed_max)
					break;

			speed_inf[routeidx][i] = min_gt_speed;

			peak_point = calc_peakpoint(point);
			speed_sup[routeidx][i] = 0.0f;
			if (peak_point < deliminator[routeidx][i]){
				speed_sup[routeidx][i] = calc_spottingspeed(point, peak_point);
			}
			speed_sup[routeidx][i] = Max(calc_spottingspeed(point, deliminator[routeidx][i])
				, speed_sup[routeidx][i]);
			min_gt_speed = Max(min_gt_speed, speed_sup[routeidx][i]);
			speed_sup[routeidx][i] = calc_spottingspeed(point, deliminator[routeidx][i]);
			speed_sup[routeidx][i] = Min(speed_sup[routeidx][i], SP_ball_speed_max);
			peakpoint_considered_idx = i;

			//add passinfo
			CalcPassinfos(speed_inf[routeidx][i], speed_sup[routeidx][i], Routes_considered[routeidx],
				Players_considered[controllers_of_line[routeidx][i]]);

			if (min_gt_speed >= SP_ball_speed_max)
					break;
		}
	}
	return true;
}

bool Pass::AddtoCtrlList(int p, int& begin){
	float t;
	int q = begin;

//find insert point
	if (!ctrl_list.IsValid(q)){
		DoLog(LOG_PASS, "bug q not valid");
		return false;
	}

	while(ctrl_list.IsValid(ctrl_list.Next(q))){
		if (ctrl_list.Next(q) > p)
			break;
		q = ctrl_list.Next(q);
	}

	begin = q;

	if (q == p){
		return false;
	}
	int q_bk = q;

	if (!ctrl_list.IsHead(q)){
		t = getdeliminator(ballcourse, q, p);

		if (t >= ctrl_list.GetElement(q).data || t < 1.0f){
			return false;
		}
	}

//add to the list
	ctrl_list.Insert(q, p);
	begin = p;

//adjust the list
	bool first_run = true;
	while(ctrl_list.IsValid(q) && !ctrl_list.IsHead(q)){
		if (!first_run){
			t = getdeliminator(ballcourse, q, p);
		}
		else{
			first_run = false;
		}

		if ((ctrl_list.IsHead(ctrl_list.Prev(q)) && t > 0) || t > ctrl_list.GetData(ctrl_list.Prev(q))){
			ctrl_list.GetData(q) = t;
			break;
		}
		if (ctrl_list.IsHead(ctrl_list.Prev(q)) && t <= 0){
			DoLog(LOG_PASS, "deleter %d %.2f at (%.2f %.2f) ", Players_considered[p], t, GetPlayer_considered(p).rel_pos_2_ball.x,
				GetPlayer_considered(p).rel_pos_2_ball.y);
			DoLog(LOG_PASS, "pline (%.2f %.2f %.2f) deletee %d", Pline(p, q).A, Pline(p, q).B, Pline(p, q).C, Players_considered[q]);
		}
		float u = ballcourse.intersection(Pline(p, q_bk));
		ctrl_list.Delete(q);
		q = ctrl_list.Prev(p);
	}
	q = ctrl_list.Next(p);
	while(ctrl_list.IsValid(q)){
		t = getdeliminator(ballcourse, q, p);
		if(t < ctrl_list.GetData(q) && t > 0.0f){
			ctrl_list.GetData(p) = t;
			break;
		}
		ctrl_list.Delete(q);
		q = ctrl_list.Next(p);
	}

	if (ctrl_list.IsTail(p)){
	// p is the tail
		ctrl_list.GetData(p) = ctrl_list.GetData(ctrl_list.Head());
	}
	return true;
}


bool Pass::CalcPassinfos(float min_speed, float max_speed, AngleDeg passangle, UNum No){
	if (max_speed - min_speed < 0.2f)
		return false;
	if (!IsMyPlayer(No))
		return false;
	if (min_speed >= CP_impossible_speed)
		return false;

	float weight = PASSWeightByDistance(MyPlayer(No).balldist);

	AddPassInfo(No, NormalizeAngle(passangle), min_speed*(1 - weight) + max_speed * weight, true);

	int num_speed_divisions = int((max_speed - min_speed) / 0.4f);
		
	for (int j = 0; j < num_speed_divisions; j ++){
		AddPassInfo(No, passangle, min_speed + (max_speed - min_speed) * (j + 1)  / (num_speed_divisions + 1), true);
	}
	return true;
}

float Pass::getdeliminator(Ray& ballcourse, int idx1, int idx2){
	if (idx1 < 0 || idx1 > num_players_considered || idx2 < 0 || idx2 > num_players_considered)
		return SP_pitch_diameter;
	float t = ballcourse.intersection(Pline(idx1, idx2));
	if(fabs(t) < 0){
		Line line(ballcourse);
		if (line.dist2(GetPlayer_considered(idx1).pos) < line.dist2(GetPlayer_considered(idx2).pos)){
			return SP_pitch_diameter;
		}
		else{
			return 0;
		}
	}
	return t;
}

void Pass::LogPass(float angle){
	int routeidx = -1;
	float min_angle_gap = 720.0f, angle_gap;
	for(int i = 0; i < num_routes_considered; i ++){
		angle_gap = (float)fabs(NormalizeAngle(angle - Routes_considered[i]));
		if (angle_gap < min_angle_gap){
			routeidx = i;
			min_angle_gap = angle_gap;
		}
	}
	if (routeidx == -1)
		return;
	DoLog(LOG_PASS, "angle %.2f controllers %d", Routes_considered[routeidx], num_controllers[routeidx].data);
	char string[1000];
	int len = 0;
	string[0] = 0;
	for(i = 0; i < num_controllers[routeidx].data; i ++){
		len += sprintf(string + len, "%d %.2f ", Players_considered[controllers_of_line[routeidx][i]], 
			deliminator[routeidx][i]);
	}
	DoLog(LOG_PASS, string);
	Vector point;
	float peak_point;
	for(i = 0; i < num_controllers[routeidx].data; i ++){
		if (GetPlayer_considered(controllers_of_line[routeidx][i]).IsMyside()){
			point = GetPlayer_considered(controllers_of_line[routeidx][i]).rel_pos_2_ball.Rotate(- Routes_considered[routeidx]);
			peak_point = calc_peakpoint(point);
			DoLog(LOG_PASS, "pass to %d spd(%.2f %.2f) (d %.2f peak %.2f) point(%.2f %.2f)", Players_considered[controllers_of_line[routeidx][i]], speed_inf[routeidx][i], speed_sup[routeidx][i]
					,deliminator[routeidx][i], peak_point, point.x, point.y);
		}
	}
}