/********************************************************************************
 This module is designed to complete visual decision-making. Parts of confusing 
 hand codes, which assign priority, are removed.
*********************************************************************************/
/*
    Copyright (C) 2001  Tsinghuaeolus

    Authors : ChenJiang, YaoJinyi, CaiYunpeng, Lishi

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

	If you make any changes or have any comments we would appreciate a 
	message to yjy01@mails.tsinghua.edu.cn.
*/
#include "visualsystem.h"
#include "global.h"
#include "strategy.h"

/***********************   Visual System   *****************************************/
VisualSystem::VisualSystem(){
	b_validt = 0;
}

void VisualSystem::SetArounds(){
	int i, j;
	num_my_visible_players = 0;
	num_their_visible_players = 0;
	DoLog(LOG_SEE, "unk my %d", unknownmyplayerlist.numplayers);
	DoLog(LOG_SEE, "unk their %d", unknowntheirplayerlist.numplayers);
	DoLog(LOG_SEE, "teamless %d", teamlessplayerlist.numplayers);
/******   Get visible players list  ********************/
	for(i = 1; i <= SP_team_size; i++){
		if (i != MyNumber && MyPlayer(i).pos_valid()){
			MyVisiblePlayers[num_my_visible_players++] = i;
			DoLog(LOG_SEE,"My %d(%.2f,%.2f)",i,MyPlayer(i).pos.x,MyPlayer(i).pos.y);
		}
		if (TheirPlayer(i).pos_valid() || TheirPlayer(i).Is_goalie){
			TheirVisiblePlayers[num_their_visible_players++] = i + SP_team_size;
			DoLog(LOG_SEE,"Opp %d(%.2f,%.2f)",i,TheirPlayer(i).pos.x,TheirPlayer(i).pos.y);
			if(TheirPlayer(i).Is_goalie){
				DoLog("their goalie %.2f %.2f posconf %.2f", TheirPlayer(i).pos.x, TheirPlayer(i).pos.y, TheirPlayer(i).pos_conf);
			}
		}
	}
	
	for(i = 0; i < unknownmyplayerlist.numplayers; i ++){
		MyVisiblePlayers[num_my_visible_players++] = 2 * SP_team_size + i +1;
		DoLog(LOG_SEE,"My unk(%.2f,%.2f)",MyPlayer(2 * SP_team_size + i +1).pos.x,MyPlayer(2 * SP_team_size + i +1).pos.y);
		if (num_my_visible_players >= CP_max_visible_players)
			break;
	}
	for(i = 0; i < unknowntheirplayerlist.numplayers; i ++){
		TheirVisiblePlayers[num_their_visible_players++] = 4 * SP_team_size + i + 1;
		DoLog(LOG_SEE,"Their unk(%.2f,%.2f)",TheirPlayer(4* SP_team_size + i +1).pos.x,TheirPlayer(4 * SP_team_size + i +1).pos.y);
		if (num_their_visible_players >= CP_max_visible_players)
			break;
	}

	for(i = 0; i < teamlessplayerlist.numplayers; i ++){
		TheirVisiblePlayers[num_their_visible_players++] = 6 * SP_team_size + i + 1;
		DoLog(LOG_SEE,"Their teamless(%.2f,%.2f)",TheirPlayer(6* SP_team_size + i +1).pos.x,TheirPlayer(6 * SP_team_size + i +1).pos.y);
		if (num_their_visible_players >= CP_max_visible_players)
			break;
	}

	for(i = 0; i < num_my_visible_players; i++){
		MyPlayers_Close2Ball[i] = MyPlayers_Close2Me[i] = MyPlayers_InFront[i] = MyVisiblePlayers[i];
	}
	for(i = 0; i < num_my_visible_players - 1; i++){
		for ( j = i + 1; j < num_my_visible_players; j ++){
			if (MyPlayer(MyPlayers_Close2Ball[j]).balldist < MyPlayer(MyPlayers_Close2Ball[i]).balldist){
				Swap(MyPlayers_Close2Ball[i], MyPlayers_Close2Ball[j]);
			}
			if (MyPlayer(MyPlayers_Close2Me[j]).distance < MyPlayer(MyPlayers_Close2Me[i]).distance){
				Swap(MyPlayers_Close2Me[i], MyPlayers_Close2Me[j]);
			}
			if (MyPlayer(MyPlayers_InFront[j]).pos.x > MyPlayer(MyPlayers_InFront[i]).pos.x){
				Swap(MyPlayers_InFront[j], MyPlayers_InFront[i]);
			}
		}
	}

	for(i = 0; i < num_their_visible_players; i++){
		TheirPlayers_Danger[i] = TheirPlayers_Close2Ball[i] = TheirPlayers_Close2Me[i] = TheirPlayers_InFront[i] = TheirVisiblePlayers[i];
	}
	for(i = 0; i < num_their_visible_players - 1; i++){
		for ( j = i + 1; j < num_their_visible_players; j ++){
			if (TheirPlayer(TheirPlayers_Close2Ball[j]).balldist < TheirPlayer(TheirPlayers_Close2Ball[i]).balldist){
				Swap(TheirPlayers_Close2Ball[i], TheirPlayers_Close2Ball[j]);
			}
			if (TheirPlayer(TheirPlayers_Close2Me[j]).distance < TheirPlayer(TheirPlayers_Close2Me[i]).distance){
				Swap(TheirPlayers_Close2Me[i], TheirPlayers_Close2Me[j]);
			}
			if (TheirPlayer(TheirPlayers_InFront[j]).pos.x > TheirPlayer(TheirPlayers_InFront[i]).pos.x){
				Swap(TheirPlayers_InFront[j], TheirPlayers_InFront[i]);
			}
			if (TheirPlayer(TheirPlayers_Danger[j]).GetDefsensitivity()
				> TheirPlayer(TheirPlayers_Danger[i]).GetDefsensitivity())
				Swap(TheirPlayers_Danger[i],TheirPlayers_Danger[j]);
		}
	}

	for(i = 0; i < num_my_visible_players; i ++)
		MyPlayers_InBehind[i] = MyPlayers_InFront[num_my_visible_players - 1 - i];


	for(i = 0; i < num_their_visible_players; i++)
		TheirPlayers_InBehind[i] = TheirPlayers_InFront[num_their_visible_players - 1 - i];

	num_visible_players = num_my_visible_players + num_their_visible_players;

	i = 0; j = 0; int num_players_close2me = 0;
	while(i < num_my_visible_players || j < num_their_visible_players){
		if (i >= num_my_visible_players){
			while(j < num_their_visible_players){
				Players_Close2Me[num_players_close2me++] = TheirPlayers_Close2Me[j++];
			}
		}
		else if(j >= num_their_visible_players){
			while(i < num_my_visible_players){
				Players_Close2Me[num_players_close2me++] = MyPlayers_Close2Me[i++];
			}
		}
		else{
			if (MyPlayer_Close2Me(i).distance < TheirPlayer_Close2Me(j).distance)
				Players_Close2Me[num_players_close2me++] = MyPlayers_Close2Me[i++];
			else
				Players_Close2Me[num_players_close2me++] = TheirPlayers_Close2Me[j++];
		}
	}

	i = 0; j = 0; int num_players_close2ball = 0;
	while(i < num_my_visible_players || j < num_their_visible_players){
		if (i >= num_my_visible_players){
			while(j < num_their_visible_players){
				Players_Close2Ball[num_players_close2ball++] = TheirPlayers_Close2Ball[j++];
			}
		}
		else if(j >= num_their_visible_players){
			while(i < num_my_visible_players){
				Players_Close2Ball[num_players_close2ball++] = MyPlayers_Close2Ball[i++];
			}
		}
		else{
			if (MyPlayer_Close2Ball(i).balldist < TheirPlayer_Close2Ball(j).balldist)
				Players_Close2Ball[num_players_close2ball++] = MyPlayers_Close2Ball[i++];
			else
				Players_Close2Ball[num_players_close2ball++] = TheirPlayers_Close2Ball[j++];
		}
	}

	SetBallInf();
	DoLog(LOG_SEE, "setarounds complete");
}

/*find out who is closes to the ball */
void VisualSystem::SetBallInf(){
	UNum closestp = -1;
	float min_dist = 100.0f;
	if(num_my_visible_players > 0){
		if(MyPlayers_Close2Ball[0] == MyNumber && num_my_visible_players > 1){
			closestp = MyPlayers_Close2Ball[1];
		}
		else{
			closestp = MyPlayers_Close2Ball[0] ;
		}
		min_dist = MyPlayer(closestp).balldist;
		ball.Setball_CPlayer(closestp,MyPlayer(closestp).pos,situation.CurrentTime);
	}
	if(num_their_visible_players > 0){
		if(TheirPlayer(TheirPlayers_Close2Ball[0]).balldist < min_dist){
			closestp = TheirPlayers_Close2Ball[0] ;
			ball.Setball_CPlayer(closestp,TheirPlayer(closestp).pos,situation.CurrentTime);
		}
	}	
}

Player& VisualSystem::MyPlayer_Close2Me(int idx){
	return MyPlayer(MyPlayers_Close2Me[idx]);
}

Player& VisualSystem::TheirPlayer_Close2Me(int idx){
	return TheirPlayer(TheirPlayers_Close2Me[idx]);
}

Player& VisualSystem::GetPlayer_Close2Me(int idx){
	return GetPlayer(Players_Close2Me[idx]);
}

Player& VisualSystem::MyPlayer_Close2Ball(int idx){
	return MyPlayer(MyPlayers_Close2Ball[idx]);
}

Player& VisualSystem::TheirPlayer_Close2Ball(int idx){
	return TheirPlayer(TheirPlayers_Close2Ball[idx]);
}

Player& VisualSystem::GetPlayer_Close2Ball(int idx){
	return GetPlayer(Players_Close2Ball[idx]);
}

Player& VisualSystem::MyPlayer_InFront(int idx){
	return MyPlayer(MyPlayers_InFront[idx]);
}

Player& VisualSystem::MyPlayer_InBehind(int idx){
	return MyPlayer(MyPlayers_InBehind[idx]);
}

Player& VisualSystem::TheirPlayer_InFront(int idx){
	return TheirPlayer(TheirPlayers_InFront[idx]);
}

Player& VisualSystem::TheirPlayer_InBehind(int idx){
	return TheirPlayer(TheirPlayers_InBehind[idx]);
}

void VisualSystem::Raise(VisualObjType otype, AngleDeg angle, float priority){
	mediator.enroll(priority, angle);
	DoLog(LOG_VDEC, "R %.0f, %.3f", angle, priority);
}

void VisualSystem::Raise(VisualObjType otype, AngleDeg angle, AngleDeg distribute, float priority){
	Raise(otype, angle, 0.7f * priority);
	Raise(otype, angle - distribute, 0.4f * priority);
	Raise(otype, angle + distribute, 0.4f * priority);
}

/*raise a visual request for observing player*/
void VisualSystem::Raiseplayer(Player& p, float priority){
	float radius = p.ActiveRadius(situation.CurrentTime);
	float dist = Max(p.distance + 0.001f, radius);
	float spreadangle = ASin(radius / dist);

	if(p.Update_seen_time() != p.original_postime){//hearpos isn't accurate enough.
		spreadangle += 10.0f;
	} 
	if(p.Isforgot()){
		DoLog(LOG_VDEC, "find %d %.2f", p.InsideNO, priority);
		Findforgetobj(VO_player, p.global_angle, spreadangle, priority);
	}
	else
		Raise(VO_player, p.global_angle, spreadangle, priority);
}

/*raise a visual request for finding */
void VisualSystem::Findforgetobj(VisualObjType otype, AngleDeg angle, AngleDeg distribute, float maxpriority){
	AngleDeg tangle = angle - distribute;
	float priority;
	for(int i = 0; i < 4; i++){
		float conf = sensory.DirConf(tangle);
		if(conf < 0.9f && !Predictseen(tangle)){
			priority = Min(4.0f * (1 - conf), 1.0f);
			Raise(otype, tangle, maxpriority * priority * 0.5f);
		}
		tangle += distribute * 0.666f;
	}	
}

void VisualSystem::GetActionInf(){
	mediator.Getbestaction(bestaction);
	nextselfpos = Self.pos + Self.global_vel;
	if(bestaction.GetCMDType() == CMD_dash){
		nextselfpos += Polar2Vector(bestaction.GetPower() * action.Dashrate(), Self.bodyfacing);
	}
	DoLog(LOG_VDEC,"next self pos(%.2f, %.2f)", nextselfpos.x, nextselfpos.y);
}

void VisualSystem::DoVisualDecision(){
	PredicInfComing();
	Updatebvalidt();

	GetActionInf();
	
	if(FindBall())	return;

	float ball_priority = 0;
	float offlook_priority = 0;
	float deflook_priority = 0;
	float offside_priority = 0;
	float extendit_priority = 0;

//you should assign priority here.	

	bool canfeelball = Isballbyfeel();
	if(!canfeelball)
		AssureBall(ball_priority);

	DoLog(LOG_VDEC,"off %.2f, def %.2f", offlook_priority, deflook_priority);
	DoPassLook(offlook_priority);
	DoExtendITLook(extendit_priority);
	DoDefendLook(deflook_priority);	
	DoOffsideLook(offside_priority);

	DoViewModeDecision(canfeelball);
}

void VisualSystem::DoOffsideLook(float maxpriority){
	if(maxpriority <=0 || Self.attackness < 0.7f) return;//Self.pos.x <= 0 || Self.pos.x < ball.pos.x || 

	if(fieldinfo.IsOffside() || !fieldinfo.Maybeoffside()) return;
	UNum opp = fieldinfo.Susoffsideopp.ChkData(situation.CurrentTime);
	DoLog(LOG_VDEC, "suspicious offside %d", opp);
	Raiseplayer(TheirPlayer(opp), maxpriority * 0.5f);
	return;
}

bool VisualSystem::FindBall(){
/*	When Ball is missing, look for it	*/
	float CP_valid_dirconf = 0.77f;
	if(b_validt <= 0 && ball.pos_conf < CP_valid_dirconf){
		DoLog(LOG_VDEC, "find ball");
		if((Predictconf(ball.global_angle) < 0.8f || (!ball.Isforgot() && ball.pos_conf > 0.5f)) && !situation.ClockStopped){
			//ball maybe still in course
			float distribute = ATan2(1.5f * (situation.CurrentTime - ball.original_postime), ball.distance);
			distribute = Max(distribute, 20.0f);
			Raise(VO_ball, ball.global_angle, 2 * distribute,1.0f);
		}else{
			DoSeek(1.0f, ball.global_angle);
		}
		return true;
	}
 	return false;
}

/*****predict next visual inf within current view angle	*********/
/*****when clockstopped, prediction is invalid now		*********/
void VisualSystem::PredicInfComing(){
	Vinfo_coming = false;
	if(situation.SightTime != situation.CurrentTime && !situation.ClockStopped){
		if(Self.viewwidth.Data(situation.CurrentTime) == VW_Narrow
			|| (situation.SightTime == situation.CurrentTime - 2 && Self.viewwidth.Data(situation.CurrentTime) == VW_Normal)
			|| situation.SightTime < situation.CurrentTime - 2){
			Vinfo_coming = true;
			Vcominginfo_angle = Self.headfacing;
			Vcominginfo_width = sensory.MyViewWidth(Self.viewwidth.Data(situation.CurrentTime));
			DoLog(LOG_VDEC, "coming visual info (%.0f %.0f)" , NormalizeAngle(Vcominginfo_angle - Vcominginfo_width/2),
				NormalizeAngle(Vcominginfo_angle + Vcominginfo_width/2));
		}
	}
}

void VisualSystem::Updatebvalidt(){
	ITconfirm = false;
	if(b_validt <= 0){
		if(!ball.newpos){
			b_validt --;
		}else if(!ball.newvel){
			b_validt = ball.original_postime - situation.CurrentTime;
		}else{
			UpdateMinITcycles();
			b_validt = DirValidateball(lastITcycles);
		}
	}else{		
		UpdateMinITcycles();
		b_validt = DirValidateball(lastITcycles);
	}
	if(situation.ClockStopped) b_validt = Max(b_validt, 0); 
	DoLog(LOG_VDEC, "valid time %d", b_validt);
	if(num_visible_players > 0 && b_validt > 0){
		if(GetPlayer_Close2Ball(0).balldist - 0.5f * Max(1.0f, ball.posgap.ChkData(situation.CurrentTime).mod()) < SP_kickable_area){
			DoLog(LOG_VDEC, "modify valid t %d %.2f %.2f", GetPlayer_Close2Ball(0).InsideNO, GetPlayer_Close2Ball(0).balldist, ball.posgap.ChkData(situation.CurrentTime).mod());
			b_validt = 0;
		}
	}
} 

void VisualSystem::UpdateMinITcycles(){
	if(!situation.BallFree){
		lastITplayer = situation.ballcontroller;
		lastITcycles = 0;
	}else{
		//continue record it player
		if(!ball.newpos && int(situation.min_IT_cycles) >= lastITcycles &&
			lastITplayer > 0 && lastITplayer < 23 && !GetPlayer(lastITplayer).newpos){
			lastITcycles --;
			DoLog(LOG_VDEC, "special deal");
		}else{
			lastITplayer = situation.most_promising_controller;
			lastITcycles = (int)situation.min_IT_cycles;
		}
	}
	DoLog(LOG_VDEC,"IT player%d cycles %d", lastITplayer, lastITcycles);
}

int VisualSystem::DirValidateball(int maxcycles){
	maxcycles = Min(maxcycles, CP_max_bvalidcycles);
	if(maxcycles <= 0) return maxcycles;
	
	float CP_valid_dirconf = 0.81f;
	Vector pos[CP_max_bvalidcycles];
	ball.PredictPos(pos, maxcycles); 
	float leftang = ball.global_angle;
	float rightang = leftang;
	Vector relpos;
	float tmpang1, tmpang2, radius, dist;
	for(int i = 0; i < maxcycles; i++){
		radius = i + 1.0f;
		relpos = pos[i] - Self.pos;
		dist = relpos.mod() + CP_max_bvalidcycles;
		tmpang1 = relpos.Angle();
		tmpang2 = ASin(radius / dist);
		if(leftang > tmpang1 - tmpang2){
			leftang = tmpang1 - tmpang2;
			if(sensory.DirConf(leftang) < CP_valid_dirconf){
				ITconfirm = true;
				ITconfirmagl = NormalizeAngle(leftang - 10);
				return i+1;
			}
		}
		if(rightang < tmpang1 + tmpang2){
			rightang = tmpang1 + tmpang2;
			if(sensory.DirConf(leftang) < CP_valid_dirconf){
				ITconfirm = true;
				ITconfirmagl = NormalizeAngle(rightang + 10); 
				return i+1;
			}
		}
	}
	return maxcycles;
}

float VisualSystem::GetPromisingAngle(VisualObjType obj, UNum idx){
	MobileObject* mobj = NULL;
	if(obj == VO_ball){
		mobj = &ball;
	}else if(obj == VO_player){
		mobj = &GetPlayer(idx);
	}
	if(mobj != NULL){
		if(mobj->pos_conf > CP_min_valid_conf){
			return mobj->global_angle; 
		}else{//
		}
	}else{
		//shouldn't be here;
	}
	return 0;
}

void VisualSystem::DoSeek(float maxpriority, AngleDeg recommendedangle){
	int CP_seek_division = 12;
	for(int i = 0; i < CP_seek_division; i++){
		float angle = recommendedangle + 360.0f/CP_seek_division * i;
		float conf = Predictconf(angle);
		if(conf > 0.9f) continue;
		float priority = 0.3f + 0.7f * (1 - conf);
		priority += 0.1f - 0.2f * (float)fabs(NormalizeAngle(360.0f/CP_seek_division * i)) / 180.0f; 
		Raise(VO_route, angle, priority * maxpriority);
	}
}

void VisualSystem::DoExtendITLook(float maxpriority){
	if(!ITconfirm || maxpriority <= 0 || Predictseen(ITconfirmagl)) return;

	maxpriority *= 0.25f;
	Raise(VO_route, ITconfirmagl, maxpriority);
	DoLog(LOG_VDEC, "extend it %.2f priority %.2f", ITconfirmagl, maxpriority);
}

float VisualSystem::Predictconf(float ang){
	if(Predictseen(ang))
		return 1.0f;
	else
		return sensory.DirConf(ang);
}

bool VisualSystem::Predictseen(float ang){
	if(Vinfo_coming && fabs(NormalizeAngle(ang - Vcominginfo_angle)) < Vcominginfo_width)
		return true;
	else
		return false;
}

void VisualSystem::AssureBall(float priority){	
	Vector relpos;
	float ballangle, distribute_angle, radius, dist;
	if(!situation.BallFree){
		relpos = ball.Pos(ball.original_postime) - nextselfpos;
		ballangle = relpos.Angle();
		dist = relpos.mod() + 0.001f;
		radius = (situation.CurrentTime - ball.original_postime + 1) * SP_ball_speed_max; 
		radius = Min(radius, dist);
		distribute_angle = ASin(radius / dist);
	}else if(b_validt > 0){
		ballangle = (ball.PredictPos(1) - nextselfpos).Angle();
		distribute_angle = 5;
	}else{
		ballangle = ball.global_angle;
		distribute_angle = 15;
	}

	
	if(!ball.Isforgot()){
		Raise(VO_ball, ballangle, distribute_angle, priority);
	}else{
		DoLog(LOG_VDEC, "ball is forgot");
		if(ball.Rs_valid())
			Findforgetobj(VO_ball, ballangle, distribute_angle, priority);
		else
			FindBall();
	}	
	
}


void VisualSystem::DoPassLook(float maxpriority){
	if(maxpriority <= 0) return;
	float priority;
	int i;

	for(i = 1; i <= SP_team_size; i++){
		if(MyPlayer(i).original_postime == 0) continue;//doesn't exit player

		priority = LookTeammatePriority(MyPlayer(i));
		if(priority <= 0.0001f) continue;
		Raiseplayer(MyPlayer(i), maxpriority * priority);
	}

	for(i = 1; i <= SP_team_size; i++){		
		if(TheirPlayer(i).original_postime == 0) continue;

		priority = LookOppPriority(TheirPlayer(i));
		if(priority <= 0.0001f) continue;
		Raiseplayer(TheirPlayer(i), maxpriority * priority * 0.7f);
	}
	//for shoot
	if((ball.kickable() || bestaction.GetActionType() == Action_interception) && fieldinfo.WithInShootableArea(Self.pos) ){
		//Rect shoot_area(SP_semi_pitch_length, 30, SP_penalty_area_width /2, - SP_penalty_area_width /2);
		bool maybe_shoot;
		if(bestaction.GetActionType() == Action_interception){
			maybe_shoot = bool(motion.GuessShoot(Self.IT_inf.IT_point, (int)Self.IT_inf.IT_cycles) > 0);
		}else if(bestaction.GetActionType() ==  Action_shoot){
			maybe_shoot = true;
		}else{
			maybe_shoot = bool(motion.GuessShoot(ball.pos + Vector(1,3), 0) >0 || motion.GuessShoot(ball.pos + Vector(1, -3), 0) >0);
		}

		if(maybe_shoot){
			//pay some more attention to goalie
		}else{
			DoLog(LOG_VDEC, "don't intend to shoot");
		}
	}
	//for pass
	if(motion.WaitPassLook()){
		DoLog(LOG_VDEC, "look at pass receiver %.2f", maxpriority * 2.0f);
		Raiseplayer(MyPlayer(motion.GetReceiver()), maxpriority * 2.0f);
		Raise(VO_route, motion.GetPassAngle(), maxpriority * 0.3f);
	}
}

float VisualSystem::LookTeammatePriority(Player& player){
	if(player.distance > CP_passable_distance) return 0.0f;

	return 0.0f;
}

float VisualSystem::LookOppPriority(Player& player){
	if(player.distance > CP_passable_distance) return 0.0f;
	
	return 0.0f;
}

float VisualSystem::DefLookOppPry(Player& player){
	return 0.0f;
}

float VisualSystem::GetPriority(float concern, float conf){
	if(concern < 0.1) return 0;
	return float(concern * sqrt(conf));
}

void VisualSystem::DoDefendLook(float maxpriority){
	if(maxpriority <= 0) return;
	for(int i = 1; i <= SP_team_size; i++){
		if(TheirPlayer(i).distance > CP_def_concern_area) continue;
		float priority = DefLookOppPry(TheirPlayer(i));
		if(priority < 0.001f) continue;
		Raiseplayer(TheirPlayer(i), maxpriority * priority / 2);
	}
}

void VisualSystem::DoGoalieLook(float maxpriority){
	float concern, conf, priority;
	for(int i = 1; i<= SP_team_size; i++){
		if(TheirPlayer(i).original_postime == 0) continue;

		if(TheirPlayer(i).GetDefsensitivity() < 0.6f) continue;
		
		concern = Sqr(TheirPlayer(i).GetDefsensitivity());
		if(!Isobjconfmax(TheirPlayer(i))){
			conf = (1 - TheirPlayer(i).pos_conf) / (1 - CP_min_valid_conf + 0.2f);
			conf = Min(conf, 1.0f);
			priority = GetPriority(concern, conf);
		}else
			priority = concern * 0.1f;

		if(priority <= 0.0001f) continue;
		DoLog(LOG_VDEC,"opp%d, concern%.2f, conf%.2f p%.2f", i, concern, conf, priority);
		Raise(VO_player, TheirPlayer(i).global_angle, priority * maxpriority);
	}
} 

bool VisualSystem::Isballbyfeel(){
	if(ball.distance > SP_feel_distance + SP_ball_speed_max) return false;

	if(ball.kickable())	return true;

	Vector ballgap = ball.posgap.ChkData(situation.CurrentTime) + ball.velgap.ChkData(situation.CurrentTime);
	if((ball.PredictPos(1) - nextselfpos).mod() < SP_feel_distance - ballgap.mod() && b_validt > 0)
		return true;
	else
		return false;
}

bool VisualSystem::Isobjconfmax(MobileObject& obj){
	if((situation.CurrentTime - obj.original_postime < 2 && obj.Rs_valid())
		|| (Predictseen(obj.global_angle) && !obj.Isforgot()))
		return true;
	else
		return false;
}

void VisualSystem::DoViewModeDecision(bool forcenarrow){
	if(Self.Is_goalie && situation.IsGoalieKick){
		mediator.SetViewWidth(VW_Wide);
		return;
	} 
	if(ball.distance > 50.0f){
		mediator.SetViewWidth(VW_Wide);
		return;
	}else if(ball.distance > 30.0f){
		mediator.SetViewWidth(VW_Normal);
		return;
	}
	
	if(Self.viewwidth.Data(situation.CurrentTime) == VW_Normal){
		float prevtimegap = watch.TimeElapsedSight(); 
		if(prevtimegap > SP_send_step * 0.5f){
			mediator.SetViewWidth(VW_Normal);
			DoLog(LOG_VDEC, "keep on visual mode %.2f", prevtimegap);
			return;
		}
	}

	if(forcenarrow){
		mediator.SetViewWidth(VW_Narrow);
		return;
	}
	if(sensory.LastSightTime != situation.CurrentTime && Self.viewwidth.Data(situation.CurrentTime) == VW_Narrow){
		if(!ball.Isforgot() && fabs(NormalizeAngle(ball.global_angle - Self.headfacing)) < sensory.MyViewWidth(VW_Narrow)){
			mediator.SetViewWidth(VW_Narrow);
			DoLog(LOG_VDEC, "keep on visual mode");
		}
	}
}
